!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
! 
#include "fftgrid.h"

program lpt2dx
  use fft_3d
  use util
  implicit none
  integer, parameter:: nfin = 10
  real(kind=8),parameter:: p2 = 6.28318530717958648d0
  integer:: iorg(3)
  character(len=8):: magic = 'TAPP:lpt'

  character(len=8):: sign
  complex(kind=8),allocatable:: vg(:,:,:,:)
  real(kind=8),allocatable:: vr(:,:,:,:), ztmp1(:,:,:,:), ztmp2(:,:,:,:)
  type (fft3_struct) :: fs
  integer:: nvx, nvy, nvz, nrx, nry, nrz, nlx, nly, nlz
  real(kind=8):: bb(3,3), aa(3,3), w
  integer:: i1, i2, i3, k1, k2, k3, ios
  character(len=256):: ifname, ofname, name, argv(7)
  logical:: error
  integer:: ic, i
  logical:: multiq, nextq
  character(len=256):: acount
  real(kind=8):: scl
  integer:: nspin, is

  ic = 1
  multiq = .false.
  call getarg(ic, argv(ic))
  if ( argv(ic) .eq. '-m' ) then
     multiq = .true.
     ic = ic + 1
  end if
  if ( argv(ic) .eq. '-s' ) then
     ic = ic + 1
     call getarg(ic, argv(ic))
     read(argv(ic),*) scl
     if (scl .lt. 1.0d0) then
        write(6,*) 'scl < 1.0'
        stop
     end if
     ic = ic + 1
  else
     scl = 1.0d0
  end if
  call getarg(ic, argv(ic))
  if (len_trim(argv(ic)) .eq. 0) then
     write(6,*) 'invalid fname'
     stop
  end if
  ifname = argv(ic)
  ic = ic + 1
  do i = ic, ic + 2
     call getarg(i, argv(i))
  end do
  read(argv(ic),*,iostat=ios) iorg(1)
  if (ios .eq. 0) then
     error = .false.
     read(argv(ic+1),*,iostat=ios) iorg(2)
     error = error .or. (ios .ne. 0)
     read(argv(ic+2),*,iostat=ios) iorg(3)
     error = error .or. (ios .ne. 0)
     if (error) then
        write(6,*) 'invalid position of origin'
        stop
     end if
  else
     iorg(1) = 0
     iorg(2) = 0
     iorg(3) = 0
  end if
  
  open(nfin, file=ifname, status='old', action='read', form='unformatted')
  if (multiq) then
     ic = 1
  end if
  read(nfin) sign
  if (sign .ne. magic) then
     write(6,*) 'Bad magic'
     stop
  end if
  nextq = .true.
  do while(nextq)
     read(nfin) bb, nspin
     read(nfin) nvx, nvy, nvz

     bb = bb/p2
     w = bb(1,1)*(bb(2,2)*bb(3,3)-bb(3,2)*bb(2,3)) &
          +bb(2,1)*(bb(3,2)*bb(1,3)-bb(1,2)*bb(3,3)) &
          +bb(3,1)*(bb(1,2)*bb(2,3)-bb(2,2)*bb(1,3))
     aa(1,1)=(bb(2,2)*bb(3,3)-bb(3,2)*bb(2,3))/w
     aa(2,1)=(bb(3,2)*bb(1,3)-bb(1,2)*bb(3,3))/w
     aa(3,1)=(bb(1,2)*bb(2,3)-bb(2,2)*bb(1,3))/w
     aa(1,2)=(bb(2,3)*bb(3,1)-bb(3,3)*bb(2,1))/w
     aa(2,2)=(bb(3,3)*bb(1,1)-bb(1,3)*bb(3,1))/w
     aa(3,2)=(bb(1,3)*bb(2,1)-bb(2,3)*bb(1,1))/w
     aa(1,3)=(bb(2,1)*bb(3,2)-bb(3,1)*bb(2,2))/w
     aa(2,3)=(bb(3,1)*bb(1,2)-bb(1,1)*bb(3,2))/w
     aa(3,3)=(bb(1,1)*bb(2,2)-bb(2,1)*bb(1,2))/w
     nrx = algn235(nint(scl*(nvx+1)),1)
     nry = algn235(nint(scl*(nvy+1)),1)
     nrz = algn235(nint(scl*(nvz+1)),1)
     nlx = FFT_PAD_NX(nrx*2)
     nly = FFT_PAD_NY(nry*2)
     nlz = FFT_PAD_NZ(nrz*2)
     allocate(vg(-nvx:nvx, -nvy:nvy, -nvz:nvz, nspin), &
          vr(nlx, nly, nlz, nspin), FFT_A3D_DEFSZ(ztmp1, nlx, nly, nlz), &
          FFT_A3D_DEFSZ(ztmp2, nlx, nly, nlz))
     read(nfin) vg


     call fft3_init(nrx*2,nry*2,nrz*2,nlx,nly,nlz,fs)

     do is = 1,nspin
        ztmp1 = 0.0d0
        do i3=-nvz,nvz
           do i2=-nvy,nvy
              do i1=-nvx,nvx
                 k1 = modulo(i1, nrx*2) + 1
                 k2 = modulo(i2, nry*2) + 1
                 k3 = modulo(i3, nrz*2) + 1
                 FFT_A3D_RL(ztmp1, k1, k2, k3) = dble(vg(i1, i2, i3, is))
                 FFT_A3D_IM(ztmp1, k1, k2, k3) = aimag(vg(i1, i2, i3, is))
              end do
           end do
        end do
        call fft3_bw(fs, ztmp1, ztmp2)

        vr(:,:,:,is) = FFT_A3D_RL(ztmp1,:,:,:)
     end do
     deallocate(ztmp1, ztmp2)

     call fft3_done(fs)

     if (multiq) then
        write(acount, '(''.'',i3.3)') ic
        ofname=trim(ifname)//trim(acount)//'.dx'//achar(0)
     else
        ofname=trim(ifname)//'.dx'//achar(0)
     end if
     name='TAPP lpt'//achar(0)
     call outfldpr(ofname, name, nrx*2, nry*2, nrz*2, aa, iorg, &
          nspin, nlx, nly, nlz, vr, 0)
     deallocate(vg, vr)

     if (.not. multiq) then
        exit
     end if

     read(nfin, iostat=ios) sign
     if (ios .ne. 0) then
        nextq = .false.
     else if (sign .ne. magic) then
        write(6,*) 'bad magic: ', sign
        stop
     end if
     ic = ic + 1
  end do
end program lpt2dx
