!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
! 
module pfft3d_const
  integer,parameter:: nfmax = 30
end module pfft3d_const

module pfft3d_subr
  use pfft3d_const
contains
  subroutine fftrd2(nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nvk,np
    real(kind=8),intent(in):: xr(nvk,np,2),xi(nvk,np,2)
    real(kind=8),intent(out):: yr(nvk,2,np),yi(nvk,2,np)
    real(kind=8),intent(in):: w(2,np)
    !
    real(kind=8):: wr1,wi1,tr,ti
    integer:: p,ivk
    !
    do p = 1,np
       wr1 = w(1,p)
       wi1 = w(2,p)
       do ivk = 1,nvk
          yr(ivk,1,p) = xr(ivk,p,1) + xr(ivk,p,2)
          yi(ivk,1,p) = xi(ivk,p,1) + xi(ivk,p,2)
          !
          tr = xr(ivk,p,1) - xr(ivk,p,2)
          ti = xi(ivk,p,1) - xi(ivk,p,2)
          yr(ivk,2,p) = wr1*tr - wi1*ti
          yi(ivk,2,p) = wr1*ti + wi1*tr
       end do
    end do
    return
  end subroutine fftrd2
  !
  subroutine bftrd2(nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nvk,np
    real(kind=8),intent(in):: xr(nvk,np,2),xi(nvk,np,2)
    real(kind=8),intent(out):: yr(nvk,2,np),yi(nvk,2,np)
    real(kind=8),intent(in):: w(2,np)
    !
    real(kind=8):: wr1,wi1,tr,ti
    integer:: p,ivk
    !
    do p = 1,np
       wr1 = w(1,p)
       wi1 = w(2,p)
       do ivk = 1,nvk
          yr(ivk,1,p) = xr(ivk,p,1) + xr(ivk,p,2)
          yi(ivk,1,p) = xi(ivk,p,1) + xi(ivk,p,2)
          !
          tr = xr(ivk,p,1) - xr(ivk,p,2)
          ti = xi(ivk,p,1) - xi(ivk,p,2)
          yr(ivk,2,p) = wr1*tr + wi1*ti
          yi(ivk,2,p) = wr1*ti - wi1*tr
       end do
    end do
    return
  end subroutine bftrd2
  !
  subroutine fftrd3(nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nvk,np
    real(kind=8),intent(in):: xr(nvk,np,3),xi(nvk,np,3)
    real(kind=8),intent(out):: yr(nvk,3,np),yi(nvk,3,np)
    real(kind=8),intent(in):: w(2,2*np)
    !
    real(kind=8),parameter:: sp3 = 0.866025403784438647d0
    !
    real(kind=8):: wr1,wi1,wr2,wi2,tr1,ti1,tr2,ti2,tr3,ti3,tr,ti
    integer:: p,ivk,iw
    !
    iw = 1
    do p = 1,np
       wr1 = w(1,iw)
       wi1 = w(2,iw)
       iw = iw+1
       wr2 = w(1,iw)
       wi2 = w(2,iw)
       iw = iw+1
       do ivk = 1,nvk
          tr1 = xr(ivk,p,2) + xr(ivk,p,3)
          ti1 = xi(ivk,p,2) + xi(ivk,p,3)
          tr2 = xr(ivk,p,1) - 0.5d0*tr1
          ti2 = xi(ivk,p,1) - 0.5d0*ti1
          tr3 = sp3*(xi(ivk,p,2) - xi(ivk,p,3))
          ti3 = -sp3*(xr(ivk,p,2) - xr(ivk,p,3))
          yr(ivk,1,p) = xr(ivk,p,1) + tr1
          yi(ivk,1,p) = xi(ivk,p,1) + ti1
          tr = tr2 + tr3
          ti = ti2 + ti3
          yr(ivk,2,p) = wr1*tr - wi1*ti
          yi(ivk,2,p) = wr1*ti + wi1*tr
          tr = tr2 - tr3
          ti = ti2 - ti3
          yr(ivk,3,p) = wr2*tr - wi2*ti
          yi(ivk,3,p) = wr2*ti + wi2*tr
       end do
    end do
    return
  end subroutine fftrd3
  !
  subroutine bftrd3(nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nvk,np
    real(kind=8),intent(in):: xr(nvk,np,3),xi(nvk,np,3)
    real(kind=8),intent(out):: yr(nvk,3,np),yi(nvk,3,np)
    real(kind=8),intent(in):: w(2,2*np)
    !
    real(kind=8),parameter:: sp3 = 0.866025403784438647d0
    !
    real(kind=8):: wr1,wi1,wr2,wi2,tr1,ti1,tr2,ti2,tr3,ti3,tr,ti
    integer:: p,ivk,iw
    !
    iw = 1
    do p = 1,np
       wr1 = w(1,iw)
       wi1 = w(2,iw)
       iw = iw+1
       wr2 = w(1,iw)
       wi2 = w(2,iw)
       iw = iw+1
       do ivk = 1,nvk
          tr1 = xr(ivk,p,2) + xr(ivk,p,3)
          ti1 = xi(ivk,p,2) + xi(ivk,p,3)
          tr2 = xr(ivk,p,1) - 0.5d0*tr1
          ti2 = xi(ivk,p,1) - 0.5d0*ti1
          tr3 = -sp3*(xi(ivk,p,2) - xi(ivk,p,3))
          ti3 = sp3*(xr(ivk,p,2) - xr(ivk,p,3))
          yr(ivk,1,p) = xr(ivk,p,1) + tr1
          yi(ivk,1,p) = xi(ivk,p,1) + ti1
          tr = tr2 + tr3
          ti = ti2 + ti3
          yr(ivk,2,p) = wr1*tr + wi1*ti
          yi(ivk,2,p) = wr1*ti - wi1*tr
          tr = tr2 - tr3
          ti = ti2 - ti3
          yr(ivk,3,p) = wr2*tr + wi2*ti
          yi(ivk,3,p) = wr2*ti - wi2*tr
       end do
    end do
    return
  end subroutine bftrd3
  !
  subroutine fftrd4(nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nvk,np
    real(kind=8),intent(in):: xr(nvk,np,4),xi(nvk,np,4)
    real(kind=8),intent(out):: yr(nvk,4,np),yi(nvk,4,np)
    real(kind=8),intent(in):: w(2,3*np)
    !
    real(kind=8):: wr1,wi1,wr2,wi2,wr3,wi3
    real(kind=8):: tr1,ti1,tr2,ti2,tr3,ti3,tr4,ti4,tr,ti
    integer:: p,ivk,iw
    !
    iw = 1
    do p = 1,np
       wr1 = w(1,iw)
       wi1 = w(2,iw)
       iw = iw+1
       wr2 = w(1,iw)
       wi2 = w(2,iw)
       iw = iw+1
       wr3 = w(1,iw)
       wi3 = w(2,iw)
       iw = iw+1
       do ivk = 1,nvk
          tr1 = xr(ivk,p,1) + xr(ivk,p,3)
          ti1 = xi(ivk,p,1) + xi(ivk,p,3)
          tr2 = xr(ivk,p,1) - xr(ivk,p,3)
          ti2 = xi(ivk,p,1) - xi(ivk,p,3)
          tr3 = xr(ivk,p,2) + xr(ivk,p,4)
          ti3 = xi(ivk,p,2) + xi(ivk,p,4)
          tr4 = xi(ivk,p,2) - xi(ivk,p,4)
          ti4 = xr(ivk,p,4) - xr(ivk,p,2)
          yr(ivk,1,p) = tr1 + tr3
          yi(ivk,1,p) = ti1 + ti3
          tr = tr2 + tr4
          ti = ti2 + ti4
          yr(ivk,2,p) = wr1*tr - wi1*ti
          yi(ivk,2,p) = wr1*ti + wi1*tr
          tr = tr1 - tr3
          ti = ti1 - ti3
          yr(ivk,3,p) = wr2*tr - wi2*ti
          yi(ivk,3,p) = wr2*ti + wi2*tr
          tr = tr2 - tr4
          ti = ti2 - ti4
          yr(ivk,4,p) = wr3*tr - wi3*ti
          yi(ivk,4,p) = wr3*ti + wi3*tr
       end do
    end do
    return
  end subroutine fftrd4
  !
  subroutine bftrd4(nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nvk,np
    real(kind=8),intent(in):: xr(nvk,np,4),xi(nvk,np,4)
    real(kind=8),intent(out):: yr(nvk,4,np),yi(nvk,4,np)
    real(kind=8),intent(in):: w(2,3*np)
    !
    real(kind=8):: wr1,wi1,wr2,wi2,wr3,wi3
    real(kind=8):: tr1,ti1,tr2,ti2,tr3,ti3,tr4,ti4,tr,ti
    integer:: p,ivk,iw
    !
    iw = 1
    do p = 1,np
       wr1 = w(1,iw)
       wi1 = w(2,iw)
       iw = iw+1
       wr2 = w(1,iw)
       wi2 = w(2,iw)
       iw = iw+1
       wr3 = w(1,iw)
       wi3 = w(2,iw)
       iw = iw+1
       do ivk = 1,nvk
          tr1 = xr(ivk,p,1) + xr(ivk,p,3)
          ti1 = xi(ivk,p,1) + xi(ivk,p,3)
          tr2 = xr(ivk,p,1) - xr(ivk,p,3)
          ti2 = xi(ivk,p,1) - xi(ivk,p,3)
          tr3 = xr(ivk,p,2) + xr(ivk,p,4)
          ti3 = xi(ivk,p,2) + xi(ivk,p,4)
          tr4 = xi(ivk,p,4) - xi(ivk,p,2)
          ti4 = xr(ivk,p,2) - xr(ivk,p,4)
          yr(ivk,1,p) = tr1 + tr3
          yi(ivk,1,p) = ti1 + ti3
          tr = tr2 + tr4
          ti = ti2 + ti4
          yr(ivk,2,p) = wr1*tr + wi1*ti
          yi(ivk,2,p) = wr1*ti - wi1*tr
          tr = tr1 - tr3
          ti = ti1 - ti3
          yr(ivk,3,p) = wr2*tr + wi2*ti
          yi(ivk,3,p) = wr2*ti - wi2*tr
          tr = tr2 - tr4
          ti = ti2 - ti4
          yr(ivk,4,p) = wr3*tr + wi3*ti
          yi(ivk,4,p) = wr3*ti - wi3*tr
       end do
    end do
    return
  end subroutine bftrd4
  !
  subroutine fftrd5(nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nvk,np
    real(kind=8),intent(in):: xr(nvk,np,5),xi(nvk,np,5)
    real(kind=8),intent(out):: yr(nvk,5,np),yi(nvk,5,np)
    real(kind=8),intent(in):: w(2,4*np)
    !
    real(kind=8),parameter:: sp25 = 0.951056516295153572d0
    real(kind=8),parameter:: sq54 = 0.559016994374947424d0
    real(kind=8),parameter:: ss = 0.618033988749894848d0
    !
    real(kind=8):: wr1,wi1,wr2,wi2,wr3,wi3,wr4,wi4
    real(kind=8):: tr1,ti1,tr2,ti2,tr3,ti3,tr4,ti4,tr5,ti5
    real(kind=8):: tr6,ti6,tr7,ti7,tr8,ti8,tr9,ti9,tr10,ti10,tr11,ti11
    real(kind=8):: tr,ti
    integer:: p,ivk,iw
    !
    do ivk = 1,nvk
       tr1 = xr(ivk,1,2) + xr(ivk,1,5)
       ti1 = xi(ivk,1,2) + xi(ivk,1,5)
       tr2 = xr(ivk,1,3) + xr(ivk,1,4)
       ti2 = xi(ivk,1,3) + xi(ivk,1,4)
       tr3 = sp25*(xr(ivk,1,2) - xr(ivk,1,5))
       ti3 = sp25*(xi(ivk,1,2) - xi(ivk,1,5))
       tr4 = sp25*(xr(ivk,1,3) - xr(ivk,1,4))
       ti4 = sp25*(xi(ivk,1,3) - xi(ivk,1,4))
       tr5 = tr1 + tr2
       ti5 = ti1 + ti2
       tr6 = sq54*(tr1 - tr2)
       ti6 = sq54*(ti1 - ti2)
       tr7 = xr(ivk,1,1) - 0.25d0*tr5
       ti7 = xi(ivk,1,1) - 0.25d0*ti5
       tr8 = tr7 + tr6
       ti8 = ti7 + ti6
       tr9 = tr7 - tr6
       ti9 = ti7 - ti6
       tr10 = ti3 + ss*ti4
       ti10 = - tr3 - ss*tr4
       tr11 = ss*ti3 - ti4
       ti11 = - ss*tr3 + tr4
       yr(ivk,1,1) = xr(ivk,1,1) + tr5
       yi(ivk,1,1) = xi(ivk,1,1) + ti5
       yr(ivk,2,1) = tr8 + tr10
       yi(ivk,2,1) = ti8 + ti10
       yr(ivk,3,1) = tr9 + tr11
       yi(ivk,3,1) = ti9 + ti11
       yr(ivk,4,1) = tr9 - tr11
       yi(ivk,4,1) = ti9 - ti11
       yr(ivk,5,1) = tr8 - tr10
       yi(ivk,5,1) = ti8 - ti10
    end do

    iw = 5
    do p = 2,np
       wr1 = w(1,iw)
       wi1 = w(2,iw)
       iw = iw+1
       wr2 = w(1,iw)
       wi2 = w(2,iw)
       iw = iw+1
       wr3 = w(1,iw)
       wi3 = w(2,iw)
       iw = iw+1
       wr4 = w(1,iw)
       wi4 = w(2,iw)
       iw = iw+1
       do ivk = 1,nvk
          tr1 = xr(ivk,p,2) + xr(ivk,p,5)
          ti1 = xi(ivk,p,2) + xi(ivk,p,5)
          tr2 = xr(ivk,p,3) + xr(ivk,p,4)
          ti2 = xi(ivk,p,3) + xi(ivk,p,4)
          tr3 = sp25*(xr(ivk,p,2) - xr(ivk,p,5))
          ti3 = sp25*(xi(ivk,p,2) - xi(ivk,p,5))
          tr4 = sp25*(xr(ivk,p,3) - xr(ivk,p,4))
          ti4 = sp25*(xi(ivk,p,3) - xi(ivk,p,4))
          tr5 = tr1 + tr2
          ti5 = ti1 + ti2
          tr6 = sq54*(tr1 - tr2)
          ti6 = sq54*(ti1 - ti2)
          tr7 = xr(ivk,p,1) - 0.25d0*tr5
          ti7 = xi(ivk,p,1) - 0.25d0*ti5
          tr8 = tr7 + tr6
          ti8 = ti7 + ti6
          tr9 = tr7 - tr6
          ti9 = ti7 - ti6
          tr10 = ti3 + ss*ti4
          ti10 = - tr3 - ss*tr4
          tr11 = ss*ti3 - ti4
          ti11 = - ss*tr3 + tr4
          yr(ivk,1,p) = xr(ivk,p,1) + tr5
          yi(ivk,1,p) = xi(ivk,p,1) + ti5
          tr = tr8 + tr10
          ti = ti8 + ti10
          yr(ivk,2,p) = wr1*tr - wi1*ti
          yi(ivk,2,p) = wr1*ti + wi1*tr
          tr = tr9 + tr11
          ti = ti9 + ti11
          yr(ivk,3,p) = wr2*tr - wi2*ti
          yi(ivk,3,p) = wr2*ti + wi2*tr
          tr = tr9 - tr11
          ti = ti9 - ti11
          yr(ivk,4,p) = wr3*tr - wi3*ti
          yi(ivk,4,p) = wr3*ti + wi3*tr
          tr = tr8 - tr10
          ti = ti8 - ti10
          yr(ivk,5,p) = wr4*tr - wi4*ti
          yi(ivk,5,p) = wr4*ti + wi4*tr
       end do
    end do
    return
  end subroutine fftrd5
  !
  subroutine bftrd5(nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nvk,np
    real(kind=8),intent(in):: xr(nvk,np,5),xi(nvk,np,5)
    real(kind=8),intent(out):: yr(nvk,5,np),yi(nvk,5,np)
    real(kind=8),intent(in):: w(2,4*np)
    !
    real(kind=8),parameter:: sp25 = 0.951056516295153572d0
    real(kind=8),parameter:: sq54 = 0.559016994374947424d0
    real(kind=8),parameter:: ss = 0.618033988749894848d0
    !
    real(kind=8):: wr1,wi1,wr2,wi2,wr3,wi3,wr4,wi4
    real(kind=8):: tr1,ti1,tr2,ti2,tr3,ti3,tr4,ti4,tr5,ti5
    real(kind=8):: tr6,ti6,tr7,ti7,tr8,ti8,tr9,ti9,tr10,ti10,tr11,ti11
    real(kind=8):: tr,ti
    integer:: p,ivk,iw
    !
    do ivk = 1,nvk
       tr1 = xr(ivk,1,2) + xr(ivk,1,5)
       ti1 = xi(ivk,1,2) + xi(ivk,1,5)
       tr2 = xr(ivk,1,3) + xr(ivk,1,4)
       ti2 = xi(ivk,1,3) + xi(ivk,1,4)
       tr3 = sp25*(xr(ivk,1,2) - xr(ivk,1,5))
       ti3 = sp25*(xi(ivk,1,2) - xi(ivk,1,5))
       tr4 = sp25*(xr(ivk,1,3) - xr(ivk,1,4))
       ti4 = sp25*(xi(ivk,1,3) - xi(ivk,1,4))
       tr5 = tr1 + tr2
       ti5 = ti1 + ti2
       tr6 = sq54*(tr1 - tr2)
       ti6 = sq54*(ti1 - ti2)
       tr7 = xr(ivk,1,1) - 0.25d0*tr5
       ti7 = xi(ivk,1,1) - 0.25d0*ti5
       tr8 = tr7 + tr6
       ti8 = ti7 + ti6
       tr9 = tr7 - tr6
       ti9 = ti7 - ti6
       tr10 = - ti3 - ss*ti4
       ti10 = tr3 + ss*tr4
       tr11 = - ss*ti3 + ti4
       ti11 = ss*tr3 - tr4
       yr(ivk,1,1) = xr(ivk,1,1) + tr5
       yi(ivk,1,1) = xi(ivk,1,1) + ti5
       yr(ivk,2,1) = tr8 + tr10
       yi(ivk,2,1) = ti8 + ti10
       yr(ivk,3,1) = tr9 + tr11
       yi(ivk,3,1) = ti9 + ti11
       yr(ivk,4,1) = tr9 - tr11
       yi(ivk,4,1) = ti9 - ti11
       yr(ivk,5,1) = tr8 - tr10
       yi(ivk,5,1) = ti8 - ti10
    end do

    iw = 5
    do p = 2,np
       wr1 = w(1,iw)
       wi1 = w(2,iw)
       iw = iw+1
       wr2 = w(1,iw)
       wi2 = w(2,iw)
       iw = iw+1
       wr3 = w(1,iw)
       wi3 = w(2,iw)
       iw = iw+1
       wr4 = w(1,iw)
       wi4 = w(2,iw)
       iw = iw+1
       do ivk = 1,nvk
          tr1 = xr(ivk,p,2) + xr(ivk,p,5)
          ti1 = xi(ivk,p,2) + xi(ivk,p,5)
          tr2 = xr(ivk,p,3) + xr(ivk,p,4)
          ti2 = xi(ivk,p,3) + xi(ivk,p,4)
          tr3 = sp25*(xr(ivk,p,2) - xr(ivk,p,5))
          ti3 = sp25*(xi(ivk,p,2) - xi(ivk,p,5))
          tr4 = sp25*(xr(ivk,p,3) - xr(ivk,p,4))
          ti4 = sp25*(xi(ivk,p,3) - xi(ivk,p,4))
          tr5 = tr1 + tr2
          ti5 = ti1 + ti2
          tr6 = sq54*(tr1 - tr2)
          ti6 = sq54*(ti1 - ti2)
          tr7 = xr(ivk,p,1) - 0.25d0*tr5
          ti7 = xi(ivk,p,1) - 0.25d0*ti5
          tr8 = tr7 + tr6
          ti8 = ti7 + ti6
          tr9 = tr7 - tr6
          ti9 = ti7 - ti6
          tr10 = - ti3 - ss*ti4
          ti10 = tr3 + ss*tr4
          tr11 = - ss*ti3 + ti4
          ti11 = ss*tr3 - tr4
          yr(ivk,1,p) = xr(ivk,p,1) + tr5
          yi(ivk,1,p) = xi(ivk,p,1) + ti5
          tr = tr8 + tr10
          ti = ti8 + ti10
          yr(ivk,2,p) = wr1*tr + wi1*ti
          yi(ivk,2,p) = wr1*ti - wi1*tr
          tr = tr9 + tr11
          ti = ti9 + ti11
          yr(ivk,3,p) = wr2*tr + wi2*ti
          yi(ivk,3,p) = wr2*ti - wi2*tr
          tr = tr9 - tr11
          ti = ti9 - ti11
          yr(ivk,4,p) = wr3*tr + wi3*ti
          yi(ivk,4,p) = wr3*ti - wi3*tr
          tr = tr8 - tr10
          ti = ti8 - ti10
          yr(ivk,5,p) = wr4*tr + wi4*ti
          yi(ivk,5,p) = wr4*ti - wi4*tr
       end do
    end do
    return
  end subroutine bftrd5

  subroutine fftrd6(nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nvk,np
    real(kind=8),intent(in):: xr(nvk,np,6),xi(nvk,np,6)
    real(kind=8),intent(out):: yr(nvk,6,np),yi(nvk,6,np)
    real(kind=8),intent(in):: w(2,5*np)
    !
    real(kind=8),parameter:: sp3 = 0.866025403784438647d0
    !
    real(kind=8):: wr1,wi1,wr2,wi2,wr3,wi3,wr4,wi4,wr5,wi5
    real(kind=8):: tr1,ti1,tr2,ti2,tr3,ti3,tr,ti
    real(kind=8):: ur1,ui1,ur2,ui2,ur3,ui3
    real(kind=8):: vr1,vi1,vr2,vi2,vr3,vi3
    integer:: p,ivk,iw
    !
    do ivk = 1,nvk
       tr1 = xr(ivk,1,3) + xr(ivk,1,5)
       ti1 = xi(ivk,1,3) + xi(ivk,1,5)
       tr2 = xr(ivk,1,1) - 0.5d0*tr1
       ti2 = xi(ivk,1,1) - 0.5d0*ti1
       tr3 = -sp3*(xr(ivk,1,3) - xr(ivk,1,5))
       ti3 = -sp3*(xi(ivk,1,3) - xi(ivk,1,5))
       ur1 = xr(ivk,1,1) + tr1
       ui1 = xi(ivk,1,1) + ti1
       ur2 = tr2 - ti3
       ui2 = ti2 + tr3
       ur3 = tr2 + ti3
       ui3 = ti2 - tr3

       tr1 = xr(ivk,1,6) + xr(ivk,1,2)
       ti1 = xi(ivk,1,6) + xi(ivk,1,2)
       tr2 = xr(ivk,1,4) - 0.5d0*tr1
       ti2 = xi(ivk,1,4) - 0.5d0*ti1
       tr3 = -sp3*(xr(ivk,1,6) - xr(ivk,1,2))
       ti3 = -sp3*(xi(ivk,1,6) - xi(ivk,1,2))
       vr1 = xr(ivk,1,4) + tr1
       vi1 = xi(ivk,1,4) + ti1
       vr2 = tr2 - ti3
       vi2 = ti2 + tr3
       vr3 = tr2 + ti3
       vi3 = ti2 - tr3

       yr(ivk,1,1) = ur1 + vr1
       yi(ivk,1,1) = ui1 + vi1
       yr(ivk,5,1) = ur2 + vr2
       yi(ivk,5,1) = ui2 + vi2
       yr(ivk,3,1) = ur3 + vr3
       yi(ivk,3,1) = ui3 + vi3
       yr(ivk,4,1) = ur1 - vr1
       yi(ivk,4,1) = ui1 - vi1
       yr(ivk,2,1) = ur2 - vr2
       yi(ivk,2,1) = ui2 - vi2
       yr(ivk,6,1) = ur3 - vr3
       yi(ivk,6,1) = ui3 - vi3
    end do

    iw = 6
    do p = 2,np
       wr1 = w(1,iw)
       wi1 = w(2,iw)
       iw = iw+1
       wr2 = w(1,iw)
       wi2 = w(2,iw)
       iw = iw+1
       wr3 = w(1,iw)
       wi3 = w(2,iw)
       iw = iw+1
       wr4 = w(1,iw)
       wi4 = w(2,iw)
       iw = iw+1
       wr5 = w(1,iw)
       wi5 = w(2,iw)
       iw = iw+1
       do ivk = 1,nvk
          tr1 = xr(ivk,p,3) + xr(ivk,p,5)
          ti1 = xi(ivk,p,3) + xi(ivk,p,5)
          tr2 = xr(ivk,p,1) - 0.5d0*tr1
          ti2 = xi(ivk,p,1) - 0.5d0*ti1
          tr3 = -sp3*(xr(ivk,p,3) - xr(ivk,p,5))
          ti3 = -sp3*(xi(ivk,p,3) - xi(ivk,p,5))
          ur1 = xr(ivk,p,1) + tr1
          ui1 = xi(ivk,p,1) + ti1
          ur2 = tr2 - ti3
          ui2 = ti2 + tr3
          ur3 = tr2 + ti3
          ui3 = ti2 - tr3

          tr1 = xr(ivk,p,6) + xr(ivk,p,2)
          ti1 = xi(ivk,p,6) + xi(ivk,p,2)
          tr2 = xr(ivk,p,4) - 0.5d0*tr1
          ti2 = xi(ivk,p,4) - 0.5d0*ti1
          tr3 = -sp3*(xr(ivk,p,6) - xr(ivk,p,2))
          ti3 = -sp3*(xi(ivk,p,6) - xi(ivk,p,2))
          vr1 = xr(ivk,p,4) + tr1
          vi1 = xi(ivk,p,4) + ti1
          vr2 = tr2 - ti3
          vi2 = ti2 + tr3
          vr3 = tr2 + ti3
          vi3 = ti2 - tr3

          yr(ivk,1,p) = ur1 + vr1
          yi(ivk,1,p) = ui1 + vi1
          tr = ur2 + vr2
          ti = ui2 + vi2
          yr(ivk,5,p) = wr4*tr - wi4*ti
          yi(ivk,5,p) = wr4*ti + wi4*tr
          tr = ur3 + vr3
          ti = ui3 + vi3
          yr(ivk,3,p) = wr2*tr - wi2*ti
          yi(ivk,3,p) = wr2*ti + wi2*tr
          tr = ur1 - vr1
          ti = ui1 - vi1
          yr(ivk,4,p) = wr3*tr - wi3*ti
          yi(ivk,4,p) = wr3*ti + wi3*tr
          tr = ur2 - vr2
          ti = ui2 - vi2
          yr(ivk,2,p) = wr1*tr - wi1*ti
          yi(ivk,2,p) = wr1*ti + wi1*tr
          tr = ur3 - vr3
          ti = ui3 - vi3
          yr(ivk,6,p) = wr5*tr - wi5*ti
          yi(ivk,6,p) = wr5*ti + wi5*tr
       end do
    end do
    return
  end subroutine fftrd6

  subroutine bftrd6(nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nvk,np
    real(kind=8),intent(in):: xr(nvk,np,6),xi(nvk,np,6)
    real(kind=8),intent(out):: yr(nvk,6,np),yi(nvk,6,np)
    real(kind=8),intent(in):: w(2,5*np)
    !
    real(kind=8),parameter:: sp3 = 0.866025403784438647d0
    !
    real(kind=8):: wr1,wi1,wr2,wi2,wr3,wi3,wr4,wi4,wr5,wi5
    real(kind=8):: tr1,ti1,tr2,ti2,tr3,ti3,tr,ti
    real(kind=8):: ur1,ui1,ur2,ui2,ur3,ui3
    real(kind=8):: vr1,vi1,vr2,vi2,vr3,vi3
    integer:: p,ivk,iw
    !
    do ivk = 1,nvk
       tr1 = xr(ivk,1,3) + xr(ivk,1,5)
       ti1 = xi(ivk,1,3) + xi(ivk,1,5)
       tr2 = xr(ivk,1,1) - 0.5d0*tr1
       ti2 = xi(ivk,1,1) - 0.5d0*ti1
       tr3 = sp3*(xr(ivk,1,3) - xr(ivk,1,5))
       ti3 = sp3*(xi(ivk,1,3) - xi(ivk,1,5))
       ur1 = xr(ivk,1,1) + tr1
       ui1 = xi(ivk,1,1) + ti1
       ur2 = tr2 - ti3
       ui2 = ti2 + tr3
       ur3 = tr2 + ti3
       ui3 = ti2 - tr3

       tr1 = xr(ivk,1,6) + xr(ivk,1,2)
       ti1 = xi(ivk,1,6) + xi(ivk,1,2)
       tr2 = xr(ivk,1,4) - 0.5d0*tr1
       ti2 = xi(ivk,1,4) - 0.5d0*ti1
       tr3 = sp3*(xr(ivk,1,6) - xr(ivk,1,2))
       ti3 = sp3*(xi(ivk,1,6) - xi(ivk,1,2))
       vr1 = xr(ivk,1,4) + tr1
       vi1 = xi(ivk,1,4) + ti1
       vr2 = tr2 - ti3
       vi2 = ti2 + tr3
       vr3 = tr2 + ti3
       vi3 = ti2 - tr3

       yr(ivk,1,1) = ur1 + vr1
       yi(ivk,1,1) = ui1 + vi1
       yr(ivk,5,1) = ur2 + vr2
       yi(ivk,5,1) = ui2 + vi2
       yr(ivk,3,1) = ur3 + vr3
       yi(ivk,3,1) = ui3 + vi3
       yr(ivk,4,1) = ur1 - vr1
       yi(ivk,4,1) = ui1 - vi1
       yr(ivk,2,1) = ur2 - vr2
       yi(ivk,2,1) = ui2 - vi2
       yr(ivk,6,1) = ur3 - vr3
       yi(ivk,6,1) = ui3 - vi3
    end do

    iw = 6
    do p = 2,np
       wr1 = w(1,iw)
       wi1 = w(2,iw)
       iw = iw+1
       wr2 = w(1,iw)
       wi2 = w(2,iw)
       iw = iw+1
       wr3 = w(1,iw)
       wi3 = w(2,iw)
       iw = iw+1
       wr4 = w(1,iw)
       wi4 = w(2,iw)
       iw = iw+1
       wr5 = w(1,iw)
       wi5 = w(2,iw)
       iw = iw+1
       do ivk = 1,nvk
          tr1 = xr(ivk,p,3) + xr(ivk,p,5)
          ti1 = xi(ivk,p,3) + xi(ivk,p,5)
          tr2 = xr(ivk,p,1) - 0.5d0*tr1
          ti2 = xi(ivk,p,1) - 0.5d0*ti1
          tr3 = sp3*(xr(ivk,p,3) - xr(ivk,p,5))
          ti3 = sp3*(xi(ivk,p,3) - xi(ivk,p,5))
          ur1 = xr(ivk,p,1) + tr1
          ui1 = xi(ivk,p,1) + ti1
          ur2 = tr2 - ti3
          ui2 = ti2 + tr3
          ur3 = tr2 + ti3
          ui3 = ti2 - tr3

          tr1 = xr(ivk,p,6) + xr(ivk,p,2)
          ti1 = xi(ivk,p,6) + xi(ivk,p,2)
          tr2 = xr(ivk,p,4) - 0.5d0*tr1
          ti2 = xi(ivk,p,4) - 0.5d0*ti1
          tr3 = sp3*(xr(ivk,p,6) - xr(ivk,p,2))
          ti3 = sp3*(xi(ivk,p,6) - xi(ivk,p,2))
          vr1 = xr(ivk,p,4) + tr1
          vi1 = xi(ivk,p,4) + ti1
          vr2 = tr2 - ti3
          vi2 = ti2 + tr3
          vr3 = tr2 + ti3
          vi3 = ti2 - tr3

          yr(ivk,1,p) = ur1 + vr1
          yi(ivk,1,p) = ui1 + vi1
          tr = ur2 + vr2
          ti = ui2 + vi2
          yr(ivk,5,p) = wr4*tr + wi4*ti
          yi(ivk,5,p) = wr4*ti - wi4*tr
          tr = ur3 + vr3
          ti = ui3 + vi3
          yr(ivk,3,p) = wr2*tr + wi2*ti
          yi(ivk,3,p) = wr2*ti - wi2*tr
          tr = ur1 - vr1
          ti = ui1 - vi1
          yr(ivk,4,p) = wr3*tr + wi3*ti
          yi(ivk,4,p) = wr3*ti - wi3*tr
          tr = ur2 - vr2
          ti = ui2 - vi2
          yr(ivk,2,p) = wr1*tr + wi1*ti
          yi(ivk,2,p) = wr1*ti - wi1*tr
          tr = ur3 - vr3
          ti = ui3 - vi3
          yr(ivk,6,p) = wr5*tr + wi5*ti
          yi(ivk,6,p) = wr5*ti - wi5*tr
       end do
    end do
    return
  end subroutine bftrd6
  !
  subroutine fftrd8(nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nvk,np
    real(kind=8),intent(in):: xr(nvk,np,8),xi(nvk,np,8)
    real(kind=8),intent(out):: yr(nvk,8,np),yi(nvk,8,np)
    real(kind=8),intent(in):: w(2,7*np)
    !
    real(kind=8),parameter:: sp2 = 0.70710678118654752440d0
    !
    real(kind=8):: wr1,wi1,wr2,wi2,wr3,wi3,wr4,wi4,wr5,wi5,wr6,wi6,wr7,wi7
    real(kind=8):: tr1,ti1,tr2,ti2,tr3,ti3,tr4,ti4,tr,ti,sr,si
    real(kind=8):: ur1,ui1,ur2,ui2,ur3,ui3,ur4,ui4
    real(kind=8):: vr1,vi1,vr2,vi2,vr3,vi3,vr4,vi4
    integer:: p,ivk,iw
    !
    do ivk = 1,nvk
       tr1 = xr(ivk,1,1) + xr(ivk,1,5)
       ti1 = xi(ivk,1,1) + xi(ivk,1,5)
       tr2 = xr(ivk,1,1) - xr(ivk,1,5)
       ti2 = xi(ivk,1,1) - xi(ivk,1,5)
       tr3 = xr(ivk,1,3) + xr(ivk,1,7)
       ti3 = xi(ivk,1,3) + xi(ivk,1,7)
       tr4 = xi(ivk,1,3) - xi(ivk,1,7)
       ti4 = xr(ivk,1,7) - xr(ivk,1,3)
       ur1 = tr1 + tr3
       ui1 = ti1 + ti3
       ur2 = tr2 + tr4
       ui2 = ti2 + ti4
       ur3 = tr1 - tr3
       ui3 = ti1 - ti3
       ur4 = tr2 - tr4
       ui4 = ti2 - ti4
       tr1 = xr(ivk,1,2) + xr(ivk,1,6)
       ti1 = xi(ivk,1,2) + xi(ivk,1,6)
       tr2 = xr(ivk,1,2) - xr(ivk,1,6)
       ti2 = xi(ivk,1,2) - xi(ivk,1,6)
       tr3 = xr(ivk,1,4) + xr(ivk,1,8)
       ti3 = xi(ivk,1,4) + xi(ivk,1,8)
       tr4 = xi(ivk,1,4) - xi(ivk,1,8)
       ti4 = xr(ivk,1,8) - xr(ivk,1,4)
       vr1 = tr1 + tr3
       vi1 = ti1 + ti3
       vr2 = tr2 + tr4
       vi2 = ti2 + ti4
       vr3 = tr1 - tr3
       vi3 = ti1 - ti3
       vr4 = tr2 - tr4
       vi4 = ti2 - ti4
       yr(ivk,1,1) = ur1 + vr1
       yi(ivk,1,1) = ui1 + vi1
       yr(ivk,5,1) = ur1 - vr1
       yi(ivk,5,1) = ui1 - vi1

       sr = sp2*(vr2 + vi2)
       si = sp2*(vi2 - vr2)
       yr(ivk,2,1) = ur2 + sr
       yi(ivk,2,1) = ui2 + si
       yr(ivk,6,1) = ur2 - sr
       yi(ivk,6,1) = ui2 - si

       yr(ivk,3,1) = ur3 + vi3
       yi(ivk,3,1) = ui3 - vr3
       yr(ivk,7,1) = ur3 - vi3
       yi(ivk,7,1) = ui3 + vr3

       sr = sp2*(vi4 - vr4)
       si = -sp2*(vr4 + vi4)
       yr(ivk,4,1) = ur4 + sr
       yi(ivk,4,1) = ui4 + si
       yr(ivk,8,1) = ur4 - sr
       yi(ivk,8,1) = ui4 - si
    end do

    iw = 8
    do p = 2,np
       wr1 = w(1,iw)
       wi1 = w(2,iw)
       iw = iw+1
       wr2 = w(1,iw)
       wi2 = w(2,iw)
       iw = iw+1
       wr3 = w(1,iw)
       wi3 = w(2,iw)
       iw = iw+1
       wr4 = w(1,iw)
       wi4 = w(2,iw)
       iw = iw+1
       wr5 = w(1,iw)
       wi5 = w(2,iw)
       iw = iw+1
       wr6 = w(1,iw)
       wi6 = w(2,iw)
       iw = iw+1
       wr7 = w(1,iw)
       wi7 = w(2,iw)
       iw = iw+1
       do ivk = 1,nvk
          tr1 = xr(ivk,p,1) + xr(ivk,p,5)
          ti1 = xi(ivk,p,1) + xi(ivk,p,5)
          tr2 = xr(ivk,p,1) - xr(ivk,p,5)
          ti2 = xi(ivk,p,1) - xi(ivk,p,5)
          tr3 = xr(ivk,p,3) + xr(ivk,p,7)
          ti3 = xi(ivk,p,3) + xi(ivk,p,7)
          tr4 = xi(ivk,p,3) - xi(ivk,p,7)
          ti4 = xr(ivk,p,7) - xr(ivk,p,3)
          ur1 = tr1 + tr3
          ui1 = ti1 + ti3
          ur2 = tr2 + tr4
          ui2 = ti2 + ti4
          ur3 = tr1 - tr3
          ui3 = ti1 - ti3
          ur4 = tr2 - tr4
          ui4 = ti2 - ti4
          tr1 = xr(ivk,p,2) + xr(ivk,p,6)
          ti1 = xi(ivk,p,2) + xi(ivk,p,6)
          tr2 = xr(ivk,p,2) - xr(ivk,p,6)
          ti2 = xi(ivk,p,2) - xi(ivk,p,6)
          tr3 = xr(ivk,p,4) + xr(ivk,p,8)
          ti3 = xi(ivk,p,4) + xi(ivk,p,8)
          tr4 = xi(ivk,p,4) - xi(ivk,p,8)
          ti4 = xr(ivk,p,8) - xr(ivk,p,4)
          vr1 = tr1 + tr3
          vi1 = ti1 + ti3
          vr2 = tr2 + tr4
          vi2 = ti2 + ti4
          vr3 = tr1 - tr3
          vi3 = ti1 - ti3
          vr4 = tr2 - tr4
          vi4 = ti2 - ti4
          yr(ivk,1,p) = ur1 + vr1
          yi(ivk,1,p) = ui1 + vi1
          tr = ur1 - vr1
          ti = ui1 - vi1
          yr(ivk,5,p) = wr4*tr - wi4*ti
          yi(ivk,5,p) = wr4*ti + wi4*tr

          sr = sp2*(vr2 + vi2)
          si = sp2*(vi2 - vr2)
          tr = ur2 + sr
          ti = ui2 + si
          yr(ivk,2,p) = wr1*tr - wi1*ti
          yi(ivk,2,p) = wr1*ti + wi1*tr
          tr = ur2 - sr
          ti = ui2 - si
          yr(ivk,6,p) = wr5*tr - wi5*ti
          yi(ivk,6,p) = wr5*ti + wi5*tr

          tr = ur3 + vi3
          ti = ui3 - vr3
          yr(ivk,3,p) = wr2*tr - wi2*ti
          yi(ivk,3,p) = wr2*ti + wi2*tr
          tr = ur3 - vi3
          ti = ui3 + vr3
          yr(ivk,7,p) = wr6*tr - wi6*ti
          yi(ivk,7,p) = wr6*ti + wi6*tr

          sr = sp2*(vi4 - vr4)
          si = -sp2*(vr4 + vi4)
          tr = ur4 + sr
          ti = ui4 + si
          yr(ivk,4,p) = wr3*tr - wi3*ti
          yi(ivk,4,p) = wr3*ti + wi3*tr
          tr = ur4 - sr
          ti = ui4 - si
          yr(ivk,8,p) = wr7*tr - wi7*ti
          yi(ivk,8,p) = wr7*ti + wi7*tr
       end do
    end do
    return
  end subroutine fftrd8
  !
  subroutine bftrd8(nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nvk,np
    real(kind=8),intent(in):: xr(nvk,np,8),xi(nvk,np,8)
    real(kind=8),intent(out):: yr(nvk,8,np),yi(nvk,8,np)
    real(kind=8),intent(in):: w(2,7*np)
    !
    real(kind=8),parameter:: sp2 = 0.70710678118654752440d0
    !
    real(kind=8):: wr1,wi1,wr2,wi2,wr3,wi3,wr4,wi4,wr5,wi5,wr6,wi6,wr7,wi7
    real(kind=8):: tr1,ti1,tr2,ti2,tr3,ti3,tr4,ti4,tr,ti,sr,si
    real(kind=8):: ur1,ui1,ur2,ui2,ur3,ui3,ur4,ui4
    real(kind=8):: vr1,vi1,vr2,vi2,vr3,vi3,vr4,vi4
    integer:: p,ivk,iw
    !
    do ivk = 1,nvk
       tr1 = xr(ivk,1,1) + xr(ivk,1,5)
       ti1 = xi(ivk,1,1) + xi(ivk,1,5)
       tr2 = xr(ivk,1,1) - xr(ivk,1,5)
       ti2 = xi(ivk,1,1) - xi(ivk,1,5)
       tr3 = xr(ivk,1,3) + xr(ivk,1,7)
       ti3 = xi(ivk,1,3) + xi(ivk,1,7)
       tr4 = xi(ivk,1,7) - xi(ivk,1,3)
       ti4 = xr(ivk,1,3) - xr(ivk,1,7)
       ur1 = tr1 + tr3
       ui1 = ti1 + ti3
       ur2 = tr2 + tr4
       ui2 = ti2 + ti4
       ur3 = tr1 - tr3
       ui3 = ti1 - ti3
       ur4 = tr2 - tr4
       ui4 = ti2 - ti4
       tr1 = xr(ivk,1,2) + xr(ivk,1,6)
       ti1 = xi(ivk,1,2) + xi(ivk,1,6)
       tr2 = xr(ivk,1,2) - xr(ivk,1,6)
       ti2 = xi(ivk,1,2) - xi(ivk,1,6)
       tr3 = xr(ivk,1,4) + xr(ivk,1,8)
       ti3 = xi(ivk,1,4) + xi(ivk,1,8)
       tr4 = xi(ivk,1,8) - xi(ivk,1,4)
       ti4 = xr(ivk,1,4) - xr(ivk,1,8)
       vr1 = tr1 + tr3
       vi1 = ti1 + ti3
       vr2 = tr2 + tr4
       vi2 = ti2 + ti4
       vr3 = tr1 - tr3
       vi3 = ti1 - ti3
       vr4 = tr2 - tr4
       vi4 = ti2 - ti4
       yr(ivk,1,1) = ur1 + vr1
       yi(ivk,1,1) = ui1 + vi1
       yr(ivk,5,1) = ur1 - vr1
       yi(ivk,5,1) = ui1 - vi1

       sr = sp2*(vr2 - vi2)
       si = sp2*(vi2 + vr2)
       yr(ivk,2,1) = ur2 + sr
       yi(ivk,2,1) = ui2 + si
       yr(ivk,6,1) = ur2 - sr
       yi(ivk,6,1) = ui2 - si

       yr(ivk,3,1) = ur3 - vi3
       yi(ivk,3,1) = ui3 + vr3
       yr(ivk,7,1) = ur3 + vi3
       yi(ivk,7,1) = ui3 - vr3

       sr = -sp2*(vr4 + vi4)
       si = sp2*(vr4 - vi4)
       yr(ivk,4,1) = ur4 + sr
       yi(ivk,4,1) = ui4 + si
       yr(ivk,8,1) = ur4 - sr
       yi(ivk,8,1) = ui4 - si
    end do

    iw = 8
    do p = 2,np
       wr1 = w(1,iw)
       wi1 = w(2,iw)
       iw = iw+1
       wr2 = w(1,iw)
       wi2 = w(2,iw)
       iw = iw+1
       wr3 = w(1,iw)
       wi3 = w(2,iw)
       iw = iw+1
       wr4 = w(1,iw)
       wi4 = w(2,iw)
       iw = iw+1
       wr5 = w(1,iw)
       wi5 = w(2,iw)
       iw = iw+1
       wr6 = w(1,iw)
       wi6 = w(2,iw)
       iw = iw+1
       wr7 = w(1,iw)
       wi7 = w(2,iw)
       iw = iw+1
       do ivk = 1,nvk
          tr1 = xr(ivk,p,1) + xr(ivk,p,5)
          ti1 = xi(ivk,p,1) + xi(ivk,p,5)
          tr2 = xr(ivk,p,1) - xr(ivk,p,5)
          ti2 = xi(ivk,p,1) - xi(ivk,p,5)
          tr3 = xr(ivk,p,3) + xr(ivk,p,7)
          ti3 = xi(ivk,p,3) + xi(ivk,p,7)
          tr4 = xi(ivk,p,7) - xi(ivk,p,3)
          ti4 = xr(ivk,p,3) - xr(ivk,p,7)
          ur1 = tr1 + tr3
          ui1 = ti1 + ti3
          ur2 = tr2 + tr4
          ui2 = ti2 + ti4
          ur3 = tr1 - tr3
          ui3 = ti1 - ti3
          ur4 = tr2 - tr4
          ui4 = ti2 - ti4
          tr1 = xr(ivk,p,2) + xr(ivk,p,6)
          ti1 = xi(ivk,p,2) + xi(ivk,p,6)
          tr2 = xr(ivk,p,2) - xr(ivk,p,6)
          ti2 = xi(ivk,p,2) - xi(ivk,p,6)
          tr3 = xr(ivk,p,4) + xr(ivk,p,8)
          ti3 = xi(ivk,p,4) + xi(ivk,p,8)
          tr4 = xi(ivk,p,8) - xi(ivk,p,4)
          ti4 = xr(ivk,p,4) - xr(ivk,p,8)
          vr1 = tr1 + tr3
          vi1 = ti1 + ti3
          vr2 = tr2 + tr4
          vi2 = ti2 + ti4
          vr3 = tr1 - tr3
          vi3 = ti1 - ti3
          vr4 = tr2 - tr4
          vi4 = ti2 - ti4
          yr(ivk,1,p) = ur1 + vr1
          yi(ivk,1,p) = ui1 + vi1
          tr = ur1 - vr1
          ti = ui1 - vi1
          yr(ivk,5,p) = wr4*tr + wi4*ti
          yi(ivk,5,p) = wr4*ti - wi4*tr

          sr = sp2*(vr2 - vi2)
          si = sp2*(vi2 + vr2)
          tr = ur2 + sr
          ti = ui2 + si
          yr(ivk,2,p) = wr1*tr + wi1*ti
          yi(ivk,2,p) = wr1*ti - wi1*tr
          tr = ur2 - sr
          ti = ui2 - si
          yr(ivk,6,p) = wr5*tr + wi5*ti
          yi(ivk,6,p) = wr5*ti - wi5*tr

          tr = ur3 - vi3
          ti = ui3 + vr3
          yr(ivk,3,p) = wr2*tr + wi2*ti
          yi(ivk,3,p) = wr2*ti - wi2*tr
          tr = ur3 + vi3
          ti = ui3 - vr3
          yr(ivk,7,p) = wr6*tr + wi6*ti
          yi(ivk,7,p) = wr6*ti - wi6*tr

          sr = -sp2*(vr4 + vi4)
          si = sp2*(vr4 - vi4)
          tr = ur4 + sr
          ti = ui4 + si
          yr(ivk,4,p) = wr3*tr + wi3*ti
          yi(ivk,4,p) = wr3*ti - wi3*tr
          tr = ur4 - sr
          ti = ui4 - si
          yr(ivk,8,p) = wr7*tr + wi7*ti
          yi(ivk,8,p) = wr7*ti - wi7*tr
       end do
    end do
    return
  end subroutine bftrd8
  !
  subroutine fftrd2b(nt,nu,iu,nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nt,nu,iu,nvk,np
    real(kind=8),intent(in):: xr(nt,nu,nvk,np,2),xi(nt,nu,nvk,np,2)
    real(kind=8),intent(out):: yr(nt,nu,nvk,2,np),yi(nt,nu,nvk,2,np)
    real(kind=8),intent(in):: w(2,np)
    !
    real(kind=8):: wr1,wi1,tr,ti
    integer:: p,ivk,it
    !
    do p = 1,np
       wr1 = w(1,p)
       wi1 = w(2,p)
       do ivk = 1,nvk
          do it = 1, nt
             yr(it,iu,ivk,1,p) = xr(it,iu,ivk,p,1) + xr(it,iu,ivk,p,2)
             yi(it,iu,ivk,1,p) = xi(it,iu,ivk,p,1) + xi(it,iu,ivk,p,2)
             !
             tr = xr(it,iu,ivk,p,1) - xr(it,iu,ivk,p,2)
             ti = xi(it,iu,ivk,p,1) - xi(it,iu,ivk,p,2)
             yr(it,iu,ivk,2,p) = wr1*tr - wi1*ti
             yi(it,iu,ivk,2,p) = wr1*ti + wi1*tr
          end do
       end do
    end do
    return
  end subroutine fftrd2b
  !
  subroutine bftrd2b(nt,nu,iu,nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nt,nu,iu,nvk,np
    real(kind=8),intent(in):: xr(nt,nu,nvk,np,2),xi(nt,nu,nvk,np,2)
    real(kind=8),intent(out):: yr(nt,nu,nvk,2,np),yi(nt,nu,nvk,2,np)
    real(kind=8),intent(in):: w(2,np)
    !
    real(kind=8):: wr1,wi1,tr,ti
    integer:: p,ivk,it
    !
    do p = 1,np
       wr1 = w(1,p)
       wi1 = w(2,p)
       do ivk = 1,nvk
          do it = 1, nt
             yr(it,iu,ivk,1,p) = xr(it,iu,ivk,p,1) + xr(it,iu,ivk,p,2)
             yi(it,iu,ivk,1,p) = xi(it,iu,ivk,p,1) + xi(it,iu,ivk,p,2)
             !
             tr = xr(it,iu,ivk,p,1) - xr(it,iu,ivk,p,2)
             ti = xi(it,iu,ivk,p,1) - xi(it,iu,ivk,p,2)
             yr(it,iu,ivk,2,p) = wr1*tr + wi1*ti
             yi(it,iu,ivk,2,p) = wr1*ti - wi1*tr
          end do
       end do
    end do
    return
  end subroutine bftrd2b
  !
  subroutine fftrd3b(nt,nu,iu,nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nt,nu,iu,nvk,np
    real(kind=8),intent(in):: xr(nt,nu,nvk,np,3),xi(nt,nu,nvk,np,3)
    real(kind=8),intent(out):: yr(nt,nu,nvk,3,np),yi(nt,nu,nvk,3,np)
    real(kind=8),intent(in):: w(2,2*np)
    !
    real(kind=8),parameter:: sp3 = 0.866025403784438647d0
    !
    real(kind=8):: wr1,wi1,wr2,wi2,tr1,ti1,tr2,ti2,tr3,ti3,tr,ti
    integer:: p,ivk,iw,it
    !
    iw = 1
    do p = 1,np
       wr1 = w(1,iw)
       wi1 = w(2,iw)
       iw = iw+1
       wr2 = w(1,iw)
       wi2 = w(2,iw)
       iw = iw+1
       do ivk = 1,nvk
          do it=1,nt
             tr1 = xr(it,iu,ivk,p,2) + xr(it,iu,ivk,p,3)
             ti1 = xi(it,iu,ivk,p,2) + xi(it,iu,ivk,p,3)
             tr2 = xr(it,iu,ivk,p,1) - 0.5d0*tr1
             ti2 = xi(it,iu,ivk,p,1) - 0.5d0*ti1
             tr3 = sp3*(xi(it,iu,ivk,p,2) - xi(it,iu,ivk,p,3))
             ti3 = -sp3*(xr(it,iu,ivk,p,2) - xr(it,iu,ivk,p,3))
             yr(it,iu,ivk,1,p) = xr(it,iu,ivk,p,1) + tr1
             yi(it,iu,ivk,1,p) = xi(it,iu,ivk,p,1) + ti1
             tr = tr2 + tr3
             ti = ti2 + ti3
             yr(it,iu,ivk,2,p) = wr1*tr - wi1*ti
             yi(it,iu,ivk,2,p) = wr1*ti + wi1*tr
             tr = tr2 - tr3
             ti = ti2 - ti3
             yr(it,iu,ivk,3,p) = wr2*tr - wi2*ti
             yi(it,iu,ivk,3,p) = wr2*ti + wi2*tr
          end do
       end do
    end do
    return
  end subroutine fftrd3b
  !
  subroutine bftrd3b(nt,nu,iu,nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nt,nu,iu,nvk,np
    real(kind=8),intent(in):: xr(nt,nu,nvk,np,3),xi(nt,nu,nvk,np,3)
    real(kind=8),intent(out):: yr(nt,nu,nvk,3,np),yi(nt,nu,nvk,3,np)
    real(kind=8),intent(in):: w(2,2*np)
    !
    real(kind=8),parameter:: sp3 = 0.866025403784438647d0
    !
    real(kind=8):: wr1,wi1,wr2,wi2,tr1,ti1,tr2,ti2,tr3,ti3,tr,ti
    integer:: p,ivk,iw,it
    !
    iw = 1
    do p = 1,np
       wr1 = w(1,iw)
       wi1 = w(2,iw)
       iw = iw+1
       wr2 = w(1,iw)
       wi2 = w(2,iw)
       iw = iw+1
       do ivk = 1,nvk
          do it=1,nt
             tr1 = xr(it,iu,ivk,p,2) + xr(it,iu,ivk,p,3)
             ti1 = xi(it,iu,ivk,p,2) + xi(it,iu,ivk,p,3)
             tr2 = xr(it,iu,ivk,p,1) - 0.5d0*tr1
             ti2 = xi(it,iu,ivk,p,1) - 0.5d0*ti1
             tr3 = -sp3*(xi(it,iu,ivk,p,2) - xi(it,iu,ivk,p,3))
             ti3 = sp3*(xr(it,iu,ivk,p,2) - xr(it,iu,ivk,p,3))
             yr(it,iu,ivk,1,p) = xr(it,iu,ivk,p,1) + tr1
             yi(it,iu,ivk,1,p) = xi(it,iu,ivk,p,1) + ti1
             tr = tr2 + tr3
             ti = ti2 + ti3
             yr(it,iu,ivk,2,p) = wr1*tr + wi1*ti
             yi(it,iu,ivk,2,p) = wr1*ti - wi1*tr
             tr = tr2 - tr3
             ti = ti2 - ti3
             yr(it,iu,ivk,3,p) = wr2*tr + wi2*ti
             yi(it,iu,ivk,3,p) = wr2*ti - wi2*tr
          end do
       end do
    end do
    return
  end subroutine bftrd3b
  !
  subroutine fftrd4b(nt,nu,iu,nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nt,nu,iu,nvk,np
    real(kind=8),intent(in):: xr(nt,nu,nvk,np,4),xi(nt,nu,nvk,np,4)
    real(kind=8),intent(out):: yr(nt,nu,nvk,4,np),yi(nt,nu,nvk,4,np)
    real(kind=8),intent(in):: w(2,3*np)
    !
    real(kind=8):: wr1,wi1,wr2,wi2,wr3,wi3
    real(kind=8):: tr1,ti1,tr2,ti2,tr3,ti3,tr4,ti4,tr,ti
    integer:: p,ivk,iw,it
    !
    iw = 1
    do p = 1,np
       wr1 = w(1,iw)
       wi1 = w(2,iw)
       iw = iw+1
       wr2 = w(1,iw)
       wi2 = w(2,iw)
       iw = iw+1
       wr3 = w(1,iw)
       wi3 = w(2,iw)
       iw = iw+1
       do ivk = 1,nvk
          do it=1,nt
             tr1 = xr(it,iu,ivk,p,1) + xr(it,iu,ivk,p,3)
             ti1 = xi(it,iu,ivk,p,1) + xi(it,iu,ivk,p,3)
             tr2 = xr(it,iu,ivk,p,1) - xr(it,iu,ivk,p,3)
             ti2 = xi(it,iu,ivk,p,1) - xi(it,iu,ivk,p,3)
             tr3 = xr(it,iu,ivk,p,2) + xr(it,iu,ivk,p,4)
             ti3 = xi(it,iu,ivk,p,2) + xi(it,iu,ivk,p,4)
             tr4 = xi(it,iu,ivk,p,2) - xi(it,iu,ivk,p,4)
             ti4 = xr(it,iu,ivk,p,4) - xr(it,iu,ivk,p,2)
             yr(it,iu,ivk,1,p) = tr1 + tr3
             yi(it,iu,ivk,1,p) = ti1 + ti3
             tr = tr2 + tr4
             ti = ti2 + ti4
             yr(it,iu,ivk,2,p) = wr1*tr - wi1*ti
             yi(it,iu,ivk,2,p) = wr1*ti + wi1*tr
             tr = tr1 - tr3
             ti = ti1 - ti3
             yr(it,iu,ivk,3,p) = wr2*tr - wi2*ti
             yi(it,iu,ivk,3,p) = wr2*ti + wi2*tr
             tr = tr2 - tr4
             ti = ti2 - ti4
             yr(it,iu,ivk,4,p) = wr3*tr - wi3*ti
             yi(it,iu,ivk,4,p) = wr3*ti + wi3*tr
          end do
       end do
    end do
    return
  end subroutine fftrd4b
  !
  subroutine bftrd4b(nt,nu,iu,nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nt,nu,iu,nvk,np
    real(kind=8),intent(in):: xr(nt,nu,nvk,np,4),xi(nt,nu,nvk,np,4)
    real(kind=8),intent(out):: yr(nt,nu,nvk,4,np),yi(nt,nu,nvk,4,np)
    real(kind=8),intent(in):: w(2,3*np)
    !
    real(kind=8):: wr1,wi1,wr2,wi2,wr3,wi3
    real(kind=8):: tr1,ti1,tr2,ti2,tr3,ti3,tr4,ti4,tr,ti
    integer:: p,ivk,iw,it
    !
    iw = 1
    do p = 1,np
       wr1 = w(1,iw)
       wi1 = w(2,iw)
       iw = iw+1
       wr2 = w(1,iw)
       wi2 = w(2,iw)
       iw = iw+1
       wr3 = w(1,iw)
       wi3 = w(2,iw)
       iw = iw+1
       do ivk = 1,nvk
          do it=1,nt
             tr1 = xr(it,iu,ivk,p,1) + xr(it,iu,ivk,p,3)
             ti1 = xi(it,iu,ivk,p,1) + xi(it,iu,ivk,p,3)
             tr2 = xr(it,iu,ivk,p,1) - xr(it,iu,ivk,p,3)
             ti2 = xi(it,iu,ivk,p,1) - xi(it,iu,ivk,p,3)
             tr3 = xr(it,iu,ivk,p,2) + xr(it,iu,ivk,p,4)
             ti3 = xi(it,iu,ivk,p,2) + xi(it,iu,ivk,p,4)
             tr4 = xi(it,iu,ivk,p,4) - xi(it,iu,ivk,p,2)
             ti4 = xr(it,iu,ivk,p,2) - xr(it,iu,ivk,p,4)
             yr(it,iu,ivk,1,p) = tr1 + tr3
             yi(it,iu,ivk,1,p) = ti1 + ti3
             tr = tr2 + tr4
             ti = ti2 + ti4
             yr(it,iu,ivk,2,p) = wr1*tr + wi1*ti
             yi(it,iu,ivk,2,p) = wr1*ti - wi1*tr
             tr = tr1 - tr3
             ti = ti1 - ti3
             yr(it,iu,ivk,3,p) = wr2*tr + wi2*ti
             yi(it,iu,ivk,3,p) = wr2*ti - wi2*tr
             tr = tr2 - tr4
             ti = ti2 - ti4
             yr(it,iu,ivk,4,p) = wr3*tr + wi3*ti
             yi(it,iu,ivk,4,p) = wr3*ti - wi3*tr
          end do
       end do
    end do
    return
  end subroutine bftrd4b
  !
  subroutine fftrd5b(nt,nu,iu,nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nt,nu,iu,nvk,np
    real(kind=8),intent(in):: xr(nt,nu,nvk,np,5),xi(nt,nu,nvk,np,5)
    real(kind=8),intent(out):: yr(nt,nu,nvk,5,np),yi(nt,nu,nvk,5,np)
    real(kind=8),intent(in):: w(2,4*np)
    !
    real(kind=8),parameter:: sp25 = 0.951056516295153572d0
    real(kind=8),parameter:: sq54 = 0.559016994374947424d0
    real(kind=8),parameter:: ss = 0.618033988749894848d0
    !
    real(kind=8):: wr1,wi1,wr2,wi2,wr3,wi3,wr4,wi4
    real(kind=8):: tr1,ti1,tr2,ti2,tr3,ti3,tr4,ti4,tr5,ti5
    real(kind=8):: tr6,ti6,tr7,ti7,tr8,ti8,tr9,ti9,tr10,ti10,tr11,ti11
    real(kind=8):: tr,ti
    integer:: p,ivk,iw,it
    !
    do ivk = 1,nvk
       do it = 1,nt
          tr1 = xr(it,iu,ivk,1,2) + xr(it,iu,ivk,1,5)
          ti1 = xi(it,iu,ivk,1,2) + xi(it,iu,ivk,1,5)
          tr2 = xr(it,iu,ivk,1,3) + xr(it,iu,ivk,1,4)
          ti2 = xi(it,iu,ivk,1,3) + xi(it,iu,ivk,1,4)
          tr3 = sp25*(xr(it,iu,ivk,1,2) - xr(it,iu,ivk,1,5))
          ti3 = sp25*(xi(it,iu,ivk,1,2) - xi(it,iu,ivk,1,5))
          tr4 = sp25*(xr(it,iu,ivk,1,3) - xr(it,iu,ivk,1,4))
          ti4 = sp25*(xi(it,iu,ivk,1,3) - xi(it,iu,ivk,1,4))
          tr5 = tr1 + tr2
          ti5 = ti1 + ti2
          tr6 = sq54*(tr1 - tr2)
          ti6 = sq54*(ti1 - ti2)
          tr7 = xr(it,iu,ivk,1,1) - 0.25d0*tr5
          ti7 = xi(it,iu,ivk,1,1) - 0.25d0*ti5
          tr8 = tr7 + tr6
          ti8 = ti7 + ti6
          tr9 = tr7 - tr6
          ti9 = ti7 - ti6
          tr10 = ti3 + ss*ti4
          ti10 = - tr3 - ss*tr4
          tr11 = ss*ti3 - ti4
          ti11 = - ss*tr3 + tr4
          yr(it,iu,ivk,1,1) = xr(it,iu,ivk,1,1) + tr5
          yi(it,iu,ivk,1,1) = xi(it,iu,ivk,1,1) + ti5
          yr(it,iu,ivk,2,1) = tr8 + tr10
          yi(it,iu,ivk,2,1) = ti8 + ti10
          yr(it,iu,ivk,3,1) = tr9 + tr11
          yi(it,iu,ivk,3,1) = ti9 + ti11
          yr(it,iu,ivk,4,1) = tr9 - tr11
          yi(it,iu,ivk,4,1) = ti9 - ti11
          yr(it,iu,ivk,5,1) = tr8 - tr10
          yi(it,iu,ivk,5,1) = ti8 - ti10
       end do
    end do

    iw = 5
    do p = 2,np
       wr1 = w(1,iw)
       wi1 = w(2,iw)
       iw = iw+1
       wr2 = w(1,iw)
       wi2 = w(2,iw)
       iw = iw+1
       wr3 = w(1,iw)
       wi3 = w(2,iw)
       iw = iw+1
       wr4 = w(1,iw)
       wi4 = w(2,iw)
       iw = iw+1
       do ivk = 1,nvk
          do it=1,nt
             tr1 = xr(it,iu,ivk,p,2) + xr(it,iu,ivk,p,5)
             ti1 = xi(it,iu,ivk,p,2) + xi(it,iu,ivk,p,5)
             tr2 = xr(it,iu,ivk,p,3) + xr(it,iu,ivk,p,4)
             ti2 = xi(it,iu,ivk,p,3) + xi(it,iu,ivk,p,4)
             tr3 = sp25*(xr(it,iu,ivk,p,2) - xr(it,iu,ivk,p,5))
             ti3 = sp25*(xi(it,iu,ivk,p,2) - xi(it,iu,ivk,p,5))
             tr4 = sp25*(xr(it,iu,ivk,p,3) - xr(it,iu,ivk,p,4))
             ti4 = sp25*(xi(it,iu,ivk,p,3) - xi(it,iu,ivk,p,4))
             tr5 = tr1 + tr2
             ti5 = ti1 + ti2
             tr6 = sq54*(tr1 - tr2)
             ti6 = sq54*(ti1 - ti2)
             tr7 = xr(it,iu,ivk,p,1) - 0.25d0*tr5
             ti7 = xi(it,iu,ivk,p,1) - 0.25d0*ti5
             tr8 = tr7 + tr6
             ti8 = ti7 + ti6
             tr9 = tr7 - tr6
             ti9 = ti7 - ti6
             tr10 = ti3 + ss*ti4
             ti10 = - tr3 - ss*tr4
             tr11 = ss*ti3 - ti4
             ti11 = - ss*tr3 + tr4
             yr(it,iu,ivk,1,p) = xr(it,iu,ivk,p,1) + tr5
             yi(it,iu,ivk,1,p) = xi(it,iu,ivk,p,1) + ti5
             tr = tr8 + tr10
             ti = ti8 + ti10
             yr(it,iu,ivk,2,p) = wr1*tr - wi1*ti
             yi(it,iu,ivk,2,p) = wr1*ti + wi1*tr
             tr = tr9 + tr11
             ti = ti9 + ti11
             yr(it,iu,ivk,3,p) = wr2*tr - wi2*ti
             yi(it,iu,ivk,3,p) = wr2*ti + wi2*tr
             tr = tr9 - tr11
             ti = ti9 - ti11
             yr(it,iu,ivk,4,p) = wr3*tr - wi3*ti
             yi(it,iu,ivk,4,p) = wr3*ti + wi3*tr
             tr = tr8 - tr10
             ti = ti8 - ti10
             yr(it,iu,ivk,5,p) = wr4*tr - wi4*ti
             yi(it,iu,ivk,5,p) = wr4*ti + wi4*tr
          end do
       end do
    end do
    return
  end subroutine fftrd5b
  !
  subroutine bftrd5b(nt,nu,iu,nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nt,nu,iu,nvk,np
    real(kind=8),intent(in):: xr(nt,nu,nvk,np,5),xi(nt,nu,nvk,np,5)
    real(kind=8),intent(out):: yr(nt,nu,nvk,5,np),yi(nt,nu,nvk,5,np)
    real(kind=8),intent(in):: w(2,4*np)
    !
    real(kind=8),parameter:: sp25 = 0.951056516295153572d0
    real(kind=8),parameter:: sq54 = 0.559016994374947424d0
    real(kind=8),parameter:: ss = 0.618033988749894848d0
    !
    real(kind=8):: wr1,wi1,wr2,wi2,wr3,wi3,wr4,wi4
    real(kind=8):: tr1,ti1,tr2,ti2,tr3,ti3,tr4,ti4,tr5,ti5
    real(kind=8):: tr6,ti6,tr7,ti7,tr8,ti8,tr9,ti9,tr10,ti10,tr11,ti11
    real(kind=8):: tr,ti
    integer:: p,ivk,iw,it
    !
    do ivk = 1,nvk
       do it = 1,nt
       tr1 = xr(it,iu,ivk,1,2) + xr(it,iu,ivk,1,5)
       ti1 = xi(it,iu,ivk,1,2) + xi(it,iu,ivk,1,5)
       tr2 = xr(it,iu,ivk,1,3) + xr(it,iu,ivk,1,4)
       ti2 = xi(it,iu,ivk,1,3) + xi(it,iu,ivk,1,4)
       tr3 = sp25*(xr(it,iu,ivk,1,2) - xr(it,iu,ivk,1,5))
       ti3 = sp25*(xi(it,iu,ivk,1,2) - xi(it,iu,ivk,1,5))
       tr4 = sp25*(xr(it,iu,ivk,1,3) - xr(it,iu,ivk,1,4))
       ti4 = sp25*(xi(it,iu,ivk,1,3) - xi(it,iu,ivk,1,4))
       tr5 = tr1 + tr2
       ti5 = ti1 + ti2
       tr6 = sq54*(tr1 - tr2)
       ti6 = sq54*(ti1 - ti2)
       tr7 = xr(it,iu,ivk,1,1) - 0.25d0*tr5
       ti7 = xi(it,iu,ivk,1,1) - 0.25d0*ti5
       tr8 = tr7 + tr6
       ti8 = ti7 + ti6
       tr9 = tr7 - tr6
       ti9 = ti7 - ti6
       tr10 = - ti3 - ss*ti4
       ti10 = tr3 + ss*tr4
       tr11 = - ss*ti3 + ti4
       ti11 = ss*tr3 - tr4
       yr(it,iu,ivk,1,1) = xr(it,iu,ivk,1,1) + tr5
       yi(it,iu,ivk,1,1) = xi(it,iu,ivk,1,1) + ti5
       yr(it,iu,ivk,2,1) = tr8 + tr10
       yi(it,iu,ivk,2,1) = ti8 + ti10
       yr(it,iu,ivk,3,1) = tr9 + tr11
       yi(it,iu,ivk,3,1) = ti9 + ti11
       yr(it,iu,ivk,4,1) = tr9 - tr11
       yi(it,iu,ivk,4,1) = ti9 - ti11
       yr(it,iu,ivk,5,1) = tr8 - tr10
       yi(it,iu,ivk,5,1) = ti8 - ti10
       end do
    end do

    iw = 5
    do p = 2,np
       wr1 = w(1,iw)
       wi1 = w(2,iw)
       iw = iw+1
       wr2 = w(1,iw)
       wi2 = w(2,iw)
       iw = iw+1
       wr3 = w(1,iw)
       wi3 = w(2,iw)
       iw = iw+1
       wr4 = w(1,iw)
       wi4 = w(2,iw)
       iw = iw+1
       do ivk = 1,nvk
          do it=1,nt
             tr1 = xr(it,iu,ivk,p,2) + xr(it,iu,ivk,p,5)
             ti1 = xi(it,iu,ivk,p,2) + xi(it,iu,ivk,p,5)
             tr2 = xr(it,iu,ivk,p,3) + xr(it,iu,ivk,p,4)
             ti2 = xi(it,iu,ivk,p,3) + xi(it,iu,ivk,p,4)
             tr3 = sp25*(xr(it,iu,ivk,p,2) - xr(it,iu,ivk,p,5))
             ti3 = sp25*(xi(it,iu,ivk,p,2) - xi(it,iu,ivk,p,5))
             tr4 = sp25*(xr(it,iu,ivk,p,3) - xr(it,iu,ivk,p,4))
             ti4 = sp25*(xi(it,iu,ivk,p,3) - xi(it,iu,ivk,p,4))
             tr5 = tr1 + tr2
             ti5 = ti1 + ti2
             tr6 = sq54*(tr1 - tr2)
             ti6 = sq54*(ti1 - ti2)
             tr7 = xr(it,iu,ivk,p,1) - 0.25d0*tr5
             ti7 = xi(it,iu,ivk,p,1) - 0.25d0*ti5
             tr8 = tr7 + tr6
             ti8 = ti7 + ti6
             tr9 = tr7 - tr6
             ti9 = ti7 - ti6
             tr10 = - ti3 - ss*ti4
             ti10 = tr3 + ss*tr4
             tr11 = - ss*ti3 + ti4
             ti11 = ss*tr3 - tr4
             yr(it,iu,ivk,1,p) = xr(it,iu,ivk,p,1) + tr5
             yi(it,iu,ivk,1,p) = xi(it,iu,ivk,p,1) + ti5
             tr = tr8 + tr10
             ti = ti8 + ti10
             yr(it,iu,ivk,2,p) = wr1*tr + wi1*ti
             yi(it,iu,ivk,2,p) = wr1*ti - wi1*tr
             tr = tr9 + tr11
             ti = ti9 + ti11
             yr(it,iu,ivk,3,p) = wr2*tr + wi2*ti
             yi(it,iu,ivk,3,p) = wr2*ti - wi2*tr
             tr = tr9 - tr11
             ti = ti9 - ti11
             yr(it,iu,ivk,4,p) = wr3*tr + wi3*ti
             yi(it,iu,ivk,4,p) = wr3*ti - wi3*tr
             tr = tr8 - tr10
             ti = ti8 - ti10
             yr(it,iu,ivk,5,p) = wr4*tr + wi4*ti
             yi(it,iu,ivk,5,p) = wr4*ti - wi4*tr
          end do
       end do
    end do
    return
  end subroutine bftrd5b

  subroutine fftrd6b(nt,nu,iu,nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nt,nu,iu,nvk,np
    real(kind=8),intent(in):: xr(nt,nu,nvk,np,6),xi(nt,nu,nvk,np,6)
    real(kind=8),intent(out):: yr(nt,nu,nvk,6,np),yi(nt,nu,nvk,6,np)
    real(kind=8),intent(in):: w(2,5*np)
    !
    real(kind=8),parameter:: sp3 = 0.866025403784438647d0
    !
    real(kind=8):: wr1,wi1,wr2,wi2,wr3,wi3,wr4,wi4,wr5,wi5
    real(kind=8):: tr1,ti1,tr2,ti2,tr3,ti3,tr,ti
    real(kind=8):: ur1,ui1,ur2,ui2,ur3,ui3
    real(kind=8):: vr1,vi1,vr2,vi2,vr3,vi3
    integer:: p,ivk,iw,it
    !
    do ivk = 1,nvk
       do it = 1, nt
          tr1 = xr(it,iu,ivk,1,3) + xr(it,iu,ivk,1,5)
          ti1 = xi(it,iu,ivk,1,3) + xi(it,iu,ivk,1,5)
          tr2 = xr(it,iu,ivk,1,1) - 0.5d0*tr1
          ti2 = xi(it,iu,ivk,1,1) - 0.5d0*ti1
          tr3 = -sp3*(xr(it,iu,ivk,1,3) - xr(it,iu,ivk,1,5))
          ti3 = -sp3*(xi(it,iu,ivk,1,3) - xi(it,iu,ivk,1,5))
          ur1 = xr(it,iu,ivk,1,1) + tr1
          ui1 = xi(it,iu,ivk,1,1) + ti1
          ur2 = tr2 - ti3
          ui2 = ti2 + tr3
          ur3 = tr2 + ti3
          ui3 = ti2 - tr3

          tr1 = xr(it,iu,ivk,1,6) + xr(it,iu,ivk,1,2)
          ti1 = xi(it,iu,ivk,1,6) + xi(it,iu,ivk,1,2)
          tr2 = xr(it,iu,ivk,1,4) - 0.5d0*tr1
          ti2 = xi(it,iu,ivk,1,4) - 0.5d0*ti1
          tr3 = -sp3*(xr(it,iu,ivk,1,6) - xr(it,iu,ivk,1,2))
          ti3 = -sp3*(xi(it,iu,ivk,1,6) - xi(it,iu,ivk,1,2))
          vr1 = xr(it,iu,ivk,1,4) + tr1
          vi1 = xi(it,iu,ivk,1,4) + ti1
          vr2 = tr2 - ti3
          vi2 = ti2 + tr3
          vr3 = tr2 + ti3
          vi3 = ti2 - tr3

          yr(it,iu,ivk,1,1) = ur1 + vr1
          yi(it,iu,ivk,1,1) = ui1 + vi1
          yr(it,iu,ivk,5,1) = ur2 + vr2
          yi(it,iu,ivk,5,1) = ui2 + vi2
          yr(it,iu,ivk,3,1) = ur3 + vr3
          yi(it,iu,ivk,3,1) = ui3 + vi3
          yr(it,iu,ivk,4,1) = ur1 - vr1
          yi(it,iu,ivk,4,1) = ui1 - vi1
          yr(it,iu,ivk,2,1) = ur2 - vr2
          yi(it,iu,ivk,2,1) = ui2 - vi2
          yr(it,iu,ivk,6,1) = ur3 - vr3
          yi(it,iu,ivk,6,1) = ui3 - vi3
       end do
    end do

    iw = 6
    do p = 2,np
       wr1 = w(1,iw)
       wi1 = w(2,iw)
       iw = iw+1
       wr2 = w(1,iw)
       wi2 = w(2,iw)
       iw = iw+1
       wr3 = w(1,iw)
       wi3 = w(2,iw)
       iw = iw+1
       wr4 = w(1,iw)
       wi4 = w(2,iw)
       iw = iw+1
       wr5 = w(1,iw)
       wi5 = w(2,iw)
       iw = iw+1
       do ivk = 1,nvk
          do it=1,nt
             tr1 = xr(it,iu,ivk,p,3) + xr(it,iu,ivk,p,5)
             ti1 = xi(it,iu,ivk,p,3) + xi(it,iu,ivk,p,5)
             tr2 = xr(it,iu,ivk,p,1) - 0.5d0*tr1
             ti2 = xi(it,iu,ivk,p,1) - 0.5d0*ti1
             tr3 = -sp3*(xr(it,iu,ivk,p,3) - xr(it,iu,ivk,p,5))
             ti3 = -sp3*(xi(it,iu,ivk,p,3) - xi(it,iu,ivk,p,5))
             ur1 = xr(it,iu,ivk,p,1) + tr1
             ui1 = xi(it,iu,ivk,p,1) + ti1
             ur2 = tr2 - ti3
             ui2 = ti2 + tr3
             ur3 = tr2 + ti3
             ui3 = ti2 - tr3

             tr1 = xr(it,iu,ivk,p,6) + xr(it,iu,ivk,p,2)
             ti1 = xi(it,iu,ivk,p,6) + xi(it,iu,ivk,p,2)
             tr2 = xr(it,iu,ivk,p,4) - 0.5d0*tr1
             ti2 = xi(it,iu,ivk,p,4) - 0.5d0*ti1
             tr3 = -sp3*(xr(it,iu,ivk,p,6) - xr(it,iu,ivk,p,2))
             ti3 = -sp3*(xi(it,iu,ivk,p,6) - xi(it,iu,ivk,p,2))
             vr1 = xr(it,iu,ivk,p,4) + tr1
             vi1 = xi(it,iu,ivk,p,4) + ti1
             vr2 = tr2 - ti3
             vi2 = ti2 + tr3
             vr3 = tr2 + ti3
             vi3 = ti2 - tr3

             yr(it,iu,ivk,1,p) = ur1 + vr1
             yi(it,iu,ivk,1,p) = ui1 + vi1
             tr = ur2 + vr2
             ti = ui2 + vi2
             yr(it,iu,ivk,5,p) = wr4*tr - wi4*ti
             yi(it,iu,ivk,5,p) = wr4*ti + wi4*tr
             tr = ur3 + vr3
             ti = ui3 + vi3
             yr(it,iu,ivk,3,p) = wr2*tr - wi2*ti
             yi(it,iu,ivk,3,p) = wr2*ti + wi2*tr
             tr = ur1 - vr1
             ti = ui1 - vi1
             yr(it,iu,ivk,4,p) = wr3*tr - wi3*ti
             yi(it,iu,ivk,4,p) = wr3*ti + wi3*tr
             tr = ur2 - vr2
             ti = ui2 - vi2
             yr(it,iu,ivk,2,p) = wr1*tr - wi1*ti
             yi(it,iu,ivk,2,p) = wr1*ti + wi1*tr
             tr = ur3 - vr3
             ti = ui3 - vi3
             yr(it,iu,ivk,6,p) = wr5*tr - wi5*ti
             yi(it,iu,ivk,6,p) = wr5*ti + wi5*tr
          end do
       end do
    end do
    return
  end subroutine fftrd6b

  subroutine bftrd6b(nt,nu,iu,nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nt,nu,iu,nvk,np
    real(kind=8),intent(in):: xr(nt,nu,nvk,np,6),xi(nt,nu,nvk,np,6)
    real(kind=8),intent(out):: yr(nt,nu,nvk,6,np),yi(nt,nu,nvk,6,np)
    real(kind=8),intent(in):: w(2,5*np)
    !
    real(kind=8),parameter:: sp3 = 0.866025403784438647d0
    !
    real(kind=8):: wr1,wi1,wr2,wi2,wr3,wi3,wr4,wi4,wr5,wi5
    real(kind=8):: tr1,ti1,tr2,ti2,tr3,ti3,tr,ti
    real(kind=8):: ur1,ui1,ur2,ui2,ur3,ui3
    real(kind=8):: vr1,vi1,vr2,vi2,vr3,vi3
    integer:: p,ivk,iw,it
    !
    do ivk = 1,nvk
       do it = 1, nt
          tr1 = xr(it,iu,ivk,1,3) + xr(it,iu,ivk,1,5)
          ti1 = xi(it,iu,ivk,1,3) + xi(it,iu,ivk,1,5)
          tr2 = xr(it,iu,ivk,1,1) - 0.5d0*tr1
          ti2 = xi(it,iu,ivk,1,1) - 0.5d0*ti1
          tr3 = sp3*(xr(it,iu,ivk,1,3) - xr(it,iu,ivk,1,5))
          ti3 = sp3*(xi(it,iu,ivk,1,3) - xi(it,iu,ivk,1,5))
          ur1 = xr(it,iu,ivk,1,1) + tr1
          ui1 = xi(it,iu,ivk,1,1) + ti1
          ur2 = tr2 - ti3
          ui2 = ti2 + tr3
          ur3 = tr2 + ti3
          ui3 = ti2 - tr3

          tr1 = xr(it,iu,ivk,1,6) + xr(it,iu,ivk,1,2)
          ti1 = xi(it,iu,ivk,1,6) + xi(it,iu,ivk,1,2)
          tr2 = xr(it,iu,ivk,1,4) - 0.5d0*tr1
          ti2 = xi(it,iu,ivk,1,4) - 0.5d0*ti1
          tr3 = sp3*(xr(it,iu,ivk,1,6) - xr(it,iu,ivk,1,2))
          ti3 = sp3*(xi(it,iu,ivk,1,6) - xi(it,iu,ivk,1,2))
          vr1 = xr(it,iu,ivk,1,4) + tr1
          vi1 = xi(it,iu,ivk,1,4) + ti1
          vr2 = tr2 - ti3
          vi2 = ti2 + tr3
          vr3 = tr2 + ti3
          vi3 = ti2 - tr3

          yr(it,iu,ivk,1,1) = ur1 + vr1
          yi(it,iu,ivk,1,1) = ui1 + vi1
          yr(it,iu,ivk,5,1) = ur2 + vr2
          yi(it,iu,ivk,5,1) = ui2 + vi2
          yr(it,iu,ivk,3,1) = ur3 + vr3
          yi(it,iu,ivk,3,1) = ui3 + vi3
          yr(it,iu,ivk,4,1) = ur1 - vr1
          yi(it,iu,ivk,4,1) = ui1 - vi1
          yr(it,iu,ivk,2,1) = ur2 - vr2
          yi(it,iu,ivk,2,1) = ui2 - vi2
          yr(it,iu,ivk,6,1) = ur3 - vr3
          yi(it,iu,ivk,6,1) = ui3 - vi3
       end do
    end do

    iw = 6
    do p = 2,np
       wr1 = w(1,iw)
       wi1 = w(2,iw)
       iw = iw+1
       wr2 = w(1,iw)
       wi2 = w(2,iw)
       iw = iw+1
       wr3 = w(1,iw)
       wi3 = w(2,iw)
       iw = iw+1
       wr4 = w(1,iw)
       wi4 = w(2,iw)
       iw = iw+1
       wr5 = w(1,iw)
       wi5 = w(2,iw)
       iw = iw+1
       do ivk = 1,nvk
          do it = 1,nt 
             tr1 = xr(it,iu,ivk,p,3) + xr(it,iu,ivk,p,5)
             ti1 = xi(it,iu,ivk,p,3) + xi(it,iu,ivk,p,5)
             tr2 = xr(it,iu,ivk,p,1) - 0.5d0*tr1
             ti2 = xi(it,iu,ivk,p,1) - 0.5d0*ti1
             tr3 = sp3*(xr(it,iu,ivk,p,3) - xr(it,iu,ivk,p,5))
             ti3 = sp3*(xi(it,iu,ivk,p,3) - xi(it,iu,ivk,p,5))
             ur1 = xr(it,iu,ivk,p,1) + tr1
             ui1 = xi(it,iu,ivk,p,1) + ti1
             ur2 = tr2 - ti3
             ui2 = ti2 + tr3
             ur3 = tr2 + ti3
             ui3 = ti2 - tr3

             tr1 = xr(it,iu,ivk,p,6) + xr(it,iu,ivk,p,2)
             ti1 = xi(it,iu,ivk,p,6) + xi(it,iu,ivk,p,2)
             tr2 = xr(it,iu,ivk,p,4) - 0.5d0*tr1
             ti2 = xi(it,iu,ivk,p,4) - 0.5d0*ti1
             tr3 = sp3*(xr(it,iu,ivk,p,6) - xr(it,iu,ivk,p,2))
             ti3 = sp3*(xi(it,iu,ivk,p,6) - xi(it,iu,ivk,p,2))
             vr1 = xr(it,iu,ivk,p,4) + tr1
             vi1 = xi(it,iu,ivk,p,4) + ti1
             vr2 = tr2 - ti3
             vi2 = ti2 + tr3
             vr3 = tr2 + ti3
             vi3 = ti2 - tr3

             yr(it,iu,ivk,1,p) = ur1 + vr1
             yi(it,iu,ivk,1,p) = ui1 + vi1
             tr = ur2 + vr2
             ti = ui2 + vi2
             yr(it,iu,ivk,5,p) = wr4*tr + wi4*ti
             yi(it,iu,ivk,5,p) = wr4*ti - wi4*tr
             tr = ur3 + vr3
             ti = ui3 + vi3
             yr(it,iu,ivk,3,p) = wr2*tr + wi2*ti
             yi(it,iu,ivk,3,p) = wr2*ti - wi2*tr
             tr = ur1 - vr1
             ti = ui1 - vi1
             yr(it,iu,ivk,4,p) = wr3*tr + wi3*ti
             yi(it,iu,ivk,4,p) = wr3*ti - wi3*tr
             tr = ur2 - vr2
             ti = ui2 - vi2
             yr(it,iu,ivk,2,p) = wr1*tr + wi1*ti
             yi(it,iu,ivk,2,p) = wr1*ti - wi1*tr
             tr = ur3 - vr3
             ti = ui3 - vi3
             yr(it,iu,ivk,6,p) = wr5*tr + wi5*ti
             yi(it,iu,ivk,6,p) = wr5*ti - wi5*tr
          end do
       end do
    end do
    return
  end subroutine bftrd6b
  !
  subroutine fftrd8b(nt,nu,iu,nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nt,nu,iu,nvk,np
    real(kind=8),intent(in):: xr(nt,nu,nvk,np,8),xi(nt,nu,nvk,np,8)
    real(kind=8),intent(out):: yr(nt,nu,nvk,8,np),yi(nt,nu,nvk,8,np)
    real(kind=8),intent(in):: w(2,7*np)
    !
    real(kind=8),parameter:: sp2 = 0.70710678118654752440d0
    !
    real(kind=8):: wr1,wi1,wr2,wi2,wr3,wi3,wr4,wi4,wr5,wi5,wr6,wi6,wr7,wi7
    real(kind=8):: tr1,ti1,tr2,ti2,tr3,ti3,tr4,ti4,tr,ti,sr,si
    real(kind=8):: ur1,ui1,ur2,ui2,ur3,ui3,ur4,ui4
    real(kind=8):: vr1,vi1,vr2,vi2,vr3,vi3,vr4,vi4
    integer:: p,ivk,iw,it
    !
    do ivk = 1,nvk
       do it=1,nt
          tr1 = xr(it,iu,ivk,1,1) + xr(it,iu,ivk,1,5)
          ti1 = xi(it,iu,ivk,1,1) + xi(it,iu,ivk,1,5)
          tr2 = xr(it,iu,ivk,1,1) - xr(it,iu,ivk,1,5)
          ti2 = xi(it,iu,ivk,1,1) - xi(it,iu,ivk,1,5)
          tr3 = xr(it,iu,ivk,1,3) + xr(it,iu,ivk,1,7)
          ti3 = xi(it,iu,ivk,1,3) + xi(it,iu,ivk,1,7)
          tr4 = xi(it,iu,ivk,1,3) - xi(it,iu,ivk,1,7)
          ti4 = xr(it,iu,ivk,1,7) - xr(it,iu,ivk,1,3)
          ur1 = tr1 + tr3
          ui1 = ti1 + ti3
          ur2 = tr2 + tr4
          ui2 = ti2 + ti4
          ur3 = tr1 - tr3
          ui3 = ti1 - ti3
          ur4 = tr2 - tr4
          ui4 = ti2 - ti4
          tr1 = xr(it,iu,ivk,1,2) + xr(it,iu,ivk,1,6)
          ti1 = xi(it,iu,ivk,1,2) + xi(it,iu,ivk,1,6)
          tr2 = xr(it,iu,ivk,1,2) - xr(it,iu,ivk,1,6)
          ti2 = xi(it,iu,ivk,1,2) - xi(it,iu,ivk,1,6)
          tr3 = xr(it,iu,ivk,1,4) + xr(it,iu,ivk,1,8)
          ti3 = xi(it,iu,ivk,1,4) + xi(it,iu,ivk,1,8)
          tr4 = xi(it,iu,ivk,1,4) - xi(it,iu,ivk,1,8)
          ti4 = xr(it,iu,ivk,1,8) - xr(it,iu,ivk,1,4)
          vr1 = tr1 + tr3
          vi1 = ti1 + ti3
          vr2 = tr2 + tr4
          vi2 = ti2 + ti4
          vr3 = tr1 - tr3
          vi3 = ti1 - ti3
          vr4 = tr2 - tr4
          vi4 = ti2 - ti4
          yr(it,iu,ivk,1,1) = ur1 + vr1
          yi(it,iu,ivk,1,1) = ui1 + vi1
          yr(it,iu,ivk,5,1) = ur1 - vr1
          yi(it,iu,ivk,5,1) = ui1 - vi1

          sr = sp2*(vr2 + vi2)
          si = sp2*(vi2 - vr2)
          yr(it,iu,ivk,2,1) = ur2 + sr
          yi(it,iu,ivk,2,1) = ui2 + si
          yr(it,iu,ivk,6,1) = ur2 - sr
          yi(it,iu,ivk,6,1) = ui2 - si

          yr(it,iu,ivk,3,1) = ur3 + vi3
          yi(it,iu,ivk,3,1) = ui3 - vr3
          yr(it,iu,ivk,7,1) = ur3 - vi3
          yi(it,iu,ivk,7,1) = ui3 + vr3

          sr = sp2*(vi4 - vr4)
          si = -sp2*(vr4 + vi4)
          yr(it,iu,ivk,4,1) = ur4 + sr
          yi(it,iu,ivk,4,1) = ui4 + si
          yr(it,iu,ivk,8,1) = ur4 - sr
          yi(it,iu,ivk,8,1) = ui4 - si
       end do
    end do

    iw = 8
    do p = 2,np
       wr1 = w(1,iw)
       wi1 = w(2,iw)
       iw = iw+1
       wr2 = w(1,iw)
       wi2 = w(2,iw)
       iw = iw+1
       wr3 = w(1,iw)
       wi3 = w(2,iw)
       iw = iw+1
       wr4 = w(1,iw)
       wi4 = w(2,iw)
       iw = iw+1
       wr5 = w(1,iw)
       wi5 = w(2,iw)
       iw = iw+1
       wr6 = w(1,iw)
       wi6 = w(2,iw)
       iw = iw+1
       wr7 = w(1,iw)
       wi7 = w(2,iw)
       iw = iw+1
       do ivk = 1,nvk
          do it=1,nt
             tr1 = xr(it,iu,ivk,p,1) + xr(it,iu,ivk,p,5)
             ti1 = xi(it,iu,ivk,p,1) + xi(it,iu,ivk,p,5)
             tr2 = xr(it,iu,ivk,p,1) - xr(it,iu,ivk,p,5)
             ti2 = xi(it,iu,ivk,p,1) - xi(it,iu,ivk,p,5)
             tr3 = xr(it,iu,ivk,p,3) + xr(it,iu,ivk,p,7)
             ti3 = xi(it,iu,ivk,p,3) + xi(it,iu,ivk,p,7)
             tr4 = xi(it,iu,ivk,p,3) - xi(it,iu,ivk,p,7)
             ti4 = xr(it,iu,ivk,p,7) - xr(it,iu,ivk,p,3)
             ur1 = tr1 + tr3
             ui1 = ti1 + ti3
             ur2 = tr2 + tr4
             ui2 = ti2 + ti4
             ur3 = tr1 - tr3
             ui3 = ti1 - ti3
             ur4 = tr2 - tr4
             ui4 = ti2 - ti4
             tr1 = xr(it,iu,ivk,p,2) + xr(it,iu,ivk,p,6)
             ti1 = xi(it,iu,ivk,p,2) + xi(it,iu,ivk,p,6)
             tr2 = xr(it,iu,ivk,p,2) - xr(it,iu,ivk,p,6)
             ti2 = xi(it,iu,ivk,p,2) - xi(it,iu,ivk,p,6)
             tr3 = xr(it,iu,ivk,p,4) + xr(it,iu,ivk,p,8)
             ti3 = xi(it,iu,ivk,p,4) + xi(it,iu,ivk,p,8)
             tr4 = xi(it,iu,ivk,p,4) - xi(it,iu,ivk,p,8)
             ti4 = xr(it,iu,ivk,p,8) - xr(it,iu,ivk,p,4)
             vr1 = tr1 + tr3
             vi1 = ti1 + ti3
             vr2 = tr2 + tr4
             vi2 = ti2 + ti4
             vr3 = tr1 - tr3
             vi3 = ti1 - ti3
             vr4 = tr2 - tr4
             vi4 = ti2 - ti4
             yr(it,iu,ivk,1,p) = ur1 + vr1
             yi(it,iu,ivk,1,p) = ui1 + vi1
             tr = ur1 - vr1
             ti = ui1 - vi1
             yr(it,iu,ivk,5,p) = wr4*tr - wi4*ti
             yi(it,iu,ivk,5,p) = wr4*ti + wi4*tr

             sr = sp2*(vr2 + vi2)
             si = sp2*(vi2 - vr2)
             tr = ur2 + sr
             ti = ui2 + si
             yr(it,iu,ivk,2,p) = wr1*tr - wi1*ti
             yi(it,iu,ivk,2,p) = wr1*ti + wi1*tr
             tr = ur2 - sr
             ti = ui2 - si
             yr(it,iu,ivk,6,p) = wr5*tr - wi5*ti
             yi(it,iu,ivk,6,p) = wr5*ti + wi5*tr

             tr = ur3 + vi3
             ti = ui3 - vr3
             yr(it,iu,ivk,3,p) = wr2*tr - wi2*ti
             yi(it,iu,ivk,3,p) = wr2*ti + wi2*tr
             tr = ur3 - vi3
             ti = ui3 + vr3
             yr(it,iu,ivk,7,p) = wr6*tr - wi6*ti
             yi(it,iu,ivk,7,p) = wr6*ti + wi6*tr

             sr = sp2*(vi4 - vr4)
             si = -sp2*(vr4 + vi4)
             tr = ur4 + sr
             ti = ui4 + si
             yr(it,iu,ivk,4,p) = wr3*tr - wi3*ti
             yi(it,iu,ivk,4,p) = wr3*ti + wi3*tr
             tr = ur4 - sr
             ti = ui4 - si
             yr(it,iu,ivk,8,p) = wr7*tr - wi7*ti
             yi(it,iu,ivk,8,p) = wr7*ti + wi7*tr
          end do
       end do
    end do
    return
  end subroutine fftrd8b
  !
  subroutine bftrd8b(nt,nu,iu,nvk,np,w,xr,xi,yr,yi)
    implicit none
    integer,intent(in):: nt,nu,iu,nvk,np
    real(kind=8),intent(in):: xr(nt,nu,nvk,np,8),xi(nt,nu,nvk,np,8)
    real(kind=8),intent(out):: yr(nt,nu,nvk,8,np),yi(nt,nu,nvk,8,np)
    real(kind=8),intent(in):: w(2,7*np)
    !
    real(kind=8),parameter:: sp2 = 0.70710678118654752440d0
    !
    real(kind=8):: wr1,wi1,wr2,wi2,wr3,wi3,wr4,wi4,wr5,wi5,wr6,wi6,wr7,wi7
    real(kind=8):: tr1,ti1,tr2,ti2,tr3,ti3,tr4,ti4,tr,ti,sr,si
    real(kind=8):: ur1,ui1,ur2,ui2,ur3,ui3,ur4,ui4
    real(kind=8):: vr1,vi1,vr2,vi2,vr3,vi3,vr4,vi4
    integer:: p,ivk,iw,it
    !
    do ivk = 1,nvk
       do it = 1,nt
          tr1 = xr(it,iu,ivk,1,1) + xr(it,iu,ivk,1,5)
          ti1 = xi(it,iu,ivk,1,1) + xi(it,iu,ivk,1,5)
          tr2 = xr(it,iu,ivk,1,1) - xr(it,iu,ivk,1,5)
          ti2 = xi(it,iu,ivk,1,1) - xi(it,iu,ivk,1,5)
          tr3 = xr(it,iu,ivk,1,3) + xr(it,iu,ivk,1,7)
          ti3 = xi(it,iu,ivk,1,3) + xi(it,iu,ivk,1,7)
          tr4 = xi(it,iu,ivk,1,7) - xi(it,iu,ivk,1,3)
          ti4 = xr(it,iu,ivk,1,3) - xr(it,iu,ivk,1,7)
          ur1 = tr1 + tr3
          ui1 = ti1 + ti3
          ur2 = tr2 + tr4
          ui2 = ti2 + ti4
          ur3 = tr1 - tr3
          ui3 = ti1 - ti3
          ur4 = tr2 - tr4
          ui4 = ti2 - ti4
          tr1 = xr(it,iu,ivk,1,2) + xr(it,iu,ivk,1,6)
          ti1 = xi(it,iu,ivk,1,2) + xi(it,iu,ivk,1,6)
          tr2 = xr(it,iu,ivk,1,2) - xr(it,iu,ivk,1,6)
          ti2 = xi(it,iu,ivk,1,2) - xi(it,iu,ivk,1,6)
          tr3 = xr(it,iu,ivk,1,4) + xr(it,iu,ivk,1,8)
          ti3 = xi(it,iu,ivk,1,4) + xi(it,iu,ivk,1,8)
          tr4 = xi(it,iu,ivk,1,8) - xi(it,iu,ivk,1,4)
          ti4 = xr(it,iu,ivk,1,4) - xr(it,iu,ivk,1,8)
          vr1 = tr1 + tr3
          vi1 = ti1 + ti3
          vr2 = tr2 + tr4
          vi2 = ti2 + ti4
          vr3 = tr1 - tr3
          vi3 = ti1 - ti3
          vr4 = tr2 - tr4
          vi4 = ti2 - ti4
          yr(it,iu,ivk,1,1) = ur1 + vr1
          yi(it,iu,ivk,1,1) = ui1 + vi1
          yr(it,iu,ivk,5,1) = ur1 - vr1
          yi(it,iu,ivk,5,1) = ui1 - vi1

          sr = sp2*(vr2 - vi2)
          si = sp2*(vi2 + vr2)
          yr(it,iu,ivk,2,1) = ur2 + sr
          yi(it,iu,ivk,2,1) = ui2 + si
          yr(it,iu,ivk,6,1) = ur2 - sr
          yi(it,iu,ivk,6,1) = ui2 - si

          yr(it,iu,ivk,3,1) = ur3 - vi3
          yi(it,iu,ivk,3,1) = ui3 + vr3
          yr(it,iu,ivk,7,1) = ur3 + vi3
          yi(it,iu,ivk,7,1) = ui3 - vr3

          sr = -sp2*(vr4 + vi4)
          si = sp2*(vr4 - vi4)
          yr(it,iu,ivk,4,1) = ur4 + sr
          yi(it,iu,ivk,4,1) = ui4 + si
          yr(it,iu,ivk,8,1) = ur4 - sr
          yi(it,iu,ivk,8,1) = ui4 - si
       end do
    end do

    iw = 8
    do p = 2,np
       wr1 = w(1,iw)
       wi1 = w(2,iw)
       iw = iw+1
       wr2 = w(1,iw)
       wi2 = w(2,iw)
       iw = iw+1
       wr3 = w(1,iw)
       wi3 = w(2,iw)
       iw = iw+1
       wr4 = w(1,iw)
       wi4 = w(2,iw)
       iw = iw+1
       wr5 = w(1,iw)
       wi5 = w(2,iw)
       iw = iw+1
       wr6 = w(1,iw)
       wi6 = w(2,iw)
       iw = iw+1
       wr7 = w(1,iw)
       wi7 = w(2,iw)
       iw = iw+1
       do ivk = 1,nvk
          do it = 1,nt
             tr1 = xr(it,iu,ivk,p,1) + xr(it,iu,ivk,p,5)
             ti1 = xi(it,iu,ivk,p,1) + xi(it,iu,ivk,p,5)
             tr2 = xr(it,iu,ivk,p,1) - xr(it,iu,ivk,p,5)
             ti2 = xi(it,iu,ivk,p,1) - xi(it,iu,ivk,p,5)
             tr3 = xr(it,iu,ivk,p,3) + xr(it,iu,ivk,p,7)
             ti3 = xi(it,iu,ivk,p,3) + xi(it,iu,ivk,p,7)
             tr4 = xi(it,iu,ivk,p,7) - xi(it,iu,ivk,p,3)
             ti4 = xr(it,iu,ivk,p,3) - xr(it,iu,ivk,p,7)
             ur1 = tr1 + tr3
             ui1 = ti1 + ti3
             ur2 = tr2 + tr4
             ui2 = ti2 + ti4
             ur3 = tr1 - tr3
             ui3 = ti1 - ti3
             ur4 = tr2 - tr4
             ui4 = ti2 - ti4
             tr1 = xr(it,iu,ivk,p,2) + xr(it,iu,ivk,p,6)
             ti1 = xi(it,iu,ivk,p,2) + xi(it,iu,ivk,p,6)
             tr2 = xr(it,iu,ivk,p,2) - xr(it,iu,ivk,p,6)
             ti2 = xi(it,iu,ivk,p,2) - xi(it,iu,ivk,p,6)
             tr3 = xr(it,iu,ivk,p,4) + xr(it,iu,ivk,p,8)
             ti3 = xi(it,iu,ivk,p,4) + xi(it,iu,ivk,p,8)
             tr4 = xi(it,iu,ivk,p,8) - xi(it,iu,ivk,p,4)
             ti4 = xr(it,iu,ivk,p,4) - xr(it,iu,ivk,p,8)
             vr1 = tr1 + tr3
             vi1 = ti1 + ti3
             vr2 = tr2 + tr4
             vi2 = ti2 + ti4
             vr3 = tr1 - tr3
             vi3 = ti1 - ti3
             vr4 = tr2 - tr4
             vi4 = ti2 - ti4
             yr(it,iu,ivk,1,p) = ur1 + vr1
             yi(it,iu,ivk,1,p) = ui1 + vi1
             tr = ur1 - vr1
             ti = ui1 - vi1
             yr(it,iu,ivk,5,p) = wr4*tr + wi4*ti
             yi(it,iu,ivk,5,p) = wr4*ti - wi4*tr

             sr = sp2*(vr2 - vi2)
             si = sp2*(vi2 + vr2)
             tr = ur2 + sr
             ti = ui2 + si
             yr(it,iu,ivk,2,p) = wr1*tr + wi1*ti
             yi(it,iu,ivk,2,p) = wr1*ti - wi1*tr
             tr = ur2 - sr
             ti = ui2 - si
             yr(it,iu,ivk,6,p) = wr5*tr + wi5*ti
             yi(it,iu,ivk,6,p) = wr5*ti - wi5*tr

             tr = ur3 - vi3
             ti = ui3 + vr3
             yr(it,iu,ivk,3,p) = wr2*tr + wi2*ti
             yi(it,iu,ivk,3,p) = wr2*ti - wi2*tr
             tr = ur3 + vi3
             ti = ui3 - vr3
             yr(it,iu,ivk,7,p) = wr6*tr + wi6*ti
             yi(it,iu,ivk,7,p) = wr6*ti - wi6*tr

             sr = -sp2*(vr4 + vi4)
             si = sp2*(vr4 - vi4)
             tr = ur4 + sr
             ti = ui4 + si
             yr(it,iu,ivk,4,p) = wr3*tr + wi3*ti
             yi(it,iu,ivk,4,p) = wr3*ti - wi3*tr
             tr = ur4 - sr
             ti = ui4 - si
             yr(it,iu,ivk,8,p) = wr7*tr + wi7*ti
             yi(it,iu,ivk,8,p) = wr7*ti - wi7*tr
          end do
       end do
    end do
    return
  end subroutine bftrd8b
  !
  subroutine fftv(nv,n,nf,fac,w,lxyz,dr,di,er,ei,ix)
    implicit none
    integer,intent(in):: nv,n,lxyz
    integer,intent(in):: nf,fac(nfmax)
    real(kind=8),intent(in):: w(2,n)
    real(kind=8),dimension(lxyz):: dr,di,er,ei
    integer:: ix
    !
    integer:: iw,nvk,np,uif,if
    !
    iw = 1
    nvk = nv
    np = n/fac(1)
    uif = mod(nf,2)
    do if = 1,nf-uif,2
       select case (fac(if))
       case (2)
          call fftrd2(nvk,np,w(1,iw),dr,di,er,ei)
       case (3)
          call fftrd3(nvk,np,w(1,iw),dr,di,er,ei)
       case (4)
          call fftrd4(nvk,np,w(1,iw),dr,di,er,ei)
       case (5)
          call fftrd5(nvk,np,w(1,iw),dr,di,er,ei)
       case (6)
          call fftrd6(nvk,np,w(1,iw),dr,di,er,ei)
       case (8)
          call fftrd8(nvk,np,w(1,iw),dr,di,er,ei)
       case default
          write(6,*)'unsupported factor'
          stop
       end select
       iw=iw+np*(fac(if)-1)
       nvk=nvk*fac(if)
       np=np/fac(if+1)
       select case (fac(if+1))
       case (2)
          call fftrd2(nvk,np,w(1,iw),er,ei,dr,di)
       case (3)
          call fftrd3(nvk,np,w(1,iw),er,ei,dr,di)
       case (4)
          call fftrd4(nvk,np,w(1,iw),er,ei,dr,di)
       case (5)
          call fftrd5(nvk,np,w(1,iw),er,ei,dr,di)
       case (6)
          call fftrd6(nvk,np,w(1,iw),er,ei,dr,di)
       case (8)
          call fftrd8(nvk,np,w(1,iw),er,ei,dr,di)
       case default
          write(6,*)'unsupported factor'
          stop
       end select
       iw=iw+np*(fac(if+1)-1)
       nvk=nvk*fac(if+1)
       if (if+2 .le. nf) then
          np=np/fac(if+2)
       end if
    end do
    if (uif .eq. 1) then
       select case (fac(nf))
       case (2)
          call fftrd2(nvk,np,w(1,iw),dr,di,er,ei)
       case (3)
          call fftrd3(nvk,np,w(1,iw),dr,di,er,ei)
       case (4)
          call fftrd4(nvk,np,w(1,iw),dr,di,er,ei)
       case (5)
          call fftrd5(nvk,np,w(1,iw),dr,di,er,ei)
       case (6)
          call fftrd6(nvk,np,w(1,iw),dr,di,er,ei)
       case (8)
          call fftrd8(nvk,np,w(1,iw),dr,di,er,ei)
       case default
          write(6,*)'unsupported factor'
          stop
       end select
       ix = - ix
    end if

    return
  end subroutine fftv
  !
  subroutine bftv(nv,n,nf,fac,w,lxyz,dr,di,er,ei,ix)
    implicit none
    integer,intent(in):: nv,n,lxyz
    integer,intent(in):: nf,fac(nfmax)
    real(kind=8),intent(in):: w(2,n)
    real(kind=8),dimension(lxyz):: dr,di,er,ei
    integer:: ix
    !
    integer:: iw,nvk,np,uif,if
    !
    iw = 1
    nvk = nv
    np = n/fac(1)
    uif = mod(nf,2)
    do if = 1,nf-uif,2
       select case (fac(if))
       case (2)
          call bftrd2(nvk,np,w(1,iw),dr,di,er,ei)
       case (3)
          call bftrd3(nvk,np,w(1,iw),dr,di,er,ei)
       case (4)
          call bftrd4(nvk,np,w(1,iw),dr,di,er,ei)
       case (5)
          call bftrd5(nvk,np,w(1,iw),dr,di,er,ei)
       case (6)
          call bftrd6(nvk,np,w(1,iw),dr,di,er,ei)
       case (8)
          call bftrd8(nvk,np,w(1,iw),dr,di,er,ei)
       case default
          write(6,*)'unsupported factor'
          stop
       end select
       iw=iw+np*(fac(if)-1)
       nvk=nvk*fac(if)
       np=np/fac(if+1)
       select case(fac(if+1))
       case (2)
          call bftrd2(nvk,np,w(1,iw),er,ei,dr,di)
       case (3)
          call bftrd3(nvk,np,w(1,iw),er,ei,dr,di)
       case (4)
          call bftrd4(nvk,np,w(1,iw),er,ei,dr,di)
       case (5)
          call bftrd5(nvk,np,w(1,iw),er,ei,dr,di)
       case (6)
          call bftrd6(nvk,np,w(1,iw),er,ei,dr,di)
       case (8)
          call bftrd8(nvk,np,w(1,iw),er,ei,dr,di)
       case default
          write(6,*)'unsupported factor'
          stop
       end select
       iw=iw+np*(fac(if+1)-1)
       nvk=nvk*fac(if+1)
       if (if+2 .le. nf) then
          np=np/fac(if+2)
       end if
    end do
    if (uif .eq. 1) then
       select case (fac(nf))
       case (2)
          call bftrd2(nvk,np,w(1,iw),dr,di,er,ei)
       case (3)
          call bftrd3(nvk,np,w(1,iw),dr,di,er,ei)
       case (4)
          call bftrd4(nvk,np,w(1,iw),dr,di,er,ei)
       case (5)
          call bftrd5(nvk,np,w(1,iw),dr,di,er,ei)
       case (6)
          call bftrd6(nvk,np,w(1,iw),dr,di,er,ei)
       case (8)
          call bftrd8(nvk,np,w(1,iw),dr,di,er,ei)
       case default
          write(6,*)'unsupported factor'
          stop
       end select
       ix = - ix
    end if

    return
  end subroutine bftv

  subroutine fftvb(nt,nu,iu,n,nf,fac,w,lxyz,dr,di,er,ei,ix)
    implicit none
    integer,intent(in):: nt,nu,iu,n,lxyz
    integer,intent(in):: nf,fac(nfmax)
    real(kind=8),intent(in):: w(2,n)
    real(kind=8),dimension(lxyz):: dr,di,er,ei
    integer:: ix
    !
    integer:: iw,nvk,np,uif,if
    !
    iw = 1
    nvk = 1
    np = n/fac(1)
    uif = mod(nf,2)
    do if = 1,nf-uif,2
       select case (fac(if))
       case (2)
          call fftrd2b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case (3)
          call fftrd3b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case (4)
          call fftrd4b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case (5)
          call fftrd5b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case (6)
          call fftrd6b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case (8)
          call fftrd8b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case default
          write(6,*)'unsupported factor'
          stop
       end select
       iw=iw+np*(fac(if)-1)
       nvk=nvk*fac(if)
       np=np/fac(if+1)
       select case (fac(if+1))
       case (2)
          call fftrd2b(nt,nu,iu,nvk,np,w(1,iw),er,ei,dr,di)
       case (3)
          call fftrd3b(nt,nu,iu,nvk,np,w(1,iw),er,ei,dr,di)
       case (4)
          call fftrd4b(nt,nu,iu,nvk,np,w(1,iw),er,ei,dr,di)
       case (5)
          call fftrd5b(nt,nu,iu,nvk,np,w(1,iw),er,ei,dr,di)
       case (6)
          call fftrd6b(nt,nu,iu,nvk,np,w(1,iw),er,ei,dr,di)
       case (8)
          call fftrd8b(nt,nu,iu,nvk,np,w(1,iw),er,ei,dr,di)
       case default
          write(6,*)'unsupported factor'
          stop
       end select
       iw=iw+np*(fac(if+1)-1)
       nvk=nvk*fac(if+1)
       if (if+2 .le. nf) then
          np=np/fac(if+2)
       end if
    end do
    if (uif .eq. 1) then
       select case (fac(nf))
       case (2)
          call fftrd2b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case (3)
          call fftrd3b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case (4)
          call fftrd4b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case (5)
          call fftrd5b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case (6)
          call fftrd6b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case (8)
          call fftrd8b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case default
          write(6,*)'unsupported factor'
          stop
       end select
       ix = - ix
    end if

    return
  end subroutine fftvb
  !
  subroutine bftvb(nt,nu,iu,n,nf,fac,w,lxyz,dr,di,er,ei,ix)
    implicit none
    integer,intent(in):: nt,nu,iu,n,lxyz
    integer,intent(in):: nf,fac(nfmax)
    real(kind=8),intent(in):: w(2,n)
    real(kind=8),dimension(lxyz):: dr,di,er,ei
    integer:: ix
    !
    integer:: iw,nvk,np,uif,if
    !
    iw = 1
    nvk = 1
    np = n/fac(1)
    uif = mod(nf,2)
    do if = 1,nf-uif,2
       select case (fac(if))
       case (2)
          call bftrd2b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case (3)
          call bftrd3b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case (4)
          call bftrd4b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case (5)
          call bftrd5b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case (6)
          call bftrd6b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case (8)
          call bftrd8b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case default
          write(6,*)'unsupported factor'
          stop
       end select
       iw=iw+np*(fac(if)-1)
       nvk=nvk*fac(if)
       np=np/fac(if+1)
       select case(fac(if+1))
       case (2)
          call bftrd2b(nt,nu,iu,nvk,np,w(1,iw),er,ei,dr,di)
       case (3)
          call bftrd3b(nt,nu,iu,nvk,np,w(1,iw),er,ei,dr,di)
       case (4)
          call bftrd4b(nt,nu,iu,nvk,np,w(1,iw),er,ei,dr,di)
       case (5)
          call bftrd5b(nt,nu,iu,nvk,np,w(1,iw),er,ei,dr,di)
       case (6)
          call bftrd6b(nt,nu,iu,nvk,np,w(1,iw),er,ei,dr,di)
       case (8)
          call bftrd8b(nt,nu,iu,nvk,np,w(1,iw),er,ei,dr,di)
       case default
          write(6,*)'unsupported factor'
          stop
       end select
       iw=iw+np*(fac(if+1)-1)
       nvk=nvk*fac(if+1)
       if (if+2 .le. nf) then
          np=np/fac(if+2)
       end if
    end do
    if (uif .eq. 1) then
       select case (fac(nf))
       case (2)
          call bftrd2b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case (3)
          call bftrd3b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case (4)
          call bftrd4b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case (5)
          call bftrd5b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case (6)
          call bftrd6b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case (8)
          call bftrd8b(nt,nu,iu,nvk,np,w(1,iw),dr,di,er,ei)
       case default
          write(6,*)'unsupported factor'
          stop
       end select
       ix = - ix
    end if

    return
  end subroutine bftvb

end module pfft3d_subr
