/*
  Copyright (c) 2012-2013 Yoshihide Yoshimoto
 */

#include <pmmintrin.h>

int fftrd2_(int *invk, int *inp, double *dw, double *dx, double *dy) {
	
	int nvk,np;
	int p,ivk;
	int iw,ix,iy;
	__m128d wr,wi,x0,x1,t1,t2,t3,t4;
	int tsd0, tsd1, tsd2;

	__m128d *w,*x,*y;

	nvk = *invk;
	np = *inp;
	w = (__m128d *)dw;
	x = (__m128d *)dx;
	y = (__m128d *)dy;

	tsd0 = nvk;
	tsd1 = nvk*2;
	tsd2 = nvk*np;

	ix = 0;
	iy = 0;
	iw = 0;
    for (p = 0; p < np; p++) {
		t1 = w[iw];
		wr = _mm_movedup_pd(t1);
		wi = _mm_unpackhi_pd(t1, t1);

		ix = tsd0*p;
		iy = tsd1*p;
		for (ivk = 0; ivk < nvk; ivk++) {
			x0 = x[ix];
			x1 = x[ix+tsd2];

			y[iy] = _mm_add_pd(x0,x1);

			t1 = _mm_sub_pd(x0,x1);

			t2 = _mm_shuffle_pd(t1,t1,_MM_SHUFFLE2(0,1));

			t3 = _mm_mul_pd(wr,t1);
			t4 = _mm_mul_pd(wi,t2);

			y[iy+tsd0] = _mm_addsub_pd(t3,t4);
			
			ix = ix + 1;
			iy = iy + 1;
		}
		iw = iw + 1;
	}
	return 0;
}

int bftrd2_(int *invk, int *inp, double *dw, double *dx, double *dy) {
	
	int nvk,np;
	int p,ivk;
	int iw,ix,iy;
	__m128d wr,wi,x0,x1,t1,t2,t3,t4;
	int tsd0, tsd1, tsd2;

	__m128d *w,*x,*y;

	nvk = *invk;
	np = *inp;
	w = (__m128d *)dw;
	x = (__m128d *)dx;
	y = (__m128d *)dy;

	tsd0 = nvk;
	tsd1 = nvk*2;
	tsd2 = nvk*np;

	ix = 0;
	iy = 0;
	iw = 0;
    for (p = 0; p < np; p++) {
		t1 = w[iw];
		t2 = _mm_setzero_pd();
		wr = _mm_movedup_pd(t1);
		wi = _mm_unpackhi_pd(t1, t1);
		wi = _mm_sub_pd(t2,wi);

		ix = tsd0*p;
		iy = tsd1*p;
		for (ivk = 0; ivk < nvk; ivk++) {
			x0 = x[ix];
			x1 = x[ix+tsd2];

			y[iy] = _mm_add_pd(x0,x1);

			t1 = _mm_sub_pd(x0,x1);

			t2 = _mm_shuffle_pd(t1,t1,_MM_SHUFFLE2(0,1));

			t3 = _mm_mul_pd(wr,t1);
			t4 = _mm_mul_pd(wi,t2);

			y[iy+tsd0] = _mm_addsub_pd(t3,t4);
			
			ix = ix + 1;
			iy = iy + 1;
		}
		iw = iw + 1;
	}
	return 0;
}

int fftrd3_(int *invk, int *inp, double *dw, double *dx, double *dy) {
	const double sp3 = 0.866025403784438647;

	int nvk,np;
	int p,ivk;
	int iw,ix,iy;
	__m128d wr1,wi1,wr2,wi2,x1,x2,x3;
	__m128d t1,t2,t3,t4,t5,t6,t7,tsp3,thlf;
	int tsd0, tsd1, tsd2;
	int tsd2m2;
	int tsd0m2;

	__m128d *w,*x,*y;

	nvk = *invk;
	np = *inp;
	w = (__m128d *)dw;
	x = (__m128d *)dx;
	y = (__m128d *)dy;

	tsd0 = nvk;
	tsd1 = nvk*3;
	tsd2 = nvk*np;

	tsd2m2 = tsd2*2;

	tsd0m2 = tsd0*2;

	tsp3 = _mm_set_pd(-sp3,sp3);
	thlf = _mm_set1_pd(0.5);

	ix = 0;
	iy = 0;
	iw = 0;
    for (p = 0; p < np; p++) {
		t1 = w[iw];
		wr1 = _mm_movedup_pd(t1);
		wi1 = _mm_unpackhi_pd(t1, t1);
		t1 = w[iw+1];
		wr2 = _mm_movedup_pd(t1);
		wi2 = _mm_unpackhi_pd(t1, t1);
		iw = iw + 2;

		ix = tsd0*p;
		iy = tsd1*p;
		for (ivk = 0; ivk < nvk; ivk++) {
			x1 = x[ix];
			x2 = x[ix+tsd2];
			x3 = x[ix+tsd2m2];

			t1 = _mm_add_pd(x2,x3);
			t3 = _mm_mul_pd(thlf,t1);
			t2 = _mm_sub_pd(x1,t3);

			t3 = _mm_sub_pd(x2,x3);
			t3 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
			t3 = _mm_mul_pd(tsp3,t3);

			y[iy] = _mm_add_pd(x1,t1);
			
			t4 = _mm_add_pd(t2,t3);
			t5 = _mm_shuffle_pd(t4,t4,_MM_SHUFFLE2(0,1));
			t6 = _mm_mul_pd(wr1,t4);
			t7 = _mm_mul_pd(wi1,t5);
			y[iy+tsd0] = _mm_addsub_pd(t6,t7);

			t4 = _mm_sub_pd(t2,t3);
			t5 = _mm_shuffle_pd(t4,t4,_MM_SHUFFLE2(0,1));
			t6 = _mm_mul_pd(wr2,t4);
			t7 = _mm_mul_pd(wi2,t5);
			y[iy+tsd0m2] = _mm_addsub_pd(t6,t7);

			ix = ix + 1;
			iy = iy + 1;
		}
	}
	return 0;
}

int bftrd3_(int *invk, int *inp, double *dw, double *dx, double *dy) {
	const double sp3 = 0.866025403784438647;

	int nvk,np;
	int p,ivk;
	int iw,ix,iy;
	__m128d wr1,wi1,wr2,wi2,x1,x2,x3;
	__m128d t1,t2,t3,t4,t5,t6,t7,tsp3,thlf;
	int tsd0, tsd1, tsd2;
	int tsd2m2;
	int tsd0m2;

	__m128d *w,*x,*y;

	nvk = *invk;
	np = *inp;
	w = (__m128d *)dw;
	x = (__m128d *)dx;
	y = (__m128d *)dy;

	tsd0 = nvk;
	tsd1 = nvk*3;
	tsd2 = nvk*np;

	tsd2m2 = tsd2*2;

	tsd0m2 = tsd0*2;

	tsp3 = _mm_set_pd(sp3,-sp3);
	thlf = _mm_set1_pd(0.5);

	ix = 0;
	iy = 0;
	iw = 0;
    for (p = 0; p < np; p++) {
		t1 = w[iw];
		t2 = _mm_setzero_pd();
		wr1 = _mm_movedup_pd(t1);
		wi1 = _mm_unpackhi_pd(t1, t1);
		wi1 = _mm_sub_pd(t2,wi1);
		t1 = w[iw+1];
		t2 = _mm_setzero_pd();
		wr2 = _mm_movedup_pd(t1);
		wi2 = _mm_unpackhi_pd(t1, t1);
		wi2 = _mm_sub_pd(t2,wi2);
		iw = iw + 2;

		ix = tsd0*p;
		iy = tsd1*p;
		for (ivk = 0; ivk < nvk; ivk++) {
			x1 = x[ix];
			x2 = x[ix+tsd2];
			x3 = x[ix+tsd2m2];

			t1 = _mm_add_pd(x2,x3);
			t3 = _mm_mul_pd(thlf,t1);
			t2 = _mm_sub_pd(x1,t3);

			t3 = _mm_sub_pd(x2,x3);
			t3 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
			t3 = _mm_mul_pd(tsp3,t3);

			y[iy] = _mm_add_pd(x1,t1);
			
			t4 = _mm_add_pd(t2,t3);
			t5 = _mm_shuffle_pd(t4,t4,_MM_SHUFFLE2(0,1));
			t6 = _mm_mul_pd(wr1,t4);
			t7 = _mm_mul_pd(wi1,t5);
			y[iy+tsd0] = _mm_addsub_pd(t6,t7);

			t4 = _mm_sub_pd(t2,t3);
			t5 = _mm_shuffle_pd(t4,t4,_MM_SHUFFLE2(0,1));
			t6 = _mm_mul_pd(wr2,t4);
			t7 = _mm_mul_pd(wi2,t5);
			y[iy+tsd0m2] = _mm_addsub_pd(t6,t7);

			ix = ix + 1;
			iy = iy + 1;
		}
	}
	return 0;
}

int fftrd4_(int *invk, int *inp, double *dw, double *dx, double *dy) {
	
	int nvk,np;
	int p,ivk;
	int iw,ix,iy;
	__m128d wr1,wi1,wr2,wi2,wr3,wi3,x1,x2,x3,x4;
	__m128d t1,t2,t3,t4,t5,t6,t7,t8;
	int tsd0, tsd1, tsd2;
	int tsd2m2,tsd2m3;
	int tsd0m2,tsd0m3;

	__m128d *w,*x,*y;

	nvk = *invk;
	np = *inp;
	w = (__m128d *)dw;
	x = (__m128d *)dx;
	y = (__m128d *)dy;

	tsd0 = nvk;
	tsd1 = nvk*4;
	tsd2 = nvk*np;

	tsd2m2 = tsd2*2;
	tsd2m3 = tsd2*3;

	tsd0m2 = tsd0*2;
	tsd0m3 = tsd0*3;

	ix = 0;
	iy = 0;
	iw = 0;
    for (p = 0; p < np; p++) {
		t1 = w[iw];
		wr1 = _mm_movedup_pd(t1);
		wi1 = _mm_unpackhi_pd(t1, t1);
		t1 = w[iw+1];
		wr2 = _mm_movedup_pd(t1);
		wi2 = _mm_unpackhi_pd(t1, t1);
		t1 = w[iw+2];
		wr3 = _mm_movedup_pd(t1);
		wi3 = _mm_unpackhi_pd(t1, t1);
		iw = iw + 3;

		ix = tsd0*p;
		iy = tsd1*p;
		for (ivk = 0; ivk < nvk; ivk++) {
			x1 = x[ix];
			x2 = x[ix+tsd2];
			x3 = x[ix+tsd2m2];
			x4 = x[ix+tsd2m3];

			t1 = _mm_add_pd(x1,x3);
			t2 = _mm_sub_pd(x1,x3);
			t3 = _mm_add_pd(x2,x4);

			t4 = _mm_sub_pd(x2,x4);
			t5 = _mm_setzero_pd();
			t4 = _mm_addsub_pd(t5,t4);
            t4 = _mm_shuffle_pd(t4,t4,_MM_SHUFFLE2(0,1));

			y[iy] = _mm_add_pd(t1,t3);

			t5 = _mm_add_pd(t2,t4);
			t6 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
			t7 = _mm_mul_pd(wr1,t5);
			t8 = _mm_mul_pd(wi1,t6);
			y[iy+tsd0] = _mm_addsub_pd(t7,t8);

			t5 = _mm_sub_pd(t1,t3);
			t6 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
			t7 = _mm_mul_pd(wr2,t5);
			t8 = _mm_mul_pd(wi2,t6);
			y[iy+tsd0m2] = _mm_addsub_pd(t7,t8);

			t5 = _mm_sub_pd(t2,t4);
			t6 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
			t7 = _mm_mul_pd(wr3,t5);
			t8 = _mm_mul_pd(wi3,t6);
			y[iy+tsd0m3] = _mm_addsub_pd(t7,t8);

			ix = ix + 1;
			iy = iy + 1;
		}
	}
	return 0;
}

int bftrd4_(int *invk, int *inp, double *dw, double *dx, double *dy) {
	
	int nvk,np;
	int p,ivk;
	int iw,ix,iy;
	__m128d wr1,wi1,wr2,wi2,wr3,wi3,x1,x2,x3,x4;
	__m128d t1,t2,t3,t4,t5,t6,t7,t8;
	int tsd0, tsd1, tsd2;
	int tsd2m2,tsd2m3;
	int tsd0m2,tsd0m3;

	__m128d *w,*x,*y;

	nvk = *invk;
	np = *inp;
	w = (__m128d *)dw;
	x = (__m128d *)dx;
	y = (__m128d *)dy;

	tsd0 = nvk;
	tsd1 = nvk*4;
	tsd2 = nvk*np;

	tsd2m2 = tsd2*2;
	tsd2m3 = tsd2*3;

	tsd0m2 = tsd0*2;
	tsd0m3 = tsd0*3;

	ix = 0;
	iy = 0;
	iw = 0;
    for (p = 0; p < np; p++) {
		t1 = w[iw];
		t2 = _mm_setzero_pd();
		wr1 = _mm_movedup_pd(t1);
		wi1 = _mm_unpackhi_pd(t1, t1);
		wi1 = _mm_sub_pd(t2,wi1);
		t1 = w[iw+1];
		t2 = _mm_setzero_pd();
		wr2 = _mm_movedup_pd(t1);
		wi2 = _mm_unpackhi_pd(t1, t1);
		wi2 = _mm_sub_pd(t2,wi2);
		t1 = w[iw+2];
		t2 = _mm_setzero_pd();
		wr3 = _mm_movedup_pd(t1);
		wi3 = _mm_unpackhi_pd(t1, t1);
		wi3 = _mm_sub_pd(t2,wi3);
		iw = iw + 3;

		ix = tsd0*p;
		iy = tsd1*p;
		for (ivk = 0; ivk < nvk; ivk++) {
			x1 = x[ix];
			x2 = x[ix+tsd2];
			x3 = x[ix+tsd2m2];
			x4 = x[ix+tsd2m3];

			t1 = _mm_add_pd(x1,x3);
			t2 = _mm_sub_pd(x1,x3);
			t3 = _mm_add_pd(x2,x4);

			t4 = _mm_sub_pd(x4,x2);
			t5 = _mm_setzero_pd();
			t4 = _mm_addsub_pd(t5,t4);
            t4 = _mm_shuffle_pd(t4,t4,_MM_SHUFFLE2(0,1));
			y[iy] = _mm_add_pd(t1,t3);

			t5 = _mm_add_pd(t2,t4);
			t6 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
			t7 = _mm_mul_pd(wr1,t5);
			t8 = _mm_mul_pd(wi1,t6);
			y[iy+tsd0] = _mm_addsub_pd(t7,t8);

			t5 = _mm_sub_pd(t1,t3);
			t6 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
			t7 = _mm_mul_pd(wr2,t5);
			t8 = _mm_mul_pd(wi2,t6);
			y[iy+tsd0m2] = _mm_addsub_pd(t7,t8);

			t5 = _mm_sub_pd(t2,t4);
			t6 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
			t7 = _mm_mul_pd(wr3,t5);
			t8 = _mm_mul_pd(wi3,t6);
			y[iy+tsd0m3] = _mm_addsub_pd(t7,t8);

			ix = ix + 1;
			iy = iy + 1;
		}
	}
	return 0;
}

int fftrd5_(int *invk, int *inp, double *dw, double *dx, double *dy) {
    const double sp25 = 0.951056516295153572;
    const double sq54 = 0.559016994374947424;
    const double ss = 0.618033988749894848;

	int nvk,np;
	int p,ivk;
	int iw,ix,iy;
	__m128d w1,w2,w3,w4,wi,wr,x1,x2,x3,x4,x5;
	__m128d t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13;
	int tsd0, tsd1, tsd2;
	int tsd2m2,tsd2m3,tsd2m4;
	int tsd0m2,tsd0m3,tsd0m4;

	__m128d *w,*x,*y;

	nvk = *invk;
	np = *inp;
	w = (__m128d *)dw;
	x = (__m128d *)dx;
	y = (__m128d *)dy;

	tsd0 = nvk;
	tsd1 = nvk*5;
	tsd2 = nvk*np;

	tsd2m2 = tsd2*2;
	tsd2m3 = tsd2*3;
	tsd2m4 = tsd2*4;

	tsd0m2 = tsd0*2;
	tsd0m3 = tsd0*3;
	tsd0m4 = tsd0*4;

	ix = 0;
	iy = 0;
	for (ivk = 0; ivk < nvk; ivk++) {
		x1 = x[ix];
		x2 = x[ix+tsd2];
		x3 = x[ix+tsd2m2];
		x4 = x[ix+tsd2m3];
		x5 = x[ix+tsd2m4];

		t1 = _mm_add_pd(x2,x5);
		t2 = _mm_add_pd(x3,x4);

		t5 = _mm_set1_pd(sp25);
		t3 = _mm_sub_pd(x2,x5);
		t4 = _mm_sub_pd(x3,x4);
		t3 = _mm_mul_pd(t5,t3);
		t4 = _mm_mul_pd(t5,t4);

		t5 = _mm_add_pd(t1,t2);
		t6 = _mm_sub_pd(t1,t2);
		t7 = _mm_set1_pd(sq54);
		t6 = _mm_mul_pd(t7,t6);

		t7 = _mm_set1_pd(0.25);
		t7 = _mm_mul_pd(t5,t7);
		t7 = _mm_sub_pd(x1,t7);

		t8 = _mm_add_pd(t7,t6);
		t9 = _mm_sub_pd(t7,t6);

		t10 = _mm_setzero_pd();
		t12 = _mm_addsub_pd(t10,t3);
		t13 = _mm_addsub_pd(t10,t4);
		t12 = _mm_shuffle_pd(t12,t12,_MM_SHUFFLE2(0,1));
		t13 = _mm_shuffle_pd(t13,t13,_MM_SHUFFLE2(0,1));
		t10 = _mm_set1_pd(ss);
		t10 = _mm_mul_pd(t13,t10);
		t10 = _mm_add_pd(t12,t10);
		t11 = _mm_set1_pd(ss);
		t11 = _mm_mul_pd(t12,t11);
		t11 = _mm_sub_pd(t11,t13);

		y[iy] = _mm_add_pd(x1,t5);

		y[iy+tsd0] = _mm_add_pd(t8,t10);

		y[iy+tsd0m2] = _mm_add_pd(t9,t11);

		y[iy+tsd0m3] = _mm_sub_pd(t9,t11);

		y[iy+tsd0m4] = _mm_sub_pd(t8,t10);

		ix = ix + 1;
		iy = iy + 1;
	}

	ix = 0;
	iy = 0;
	iw = 4;
    for (p = 1; p < np; p++) {
		w1 = w[iw];
		w2 = w[iw+1];
		w3 = w[iw+2];
		w4 = w[iw+3];
		iw = iw + 4;

		ix = tsd0*p;
		iy = tsd1*p;

		for (ivk = 0; ivk < nvk; ivk++) {
			x1 = x[ix];
			x2 = x[ix+tsd2];
			x3 = x[ix+tsd2m2];
			x4 = x[ix+tsd2m3];
			x5 = x[ix+tsd2m4];

			t1 = _mm_add_pd(x2,x5);
			t2 = _mm_add_pd(x3,x4);

			t5 = _mm_set1_pd(sp25);
			t3 = _mm_sub_pd(x2,x5);
			t4 = _mm_sub_pd(x3,x4);
			t3 = _mm_mul_pd(t5,t3);
			t4 = _mm_mul_pd(t5,t4);

			t5 = _mm_add_pd(t1,t2);
			t6 = _mm_sub_pd(t1,t2);
			t7 = _mm_set1_pd(sq54);
			t6 = _mm_mul_pd(t7,t6);

			t7 = _mm_set1_pd(0.25);
			t7 = _mm_mul_pd(t5,t7);
			t7 = _mm_sub_pd(x1,t7);

			t8 = _mm_add_pd(t7,t6);
			t9 = _mm_sub_pd(t7,t6);

			t10 = _mm_setzero_pd();
			t12 = _mm_addsub_pd(t10,t3);
			t13 = _mm_addsub_pd(t10,t4);
			t12 = _mm_shuffle_pd(t12,t12,_MM_SHUFFLE2(0,1));
			t13 = _mm_shuffle_pd(t13,t13,_MM_SHUFFLE2(0,1));
			t10 = _mm_set1_pd(ss);
			t10 = _mm_mul_pd(t13,t10);
			t10 = _mm_add_pd(t12,t10);
			t11 = _mm_set1_pd(ss);
			t11 = _mm_mul_pd(t12,t11);
			t11 = _mm_sub_pd(t11,t13);

			y[iy] = _mm_add_pd(x1,t5);

			t12 = _mm_add_pd(t8,t10);
			t13 = _mm_shuffle_pd(t12,t12,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w1);
			wi = _mm_unpackhi_pd(w1,w1);
			t12 = _mm_mul_pd(wr,t12);
			t13 = _mm_mul_pd(wi,t13);
			y[iy+tsd0] = _mm_addsub_pd(t12,t13);

			t12 = _mm_add_pd(t9,t11);
			t13 = _mm_shuffle_pd(t12,t12,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w2);
			wi = _mm_unpackhi_pd(w2,w2);
			t12 = _mm_mul_pd(wr,t12);
			t13 = _mm_mul_pd(wi,t13);
			y[iy+tsd0m2] = _mm_addsub_pd(t12,t13);

			t12 = _mm_sub_pd(t9,t11);
			t13 = _mm_shuffle_pd(t12,t12,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w3);
			wi = _mm_unpackhi_pd(w3,w3);
			t12 = _mm_mul_pd(wr,t12);
			t13 = _mm_mul_pd(wi,t13);
			y[iy+tsd0m3] = _mm_addsub_pd(t12,t13);

			t12 = _mm_sub_pd(t8,t10);
			t13 = _mm_shuffle_pd(t12,t12,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w4);
			wi = _mm_unpackhi_pd(w4,w4);
			t12 = _mm_mul_pd(wr,t12);
			t13 = _mm_mul_pd(wi,t13);
			y[iy+tsd0m4] = _mm_addsub_pd(t12,t13);

			ix = ix + 1;
			iy = iy + 1;
		}
	}
	return 0;
}

int bftrd5_(int *invk, int *inp, double *dw, double *dx, double *dy) {
    const double sp25 = 0.951056516295153572;
    const double sq54 = 0.559016994374947424;
    const double ss = 0.618033988749894848;

	int nvk,np;
	int p,ivk;
	int iw,ix,iy;
	__m128d w1,w2,w3,w4,wi,wr,x1,x2,x3,x4,x5;
	__m128d t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13;
	int tsd0, tsd1, tsd2;
	int tsd2m2,tsd2m3,tsd2m4;
	int tsd0m2,tsd0m3,tsd0m4;

	__m128d *w,*x,*y;

	nvk = *invk;
	np = *inp;
	w = (__m128d *)dw;
	x = (__m128d *)dx;
	y = (__m128d *)dy;

	tsd0 = nvk;
	tsd1 = nvk*5;
	tsd2 = nvk*np;

	tsd2m2 = tsd2*2;
	tsd2m3 = tsd2*3;
	tsd2m4 = tsd2*4;

	tsd0m2 = tsd0*2;
	tsd0m3 = tsd0*3;
	tsd0m4 = tsd0*4;

	ix = 0;
	iy = 0;
	for (ivk = 0; ivk < nvk; ivk++) {
		x1 = x[ix];
		x2 = x[ix+tsd2];
		x3 = x[ix+tsd2m2];
		x4 = x[ix+tsd2m3];
		x5 = x[ix+tsd2m4];

		t1 = _mm_add_pd(x2,x5);
		t2 = _mm_add_pd(x3,x4);

		t5 = _mm_set1_pd(sp25);
		t3 = _mm_sub_pd(x2,x5);
		t4 = _mm_sub_pd(x3,x4);
		t3 = _mm_mul_pd(t5,t3);
		t4 = _mm_mul_pd(t5,t4);

		t5 = _mm_add_pd(t1,t2);
		t6 = _mm_sub_pd(t1,t2);
		t7 = _mm_set1_pd(sq54);
		t6 = _mm_mul_pd(t7,t6);

		t7 = _mm_set1_pd(0.25);
		t7 = _mm_mul_pd(t5,t7);
		t7 = _mm_sub_pd(x1,t7);

		t8 = _mm_add_pd(t7,t6);
		t9 = _mm_sub_pd(t7,t6);

		t10 = _mm_setzero_pd();
		t12 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
		t13 = _mm_shuffle_pd(t4,t4,_MM_SHUFFLE2(0,1));
		t12 = _mm_addsub_pd(t10,t12);
		t13 = _mm_addsub_pd(t10,t13);
		t10 = _mm_set1_pd(ss);
		t10 = _mm_mul_pd(t13,t10);
		t10 = _mm_add_pd(t12,t10);
		t11 = _mm_set1_pd(ss);
		t11 = _mm_mul_pd(t12,t11);
		t11 = _mm_sub_pd(t11,t13);

		y[iy] = _mm_add_pd(x1,t5);

		y[iy+tsd0] = _mm_add_pd(t8,t10);

		y[iy+tsd0m2] = _mm_add_pd(t9,t11);

		y[iy+tsd0m3] = _mm_sub_pd(t9,t11);

		y[iy+tsd0m4] = _mm_sub_pd(t8,t10);

		ix = ix + 1;
		iy = iy + 1;
	}

	ix = 0;
	iy = 0;
	iw = 4;
    for (p = 1; p < np; p++) {
		t1 = _mm_set_pd(-0.0,0.0);
		w1 = w[iw];
		w1 = _mm_xor_pd(w1,t1);
		w2 = w[iw+1];
		w2 = _mm_xor_pd(w2,t1);
		w3 = w[iw+2];
		w3 = _mm_xor_pd(w3,t1);
		w4 = w[iw+3];
		w4 = _mm_xor_pd(w4,t1);
		iw = iw + 4;

		ix = tsd0*p;
		iy = tsd1*p;

		for (ivk = 0; ivk < nvk; ivk++) {
			x1 = x[ix];
			x2 = x[ix+tsd2];
			x3 = x[ix+tsd2m2];
			x4 = x[ix+tsd2m3];
			x5 = x[ix+tsd2m4];

			t1 = _mm_add_pd(x2,x5);
			t2 = _mm_add_pd(x3,x4);

			t5 = _mm_set1_pd(sp25);
			t3 = _mm_sub_pd(x2,x5);
			t4 = _mm_sub_pd(x3,x4);
			t3 = _mm_mul_pd(t5,t3);
			t4 = _mm_mul_pd(t5,t4);

			t5 = _mm_add_pd(t1,t2);
			t6 = _mm_sub_pd(t1,t2);
			t7 = _mm_set1_pd(sq54);
			t6 = _mm_mul_pd(t7,t6);

			t7 = _mm_set1_pd(0.25);
			t7 = _mm_mul_pd(t5,t7);
			t7 = _mm_sub_pd(x1,t7);

			t8 = _mm_add_pd(t7,t6);
			t9 = _mm_sub_pd(t7,t6);

			t10 = _mm_setzero_pd();
			t12 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
			t13 = _mm_shuffle_pd(t4,t4,_MM_SHUFFLE2(0,1));
			t12 = _mm_addsub_pd(t10,t12);
			t13 = _mm_addsub_pd(t10,t13);
			t10 = _mm_set1_pd(ss);
			t10 = _mm_mul_pd(t13,t10);
			t10 = _mm_add_pd(t12,t10);
			t11 = _mm_set1_pd(ss);
			t11 = _mm_mul_pd(t12,t11);
			t11 = _mm_sub_pd(t11,t13);

			y[iy] = _mm_add_pd(x1,t5);

			t12 = _mm_add_pd(t8,t10);
			t13 = _mm_shuffle_pd(t12,t12,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w1);
			wi = _mm_unpackhi_pd(w1,w1);
			t12 = _mm_mul_pd(wr,t12);
			t13 = _mm_mul_pd(wi,t13);
			y[iy+tsd0] = _mm_addsub_pd(t12,t13);

			t12 = _mm_add_pd(t9,t11);
			t13 = _mm_shuffle_pd(t12,t12,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w2);
			wi = _mm_unpackhi_pd(w2,w2);
			t12 = _mm_mul_pd(wr,t12);
			t13 = _mm_mul_pd(wi,t13);
			y[iy+tsd0m2] = _mm_addsub_pd(t12,t13);

			t12 = _mm_sub_pd(t9,t11);
			t13 = _mm_shuffle_pd(t12,t12,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w3);
			wi = _mm_unpackhi_pd(w3,w3);
			t12 = _mm_mul_pd(wr,t12);
			t13 = _mm_mul_pd(wi,t13);
			y[iy+tsd0m3] = _mm_addsub_pd(t12,t13);

			t12 = _mm_sub_pd(t8,t10);
			t13 = _mm_shuffle_pd(t12,t12,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w4);
			wi = _mm_unpackhi_pd(w4,w4);
			t12 = _mm_mul_pd(wr,t12);
			t13 = _mm_mul_pd(wi,t13);
			y[iy+tsd0m4] = _mm_addsub_pd(t12,t13);

			ix = ix + 1;
			iy = iy + 1;
		}
	}
	return 0;
}

int fftrd6_(int *invk, int *inp, double *dw, double *dx, double *dy) {
    const double sp3 = 0.866025403784438647;

	int nvk,np;
	int p,ivk;
	int iw,ix,iy;
	__m128d w1,w2,w3,w4,w5,wi,wr,x1,x2,x3,x4,x5,x6;
	__m128d t1,t2,t3,t4,u1,u2,u3,v1,v2,v3;
	int tsd0, tsd1, tsd2;
	int tsd2m2,tsd2m3,tsd2m4,tsd2m5;
	int tsd0m2,tsd0m3,tsd0m4,tsd0m5;

	__m128d *w,*x,*y;

	nvk = *invk;
	np = *inp;
	w = (__m128d *)dw;
	x = (__m128d *)dx;
	y = (__m128d *)dy;

	tsd0 = nvk;
	tsd1 = nvk*6;
	tsd2 = nvk*np;

	tsd2m2 = tsd2*2;
	tsd2m3 = tsd2*3;
	tsd2m4 = tsd2*4;
	tsd2m5 = tsd2*5;

	tsd0m2 = tsd0*2;
	tsd0m3 = tsd0*3;
	tsd0m4 = tsd0*4;
	tsd0m5 = tsd0*5;

	ix = 0;
	iy = 0;
	for (ivk = 0; ivk < nvk; ivk++) {
		x1 = x[ix];
		x2 = x[ix+tsd2];
		x3 = x[ix+tsd2m2];
		x4 = x[ix+tsd2m3];
		x5 = x[ix+tsd2m4];
		x6 = x[ix+tsd2m5];

		t1 = _mm_add_pd(x3,x5);
		t2 = _mm_set1_pd(0.5);
		t2 = _mm_mul_pd(t1,t2);
		t2 = _mm_sub_pd(x1,t2);
		t3 = _mm_sub_pd(x3,x5);
		t4 = _mm_set1_pd(-sp3);
		t3 = _mm_mul_pd(t4,t3);
		u1 = _mm_add_pd(x1,t1);
		t4 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
		u2 = _mm_addsub_pd(t2,t4);
		t4 = _mm_xor_pd(t4,_mm_set_pd(-0.0,-0.0));
		u3 = _mm_addsub_pd(t2,t4);

		t1 = _mm_add_pd(x6,x2);
		t2 = _mm_set1_pd(0.5);
		t2 = _mm_mul_pd(t1,t2);
		t2 = _mm_sub_pd(x4,t2);
		t4 = _mm_set1_pd(-sp3);
		t3 = _mm_sub_pd(x6,x2);
		t3 = _mm_mul_pd(t4,t3);
		v1 = _mm_add_pd(x4,t1);
		t4 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
		v2 = _mm_addsub_pd(t2,t4);
		t4 = _mm_xor_pd(t4,_mm_set_pd(-0.0,-0.0));
		v3 = _mm_addsub_pd(t2,t4);

		y[iy] = _mm_add_pd(u1,v1);

		y[iy+tsd0] = _mm_sub_pd(u2,v2);

		y[iy+tsd0m2] = _mm_add_pd(u3,v3);

		y[iy+tsd0m3] = _mm_sub_pd(u1,v1);

		y[iy+tsd0m4] = _mm_add_pd(u2,v2);

		y[iy+tsd0m5] = _mm_sub_pd(u3,v3);

		ix = ix + 1;
		iy = iy + 1;
	}

	ix = 0;
	iy = 0;
	iw = 5;
    for (p = 1; p < np; p++) {
		w1 = w[iw];
		w2 = w[iw+1];
		w3 = w[iw+2];
		w4 = w[iw+3];
		w5 = w[iw+4];
		iw = iw + 5;

		ix = tsd0*p;
		iy = tsd1*p;

		for (ivk = 0; ivk < nvk; ivk++) {
			x1 = x[ix];
			x2 = x[ix+tsd2];
			x3 = x[ix+tsd2m2];
			x4 = x[ix+tsd2m3];
			x5 = x[ix+tsd2m4];
			x6 = x[ix+tsd2m5];

			t1 = _mm_add_pd(x3,x5);
			t2 = _mm_set1_pd(0.5);
			t2 = _mm_mul_pd(t1,t2);
			t2 = _mm_sub_pd(x1,t2);
			t3 = _mm_sub_pd(x3,x5);
			t4 = _mm_set1_pd(-sp3);
			t3 = _mm_mul_pd(t4,t3);
			u1 = _mm_add_pd(x1,t1);
			t3 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
			u2 = _mm_addsub_pd(t2,t3);
			t3 = _mm_xor_pd(t3,_mm_set_pd(-0.0,-0.0));
			u3 = _mm_addsub_pd(t2,t3);

			t1 = _mm_add_pd(x6,x2);
			t2 = _mm_set1_pd(0.5);
			t2 = _mm_mul_pd(t1,t2);
			t2 = _mm_sub_pd(x4,t2);
			t4 = _mm_set1_pd(-sp3);
			t3 = _mm_sub_pd(x6,x2);
			t3 = _mm_mul_pd(t4,t3);
			v1 = _mm_add_pd(x4,t1);
			t3 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
			v2 = _mm_addsub_pd(t2,t3);
			t3 = _mm_xor_pd(t3,_mm_set_pd(-0.0,-0.0));
			v3 = _mm_addsub_pd(t2,t3);

			y[iy] = _mm_add_pd(u1,v1);

			t1 = _mm_sub_pd(u2,v2);
			t2 = _mm_shuffle_pd(t1,t1,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w1);
			wi = _mm_unpackhi_pd(w1,w1);
			t1 = _mm_mul_pd(wr,t1);
			t2 = _mm_mul_pd(wi,t2);
			y[iy+tsd0] = _mm_addsub_pd(t1,t2);

			t1 = _mm_add_pd(u3,v3);
			t2 = _mm_shuffle_pd(t1,t1,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w2);
			wi = _mm_unpackhi_pd(w2,w2);
			t1 = _mm_mul_pd(wr,t1);
			t2 = _mm_mul_pd(wi,t2);
			y[iy+tsd0m2] = _mm_addsub_pd(t1,t2);

			t1 = _mm_sub_pd(u1,v1);
			t2 = _mm_shuffle_pd(t1,t1,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w3);
			wi = _mm_unpackhi_pd(w3,w3);
			t1 = _mm_mul_pd(wr,t1);
			t2 = _mm_mul_pd(wi,t2);
			y[iy+tsd0m3] = _mm_addsub_pd(t1,t2);

			t1 = _mm_add_pd(u2,v2);
			t2 = _mm_shuffle_pd(t1,t1,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w4);
			wi = _mm_unpackhi_pd(w4,w4);
			t1 = _mm_mul_pd(wr,t1);
			t2 = _mm_mul_pd(wi,t2);
			y[iy+tsd0m4] = _mm_addsub_pd(t1,t2);

			t1 = _mm_sub_pd(u3,v3);
			t2 = _mm_shuffle_pd(t1,t1,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w5);
			wi = _mm_unpackhi_pd(w5,w5);
			t1 = _mm_mul_pd(wr,t1);
			t2 = _mm_mul_pd(wi,t2);
			y[iy+tsd0m5] = _mm_addsub_pd(t1,t2);

			ix = ix + 1;
			iy = iy + 1;
		}
	}
	return 0;
}

int bftrd6_(int *invk, int *inp, double *dw, double *dx, double *dy) {
    const double sp3 = 0.866025403784438647;

	int nvk,np;
	int p,ivk;
	int iw,ix,iy;
	__m128d w1,w2,w3,w4,w5,wi,wr,x1,x2,x3,x4,x5,x6;
	__m128d t1,t2,t3,t4,u1,u2,u3,v1,v2,v3;
	int tsd0, tsd1, tsd2;
	int tsd2m2,tsd2m3,tsd2m4,tsd2m5;
	int tsd0m2,tsd0m3,tsd0m4,tsd0m5;

	__m128d *w,*x,*y;

	nvk = *invk;
	np = *inp;
	w = (__m128d *)dw;
	x = (__m128d *)dx;
	y = (__m128d *)dy;

	tsd0 = nvk;
	tsd1 = nvk*6;
	tsd2 = nvk*np;

	tsd2m2 = tsd2*2;
	tsd2m3 = tsd2*3;
	tsd2m4 = tsd2*4;
	tsd2m5 = tsd2*5;

	tsd0m2 = tsd0*2;
	tsd0m3 = tsd0*3;
	tsd0m4 = tsd0*4;
	tsd0m5 = tsd0*5;

	ix = 0;
	iy = 0;
	for (ivk = 0; ivk < nvk; ivk++) {
		x1 = x[ix];
		x2 = x[ix+tsd2];
		x3 = x[ix+tsd2m2];
		x4 = x[ix+tsd2m3];
		x5 = x[ix+tsd2m4];
		x6 = x[ix+tsd2m5];

		t1 = _mm_add_pd(x3,x5);
		t2 = _mm_set1_pd(0.5);
		t2 = _mm_mul_pd(t1,t2);
		t2 = _mm_sub_pd(x1,t2);
		t3 = _mm_sub_pd(x3,x5);
		t4 = _mm_set1_pd(sp3);
		t3 = _mm_mul_pd(t4,t3);
		u1 = _mm_add_pd(x1,t1);
		t4 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
		u2 = _mm_addsub_pd(t2,t4);
		t4 = _mm_xor_pd(t4,_mm_set_pd(-0.0,-0.0));
		u3 = _mm_addsub_pd(t2,t4);

		t1 = _mm_add_pd(x6,x2);
		t2 = _mm_set1_pd(0.5);
		t2 = _mm_mul_pd(t1,t2);
		t2 = _mm_sub_pd(x4,t2);
		t4 = _mm_set1_pd(sp3);
		t3 = _mm_sub_pd(x6,x2);
		t3 = _mm_mul_pd(t4,t3);
		v1 = _mm_add_pd(x4,t1);
		t4 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
		v2 = _mm_addsub_pd(t2,t4);
		t4 = _mm_xor_pd(t4,_mm_set_pd(-0.0,-0.0));
		v3 = _mm_addsub_pd(t2,t4);

		y[iy] = _mm_add_pd(u1,v1);

		y[iy+tsd0] = _mm_sub_pd(u2,v2);

		y[iy+tsd0m2] = _mm_add_pd(u3,v3);

		y[iy+tsd0m3] = _mm_sub_pd(u1,v1);

		y[iy+tsd0m4] = _mm_add_pd(u2,v2);

		y[iy+tsd0m5] = _mm_sub_pd(u3,v3);

		ix = ix + 1;
		iy = iy + 1;
	}

	ix = 0;
	iy = 0;
	iw = 5;
    for (p = 1; p < np; p++) {
		t1 = _mm_set_pd(-0.0,0.0);
		w1 = w[iw];
		w1 = _mm_xor_pd(w1,t1);
		w2 = w[iw+1];
		w2 = _mm_xor_pd(w2,t1);
		w3 = w[iw+2];
		w3 = _mm_xor_pd(w3,t1);
		w4 = w[iw+3];
		w4 = _mm_xor_pd(w4,t1);
		w5 = w[iw+4];
		w5 = _mm_xor_pd(w5,t1);
		iw = iw + 5;

		ix = tsd0*p;
		iy = tsd1*p;

		for (ivk = 0; ivk < nvk; ivk++) {
			x1 = x[ix];
			x2 = x[ix+tsd2];
			x3 = x[ix+tsd2m2];
			x4 = x[ix+tsd2m3];
			x5 = x[ix+tsd2m4];
			x6 = x[ix+tsd2m5];

			t1 = _mm_add_pd(x3,x5);
			t2 = _mm_set1_pd(0.5);
			t2 = _mm_mul_pd(t1,t2);
			t2 = _mm_sub_pd(x1,t2);
			t3 = _mm_sub_pd(x3,x5);
			t4 = _mm_set1_pd(sp3);
			t3 = _mm_mul_pd(t4,t3);
			u1 = _mm_add_pd(x1,t1);
			t3 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
			u2 = _mm_addsub_pd(t2,t3);
			t3 = _mm_xor_pd(t3,_mm_set_pd(-0.0,-0.0));
			u3 = _mm_addsub_pd(t2,t3);

			t1 = _mm_add_pd(x6,x2);
			t2 = _mm_set1_pd(0.5);
			t2 = _mm_mul_pd(t1,t2);
			t2 = _mm_sub_pd(x4,t2);
			t4 = _mm_set1_pd(sp3);
			t3 = _mm_sub_pd(x6,x2);
			t3 = _mm_mul_pd(t4,t3);
			v1 = _mm_add_pd(x4,t1);
			t3 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
			v2 = _mm_addsub_pd(t2,t3);
			t3 = _mm_xor_pd(t3,_mm_set_pd(-0.0,-0.0));
			v3 = _mm_addsub_pd(t2,t3);

			y[iy] = _mm_add_pd(u1,v1);

			t1 = _mm_sub_pd(u2,v2);
			t2 = _mm_shuffle_pd(t1,t1,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w1);
			wi = _mm_unpackhi_pd(w1,w1);
			t1 = _mm_mul_pd(wr,t1);
			t2 = _mm_mul_pd(wi,t2);
			y[iy+tsd0] = _mm_addsub_pd(t1,t2);

			t1 = _mm_add_pd(u3,v3);
			t2 = _mm_shuffle_pd(t1,t1,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w2);
			wi = _mm_unpackhi_pd(w2,w2);
			t1 = _mm_mul_pd(wr,t1);
			t2 = _mm_mul_pd(wi,t2);
			y[iy+tsd0m2] = _mm_addsub_pd(t1,t2);

			t1 = _mm_sub_pd(u1,v1);
			t2 = _mm_shuffle_pd(t1,t1,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w3);
			wi = _mm_unpackhi_pd(w3,w3);
			t1 = _mm_mul_pd(wr,t1);
			t2 = _mm_mul_pd(wi,t2);
			y[iy+tsd0m3] = _mm_addsub_pd(t1,t2);

			t1 = _mm_add_pd(u2,v2);
			t2 = _mm_shuffle_pd(t1,t1,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w4);
			wi = _mm_unpackhi_pd(w4,w4);
			t1 = _mm_mul_pd(wr,t1);
			t2 = _mm_mul_pd(wi,t2);
			y[iy+tsd0m4] = _mm_addsub_pd(t1,t2);

			t1 = _mm_sub_pd(u3,v3);
			t2 = _mm_shuffle_pd(t1,t1,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w5);
			wi = _mm_unpackhi_pd(w5,w5);
			t1 = _mm_mul_pd(wr,t1);
			t2 = _mm_mul_pd(wi,t2);
			y[iy+tsd0m5] = _mm_addsub_pd(t1,t2);

			ix = ix + 1;
			iy = iy + 1;
		}
	}
	return 0;
}

int fftrd8_(int *invk, int *inp, double *dw, double *dx, double *dy) {
    const double sp2 = 0.70710678118654752440;

	int nvk,np;
	int p,ivk;
	int iw,ix,iy;
	__m128d w1,w2,w3,w4,w5,w6,w7,wi,wr,x1,x2,x3,x4,x5,x6,x7,x8;
	__m128d t1,t2,t3,t4,t5,u1,u2,u3,u4,v1,v2,v3,v4,s;
	int tsd0, tsd1, tsd2;
	int tsd2m2,tsd2m3,tsd2m4,tsd2m5,tsd2m6,tsd2m7;
	int tsd0m2,tsd0m3,tsd0m4,tsd0m5,tsd0m6,tsd0m7;

	__m128d *w,*x,*y;

	nvk = *invk;
	np = *inp;
	w = (__m128d *)dw;
	x = (__m128d *)dx;
	y = (__m128d *)dy;

	tsd0 = nvk;
	tsd1 = nvk*8;
	tsd2 = nvk*np;

	tsd2m2 = tsd2*2;
	tsd2m3 = tsd2*3;
	tsd2m4 = tsd2*4;
	tsd2m5 = tsd2*5;
	tsd2m6 = tsd2*6;
	tsd2m7 = tsd2*7;

	tsd0m2 = tsd0*2;
	tsd0m3 = tsd0*3;
	tsd0m4 = tsd0*4;
	tsd0m5 = tsd0*5;
	tsd0m6 = tsd0*6;
	tsd0m7 = tsd0*7;

	ix = 0;
	iy = 0;
	for (ivk = 0; ivk < nvk; ivk++) {
		x1 = x[ix];
		x2 = x[ix+tsd2];
		x3 = x[ix+tsd2m2];
		x4 = x[ix+tsd2m3];
		x5 = x[ix+tsd2m4];
		x6 = x[ix+tsd2m5];
		x7 = x[ix+tsd2m6];
		x8 = x[ix+tsd2m7];

		t1 = _mm_add_pd(x1,x5);
		t2 = _mm_sub_pd(x1,x5);
		t3 = _mm_add_pd(x3,x7);
		t4 = _mm_xor_pd(x3,_mm_set_pd(0.0,-0.0));
		t5 = _mm_xor_pd(x7,_mm_set_pd(-0.0,0.0));
		t5 = _mm_add_pd(t4,t5);
		t4 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
		u1 = _mm_add_pd(t1,t3);
		u2 = _mm_add_pd(t2,t4);
		u3 = _mm_sub_pd(t1,t3);
		u4 = _mm_sub_pd(t2,t4);

		t1 = _mm_add_pd(x2,x6);
		t2 = _mm_sub_pd(x2,x6);
		t3 = _mm_add_pd(x4,x8);
		t4 = _mm_xor_pd(x4,_mm_set_pd(0.0,-0.0));
		t5 = _mm_xor_pd(x8,_mm_set_pd(-0.0,0.0));
		t5 = _mm_add_pd(t4,t5);
		t4 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
		v1 = _mm_add_pd(t1,t3);
		v2 = _mm_add_pd(t2,t4);
		v3 = _mm_sub_pd(t1,t3);
		v4 = _mm_sub_pd(t2,t4);

		y[iy] = _mm_add_pd(u1,v1);
		y[iy+tsd0m4] = _mm_sub_pd(u1,v1);

		t5 = _mm_set1_pd(sp2);
		s = _mm_shuffle_pd(v2,v2,_MM_SHUFFLE2(0,1));
		s = _mm_xor_pd(s,_mm_set_pd(-0.0,0.0));
		s = _mm_add_pd(v2,s);
		s = _mm_mul_pd(t5,s);

		y[iy+tsd0] =_mm_add_pd(u2,s);

		y[iy+tsd0m5] = _mm_sub_pd(u2,s);

		t2 = _mm_shuffle_pd(v3,v3,_MM_SHUFFLE2(0,1));
		t1 = _mm_xor_pd(t2,_mm_set_pd(-0.0,-0.0));
		y[iy+tsd0m2] = _mm_addsub_pd(u3,t1);

		y[iy+tsd0m6] = _mm_addsub_pd(u3,t2);

		t5 = _mm_set1_pd(sp2);
		s = _mm_shuffle_pd(v4,v4,_MM_SHUFFLE2(0,1));
		s = _mm_xor_pd(s,_mm_set_pd(-0.0,0.0));
		s = _mm_sub_pd(s,v4);
		s = _mm_mul_pd(t5,s);

		y[iy+tsd0m3] = _mm_add_pd(u4,s);

		y[iy+tsd0m7] = _mm_sub_pd(u4,s);

		ix = ix + 1;
		iy = iy + 1;
	}

	ix = 0;
	iy = 0;
	iw = 7;
    for (p = 1; p < np; p++) {
		w1 = w[iw];
		w2 = w[iw+1];
		w3 = w[iw+2];
		w4 = w[iw+3];
		w5 = w[iw+4];
		w6 = w[iw+5];
		w7 = w[iw+6];
		iw = iw + 7;

		ix = tsd0*p;
		iy = tsd1*p;

		for (ivk = 0; ivk < nvk; ivk++) {
			x1 = x[ix];
			x2 = x[ix+tsd2];
			x3 = x[ix+tsd2m2];
			x4 = x[ix+tsd2m3];
			x5 = x[ix+tsd2m4];
			x6 = x[ix+tsd2m5];
			x7 = x[ix+tsd2m6];
			x8 = x[ix+tsd2m7];

			t1 = _mm_add_pd(x1,x5);
			t2 = _mm_sub_pd(x1,x5);
			t3 = _mm_add_pd(x3,x7);
			t4 = _mm_xor_pd(x3,_mm_set_pd(0.0,-0.0));
			t5 = _mm_xor_pd(x7,_mm_set_pd(-0.0,0.0));
			t5 = _mm_add_pd(t4,t5);
			t4 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
			u1 = _mm_add_pd(t1,t3);
			u2 = _mm_add_pd(t2,t4);
			u3 = _mm_sub_pd(t1,t3);
			u4 = _mm_sub_pd(t2,t4);

			t1 = _mm_add_pd(x2,x6);
			t2 = _mm_sub_pd(x2,x6);
			t3 = _mm_add_pd(x4,x8);
			t4 = _mm_xor_pd(x4,_mm_set_pd(0.0,-0.0));
			t5 = _mm_xor_pd(x8,_mm_set_pd(-0.0,0.0));
			t5 = _mm_add_pd(t4,t5);
			t4 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
			v1 = _mm_add_pd(t1,t3);
			v2 = _mm_add_pd(t2,t4);
			v3 = _mm_sub_pd(t1,t3);
			v4 = _mm_sub_pd(t2,t4);

			y[iy] = _mm_add_pd(u1,v1);

			t3 = _mm_sub_pd(u1,v1);
			t4 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w4);
			wi = _mm_unpackhi_pd(w4,w4);
			t3 = _mm_mul_pd(wr,t3);
			t4 = _mm_mul_pd(wi,t4);
			y[iy+tsd0m4] = _mm_addsub_pd(t3,t4);

			t5 = _mm_set1_pd(sp2);
			s = _mm_shuffle_pd(v2,v2,_MM_SHUFFLE2(0,1));
			s = _mm_xor_pd(s,_mm_set_pd(-0.0,0.0));
			s = _mm_add_pd(v2,s);
			s = _mm_mul_pd(t5,s);

			t3 = _mm_add_pd(u2,s);
			t4 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w1);
			wi = _mm_unpackhi_pd(w1,w1);
			t3 = _mm_mul_pd(wr,t3);
			t4 = _mm_mul_pd(wi,t4);
			y[iy+tsd0] = _mm_addsub_pd(t3,t4);

			t3 = _mm_sub_pd(u2,s);
			t4 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w5);
			wi = _mm_unpackhi_pd(w5,w5);
			t3 = _mm_mul_pd(wr,t3);
			t4 = _mm_mul_pd(wi,t4);
			y[iy+tsd0m5] = _mm_addsub_pd(t3,t4);

			t2 = _mm_shuffle_pd(v3,v3,_MM_SHUFFLE2(0,1));
			t1 = _mm_xor_pd(t2,_mm_set_pd(-0.0,-0.0));

			t3 = _mm_addsub_pd(u3,t1);
			t4 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w2);
			wi = _mm_unpackhi_pd(w2,w2);
			t3 = _mm_mul_pd(wr,t3);
			t4 = _mm_mul_pd(wi,t4);
			y[iy+tsd0m2] = _mm_addsub_pd(t3,t4);

			t3 = _mm_addsub_pd(u3,t2);
			t4 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w6);
			wi = _mm_unpackhi_pd(w6,w6);
			t3 = _mm_mul_pd(wr,t3);
			t4 = _mm_mul_pd(wi,t4);
			y[iy+tsd0m6] = _mm_addsub_pd(t3,t4);

			t5 = _mm_set1_pd(sp2);
			s = _mm_shuffle_pd(v4,v4,_MM_SHUFFLE2(0,1));
			s = _mm_xor_pd(s,_mm_set_pd(-0.0,0.0));
			s = _mm_sub_pd(s,v4);
			s = _mm_mul_pd(t5,s);

			t3 = _mm_add_pd(u4,s);
			t4 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w3);
			wi = _mm_unpackhi_pd(w3,w3);
			t3 = _mm_mul_pd(wr,t3);
			t4 = _mm_mul_pd(wi,t4);
			y[iy+tsd0m3] = _mm_addsub_pd(t3,t4);

			t3 = _mm_sub_pd(u4,s);
			t4 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w7);
			wi = _mm_unpackhi_pd(w7,w7);
			t3 = _mm_mul_pd(wr,t3);
			t4 = _mm_mul_pd(wi,t4);
			y[iy+tsd0m7] = _mm_addsub_pd(t3,t4);

			ix = ix + 1;
			iy = iy + 1;
		}
	}
	return 0;
}

int bftrd8_(int *invk, int *inp, double *dw, double *dx, double *dy) {
    const double sp2 = 0.70710678118654752440;

	int nvk,np;
	int p,ivk;
	int iw,ix,iy;
	__m128d w1,w2,w3,w4,w5,w6,w7,wi,wr,x1,x2,x3,x4,x5,x6,x7,x8;
	__m128d t1,t2,t3,t4,t5,u1,u2,u3,u4,v1,v2,v3,v4,s;
	int tsd0, tsd1, tsd2;
	int tsd2m2,tsd2m3,tsd2m4,tsd2m5,tsd2m6,tsd2m7;
	int tsd0m2,tsd0m3,tsd0m4,tsd0m5,tsd0m6,tsd0m7;

	__m128d *w,*x,*y;

	nvk = *invk;
	np = *inp;
	w = (__m128d *)dw;
	x = (__m128d *)dx;
	y = (__m128d *)dy;

	tsd0 = nvk;
	tsd1 = nvk*8;
	tsd2 = nvk*np;

	tsd2m2 = tsd2*2;
	tsd2m3 = tsd2*3;
	tsd2m4 = tsd2*4;
	tsd2m5 = tsd2*5;
	tsd2m6 = tsd2*6;
	tsd2m7 = tsd2*7;

	tsd0m2 = tsd0*2;
	tsd0m3 = tsd0*3;
	tsd0m4 = tsd0*4;
	tsd0m5 = tsd0*5;
	tsd0m6 = tsd0*6;
	tsd0m7 = tsd0*7;

	ix = 0;
	iy = 0;
	for (ivk = 0; ivk < nvk; ivk++) {
		x1 = x[ix];
		x2 = x[ix+tsd2];
		x3 = x[ix+tsd2m2];
		x4 = x[ix+tsd2m3];
		x5 = x[ix+tsd2m4];
		x6 = x[ix+tsd2m5];
		x7 = x[ix+tsd2m6];
		x8 = x[ix+tsd2m7];

		t1 = _mm_add_pd(x1,x5);
		t2 = _mm_sub_pd(x1,x5);
		t3 = _mm_add_pd(x3,x7);
		t4 = _mm_xor_pd(x3,_mm_set_pd(-0.0,0.0));
		t5 = _mm_xor_pd(x7,_mm_set_pd(0.0,-0.0));
		t5 = _mm_add_pd(t4,t5);
		t4 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
		u1 = _mm_add_pd(t1,t3);
		u2 = _mm_add_pd(t2,t4);
		u3 = _mm_sub_pd(t1,t3);
		u4 = _mm_sub_pd(t2,t4);

		t1 = _mm_add_pd(x2,x6);
		t2 = _mm_sub_pd(x2,x6);
		t3 = _mm_add_pd(x4,x8);
		t4 = _mm_xor_pd(x4,_mm_set_pd(-0.0,0.0));
		t5 = _mm_xor_pd(x8,_mm_set_pd(0.0,-0.0));
		t5 = _mm_add_pd(t4,t5);
		t4 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
		v1 = _mm_add_pd(t1,t3);
		v2 = _mm_add_pd(t2,t4);
		v3 = _mm_sub_pd(t1,t3);
		v4 = _mm_sub_pd(t2,t4);

		y[iy] = _mm_add_pd(u1,v1);
		y[iy+tsd0m4] = _mm_sub_pd(u1,v1);

		t5 = _mm_set1_pd(sp2);
		s = _mm_shuffle_pd(v2,v2,_MM_SHUFFLE2(0,1));
		s = _mm_addsub_pd(v2,s);
		s = _mm_mul_pd(t5,s);

		y[iy+tsd0] = _mm_add_pd(u2,s);

		y[iy+tsd0m5] = _mm_sub_pd(u2,s);

		t1 = _mm_shuffle_pd(v3,v3,_MM_SHUFFLE2(0,1));
		t2 = _mm_xor_pd(t1,_mm_set_pd(-0.0,-0.0));
		y[iy+tsd0m2] = _mm_addsub_pd(u3,t1);

		y[iy+tsd0m6] = _mm_addsub_pd(u3,t2);

		t5 = _mm_set1_pd(sp2);
		s = _mm_shuffle_pd(v4,v4,_MM_SHUFFLE2(0,1));
		s = _mm_xor_pd(s,_mm_set_pd(0.0,-0.0));
		s = _mm_sub_pd(s,v4);
		s = _mm_mul_pd(t5,s);

		y[iy+tsd0m3] = _mm_add_pd(u4,s);

		y[iy+tsd0m7] = _mm_sub_pd(u4,s);

		ix = ix + 1;
		iy = iy + 1;
	}

	ix = 0;
	iy = 0;
	iw = 7;
    for (p = 1; p < np; p++) {
		t1 = _mm_set_pd(-0.0,0.0);
		w1 = w[iw];
		w1 = _mm_xor_pd(w1,t1);
		w2 = w[iw+1];
		w2 = _mm_xor_pd(w2,t1);
		w3 = w[iw+2];
		w3 = _mm_xor_pd(w3,t1);
		w4 = w[iw+3];
		w4 = _mm_xor_pd(w4,t1);
		w5 = w[iw+4];
		w5 = _mm_xor_pd(w5,t1);
		w6 = w[iw+5];
		w6 = _mm_xor_pd(w6,t1);
		w7 = w[iw+6];
		w7 = _mm_xor_pd(w7,t1);
		iw = iw + 7;

		ix = tsd0*p;
		iy = tsd1*p;

		for (ivk = 0; ivk < nvk; ivk++) {
			x1 = x[ix];
			x2 = x[ix+tsd2];
			x3 = x[ix+tsd2m2];
			x4 = x[ix+tsd2m3];
			x5 = x[ix+tsd2m4];
			x6 = x[ix+tsd2m5];
			x7 = x[ix+tsd2m6];
			x8 = x[ix+tsd2m7];

			t1 = _mm_add_pd(x1,x5);
			t2 = _mm_sub_pd(x1,x5);
			t3 = _mm_add_pd(x3,x7);
			t4 = _mm_xor_pd(x3,_mm_set_pd(-0.0,0.0));
			t5 = _mm_xor_pd(x7,_mm_set_pd(0.0,-0.0));
			t5 = _mm_add_pd(t4,t5);
			t4 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
			u1 = _mm_add_pd(t1,t3);
			u2 = _mm_add_pd(t2,t4);
			u3 = _mm_sub_pd(t1,t3);
			u4 = _mm_sub_pd(t2,t4);

			t1 = _mm_add_pd(x2,x6);
			t2 = _mm_sub_pd(x2,x6);
			t3 = _mm_add_pd(x4,x8);
			t4 = _mm_xor_pd(x4,_mm_set_pd(-0.0,0.0));
			t5 = _mm_xor_pd(x8,_mm_set_pd(0.0,-0.0));
			t5 = _mm_add_pd(t4,t5);
			t4 = _mm_shuffle_pd(t5,t5,_MM_SHUFFLE2(0,1));
			v1 = _mm_add_pd(t1,t3);
			v2 = _mm_add_pd(t2,t4);
			v3 = _mm_sub_pd(t1,t3);
			v4 = _mm_sub_pd(t2,t4);

			y[iy] = _mm_add_pd(u1,v1);

			t3 = _mm_sub_pd(u1,v1);
			t4 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w4);
			wi = _mm_unpackhi_pd(w4,w4);
			t3 = _mm_mul_pd(wr,t3);
			t4 = _mm_mul_pd(wi,t4);
			y[iy+tsd0m4] = _mm_addsub_pd(t3,t4);

			t5 = _mm_set1_pd(sp2);
			s = _mm_shuffle_pd(v2,v2,_MM_SHUFFLE2(0,1));
			s = _mm_addsub_pd(v2,s);
			s = _mm_mul_pd(t5,s);

			t3 = _mm_add_pd(u2,s);
			t4 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w1);
			wi = _mm_unpackhi_pd(w1,w1);
			t3 = _mm_mul_pd(wr,t3);
			t4 = _mm_mul_pd(wi,t4);
			y[iy+tsd0] = _mm_addsub_pd(t3,t4);

			t3 = _mm_sub_pd(u2,s);
			t4 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w5);
			wi = _mm_unpackhi_pd(w5,w5);
			t3 = _mm_mul_pd(wr,t3);
			t4 = _mm_mul_pd(wi,t4);
			y[iy+tsd0m5] = _mm_addsub_pd(t3,t4);

			t1 = _mm_shuffle_pd(v3,v3,_MM_SHUFFLE2(0,1));
			t2 = _mm_xor_pd(t1,_mm_set_pd(-0.0,-0.0));

			t3 = _mm_addsub_pd(u3,t1);
			t4 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w2);
			wi = _mm_unpackhi_pd(w2,w2);
			t3 = _mm_mul_pd(wr,t3);
			t4 = _mm_mul_pd(wi,t4);
			y[iy+tsd0m2] = _mm_addsub_pd(t3,t4);

			t3 = _mm_addsub_pd(u3,t2);
			t4 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w6);
			wi = _mm_unpackhi_pd(w6,w6);
			t3 = _mm_mul_pd(wr,t3);
			t4 = _mm_mul_pd(wi,t4);
			y[iy+tsd0m6] = _mm_addsub_pd(t3,t4);

			t5 = _mm_set1_pd(sp2);
			s = _mm_shuffle_pd(v4,v4,_MM_SHUFFLE2(0,1));
			s = _mm_xor_pd(s,_mm_set_pd(0.0,-0.0));
			s = _mm_sub_pd(s,v4);
			s = _mm_mul_pd(t5,s);

			t3 = _mm_add_pd(u4,s);
			t4 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w3);
			wi = _mm_unpackhi_pd(w3,w3);
			t3 = _mm_mul_pd(wr,t3);
			t4 = _mm_mul_pd(wi,t4);
			y[iy+tsd0m3] = _mm_addsub_pd(t3,t4);

			t3 = _mm_sub_pd(u4,s);
			t4 = _mm_shuffle_pd(t3,t3,_MM_SHUFFLE2(0,1));
			wr = _mm_movedup_pd(w7);
			wi = _mm_unpackhi_pd(w7,w7);
			t3 = _mm_mul_pd(wr,t3);
			t4 = _mm_mul_pd(wi,t4);
			y[iy+tsd0m7] = _mm_addsub_pd(t3,t4);

			ix = ix + 1;
			iy = iy + 1;
		}
	}
	return 0;
}

int fft_tpb_(int *ilxy, int *ibxy, int *iexy, int *ilz,
			 double *dsrc, double *ddst) {
	int lxy,bxy,exy,lz;
	__m128d *src, *dst;

	int i,j,is,id,pxy;
	__m128d t1,t2;
	int tsd1, tsd2, tbs;

	const int cpfr = 900;
	const int cpf = 450;
	int ipf,ipfr,r;

	lxy = *ilxy;
	bxy = *ibxy;
	exy = *iexy;
	lz = *ilz;

	src = (__m128d *)dsrc;
	dst = (__m128d *)ddst;

	ipf = cpf/lz;
	ipf = (ipf>2) ? ipf : 2;

	ipfr = cpfr/lz;
	ipfr = (ipfr>1) ? ipfr : 1;

	pxy = exy-bxy+1;

	r = pxy % 2;

	tsd1 = lxy;
	tsd2 = lz;
	tbs = bxy-1;

	if (r == 1) {
		is = tbs;
		id = 0;

		for (j = 0; j < lz; j++) {
			_mm_prefetch((char *)(&src[is+ipfr]),_MM_HINT_T0);

			t1 = src[is];
			is = is + tsd1;
			dst[id] = t1;
			id = id + 1;
		}
	}
	for (i = r; i < pxy; i = i+2) {
		is = i+tbs;
		id = tsd2*i;

		for (j = 0; j < lz; j++) {
			_mm_prefetch((char *)(&src[is+ipf]),_MM_HINT_T0);

			t1 = src[is];
			t2 = src[is+1];
			is = is + tsd1;
			dst[id] = t1;
			dst[id+tsd2] = t2;
			id = id + 1;
		}
	}
    return 0;
}

int fft_tpbs_(int *ilxy, int *ibxy, int *iexy, int *ilz,
			  double *dsrc, double *ddst, double *scl) {
	int lxy,bxy,exy,lz;
	__m128d *src, *dst;

	int i,j,is,id,pxy;
	__m128d t1,t2,ts;
	int tsd1, tsd2, tbs;

	const int cpfr = 900;
	const int cpf = 450;
	int ipf,ipfr,r;

	lxy = *ilxy;
	bxy = *ibxy;
	exy = *iexy;
	lz = *ilz;

	src = (__m128d *)dsrc;
	dst = (__m128d *)ddst;

	ipf = cpf/lz;
	ipf = (ipf>2) ? ipf : 2;

	ipfr = cpfr/lz;
	ipfr = (ipfr>1) ? ipfr : 1;

	pxy = exy-bxy+1;

	r = pxy % 2;

	tsd1 = lxy;
	tsd2 = lz;
	tbs = bxy-1;

	ts = _mm_set1_pd(*scl);

	if (r == 1) {
		is = tbs;
		id = 0;

		for (j = 0; j < lz; j++) {
			_mm_prefetch((char *)(&src[is+ipfr]),_MM_HINT_T0);

			t1 = src[is];
			is = is + tsd1;
			t1 = _mm_mul_pd(ts,t1);
			dst[id] = t1;
			id = id + 1;
		}
	}
	for (i = r; i < pxy; i = i+2) {
		is = i+tbs;
		id = tsd2*i;

		for (j = 0; j < lz; j++) {
			_mm_prefetch((char *)(&src[is+ipf]),_MM_HINT_T0);

			t1 = src[is];
			t2 = src[is+1];
			is = is + tsd1;
			t1 = _mm_mul_pd(ts,t1);
			dst[id] = t1;
			t2 = _mm_mul_pd(ts,t2);
			dst[id+tsd2] = t2;
			id = id + 1;
		}
	}
    return 0;
}
