!
! Copyright (c) 2000-2017 Yoshihide Yoshimoto
! 
program main
  implicit none
  integer,parameter:: nfin = 10, nfout = 11
  real(kind=8),parameter:: oneau = 0.529177249d0
  integer:: nrx2, nry2, nrz2
  real(kind=8):: aa(3,3)
  character(len=256):: ifname, ofname, argv(6)
  integer:: ios
  logical:: error
  real(kind=8),allocatable:: field(:,:,:,:)
  character(len=8),parameter:: magic = 'TAPP:rho'
  character(len=8):: sign
  integer:: ic, i
  logical:: multiq, nextq
  character(len=256):: acount
  integer:: nspin
  integer:: ix,iy,iz,is
  real(kind=8):: rorg(3)
  character(len=6):: namespin(4,4) = reshape( &
   (/ 'charge', & ! nspin = 1
      'inval ', &
      'inval ', &
      'inval ', &
      'spin1 ', & ! nspin = 2
      'spin2 ', &
      'inval ', &
      'inval ', &
      'inval ', & ! nspin = 3
      'inval ', &
      'inval ', &
      'inval ', &
      'charge', & ! nspin = 4
      'spinx ', &
      'spiny ', &
      'spinz ' /), shape(namespin) )

  ic = 1
  multiq = .false.
  call getarg(ic, argv(ic))
  if ( argv(ic) .eq. '-m' ) then
     multiq = .true.
     ic = ic + 1
  end if
  call getarg(ic, argv(ic))
  if (len_trim(argv(ic)) .eq. 0) then
     write(6,*) 'invalid fname'
     stop
  end if
  ifname = argv(ic)
  open(nfin, file=ifname, status='old', action='read', form='unformatted')
  ic = ic + 1
  do i = ic, ic + 2
     call getarg(i, argv(i))
  end do
  read(argv(ic),*,iostat=ios) rorg(1)
  if (ios .eq. 0) then
     error = .false.
     read(argv(ic+1),*,iostat=ios) rorg(2)
     error = error .or. (ios .ne. 0)
     read(argv(ic+2),*,iostat=ios) rorg(3)
     error = error .or. (ios .ne. 0)
     if (error) then
        write(6,*) 'invalid position of origin'
        stop
     end if
  else
     rorg(1) = 0.0d0
     rorg(2) = 0.0d0
     rorg(3) = 0.0d0
  end if

  rewind nfin
  if (multiq) then
     ic = 1
  end if
  read(nfin) sign
  if (sign .ne. magic) then
     write(6,*) 'bad magic: ', sign
     stop
  end if
  nextq = .true.
  do while (nextq)
     read(nfin) aa, nspin
     read(nfin) nrx2, nry2, nrz2
     allocate(field(nrx2,nry2,nrz2,nspin))
     read(nfin) field

     if (multiq) then
        write(acount, '(''.'',i3.3)') ic
        ofname=trim(ifname)//trim(acount)//'.xsf'//achar(0)
     else
        ofname=trim(ifname)//'.xsf'//achar(0)
     end if

     open(nfout, file=ofname)
     write(nfout, *) 'CRYSTAL'
     write(nfout, *) 'PRIMVEC'
     write(nfout,'(3(1x,e18.11))') aa(:,1)*oneau
     write(nfout,'(3(1x,e18.11))') aa(:,2)*oneau
     write(nfout,'(3(1x,e18.11))') aa(:,3)*oneau
     write(nfout,*)
     write(nfout,*) 'BEGIN_BLOCK_DATAGRID_3D'
     write(nfout,*) 'TAPP_rho'
     do is=1,nspin
        write(nfout,*) 'BEGIN_DATAGRID_3D_'//trim(namespin(is,nspin))
        write(nfout,'(3(1x,i6))') nrx2+1,nry2+1,nrz2+1
        write(nfout,'(3(1x,e18.11))') rorg(1),rorg(2),rorg(3)
        write(nfout,'(3(1x,e18.11))') aa(:,1)*oneau
        write(nfout,'(3(1x,e18.11))') aa(:,2)*oneau
        write(nfout,'(3(1x,e18.11))') aa(:,3)*oneau
        write(nfout,*) (((field(modulo(ix,nrx2)+1,modulo(iy,nry2)+1, &
             modulo(iz,nrz2)+1,is),ix=0,nrx2),iy=0,nry2),iz=0,nrz2)
        write(nfout,*) 'END_DATAGRID_3D'
     end do
     write(nfout,*) 'END_BLOCK_DATAGRID_3D'

     deallocate(field)

     if (.not. multiq) then
        exit
     end if

     read(nfin, iostat=ios) sign
     if (ios .ne. 0) then
        nextq = .false.
     else if (sign .ne. magic) then
        write(6,*) 'bad magic: ', sign
        stop
     end if
     ic = ic + 1
  end do
  stop
end program main
