!
! Copyright (c) 2000-2013,2016 Yoshihide Yoshimoto
! 
#include "fftgrid.h"

program wfn2dx
  use fft_3d
  use util
  implicit none
  integer, parameter:: nfin = 10, nfout = 11
  real(kind=8),parameter:: p2 = 6.28318530717958648d0
  real(kind=8),parameter:: oneau = 0.529177249d0

  integer:: iorg(3)
  character(len=8):: sign
  integer:: ver
  integer:: nks, nw, ng, nnp
  integer:: nkmmax
  integer,allocatable:: nkmary(:), kg(:,:), rg(:,:,:), pg(:,:), srg(:)
  real(kind=8),allocatable:: sk(:,:)
  real(kind=8),allocatable:: pr(:)
  complex(kind=8),allocatable:: p(:,:)
  real(kind=8),allocatable:: psir(:,:,:,:,:), ztmp(:,:,:,:), rtmp(:,:,:,:,:)
  type (fft3_struct) :: fs
  integer:: nvx, nvy, nvz, nrx, nry, nrz, nlx, nly, nlz, nspin, ncomp
  real(kind=8):: bb(3,3), aa(3,3)
  real(kind=8):: delta(3,3), origin(3) = (/0.0d0, 0.0d0, 0.0d0/)
  integer:: i1, i2, i3, k1, k2, k3, i, j, ios, ik, is, ig, ib, ibt
  character(len=256):: ifname, ofname, name, buf1, buf2, argv(4)
  logical:: error,ifinvsymi
  real(kind=8):: w(2) = (/0.0d0, 0.0d0/), ee(2) = (/0.0d0, 0.0d0/)

  call getarg(1, argv(1))
  if (len_trim(argv(1)) .eq. 0) then
     write(6,*) 'invalid fname'
     stop
  end if
  ifname = argv(1)
  call getarg(2, argv(2))
  call getarg(3, argv(3))
  call getarg(4, argv(4))
  read(argv(2),*,iostat=ios) iorg(1)
  if (ios .eq. 0) then
     error = .false.
     read(argv(3),*,iostat=ios) iorg(2)
     error = error .or. (ios .ne. 0)
     read(argv(4),*,iostat=ios) iorg(3)
     error = error .or. (ios .ne. 0)
     if (error) then
        write(6,*) 'invalid position of origin'
        stop
     end if
  else
     iorg(1) = 0
     iorg(2) = 0
     iorg(3) = 0
  end if

  open(nfin, file=ifname, status='old', action='read', form='unformatted')
  ifinvsymi = .false.
  read(nfin, iostat=ios) sign, ver
  if (ios.ne.0) then
     write(6,*) 'unknown format'
     stop
  end if
  if (ver.ne.3) then
     write(6,*) 'only version 3 format is supported by this program'
  end if
  if (sign .ne. 'TAPP:wfn') then
     if (sign .ne. 'TAPP:wfi') then
        write(6,*) 'Bad magic'
        stop
     else
        ifinvsymi = .true.
     end if
  end if
  read(nfin) aa, bb
  read(nfin) nspin, nks, nw, ng, nnp, ncomp
  allocate(nkmary(nks), sk(3,nks), rg(3,3,2*ng), pg(3,2*ng) ,srg(2*ng))
  read(nfin,iostat=ios) rg, pg, srg
  if (ios.ne.0) then
     !
     ! make it a grey group.
     !
     write(6,*) 'old symmetry dat: make the symmetry a grey group'
     srg = 1
     rg(:,:,ng+1:2*ng) = rg(:,:,1:ng)
     pg(:,ng+1:2*ng) = pg(:,1:ng)
     srg(1:ng) = 1
     srg(ng+1:2*ng) = -1
     ng = 2*ng
  end if
  read(nfin) (sk(:,ik),ik=1,nks)
  read(nfin) nkmary
  nkmmax=maxval(nkmary)
  allocate(kg(3,nkmmax), pr(nkmmax), p(ncomp,nkmmax))
  !
  ! determine the range of kg.
  ! read kg only and skip wave functions.
  !
  nvx=0
  nvy=0
  nvz=0
  do ik=1,nks
     read(nfin) ((kg(i,ig),i=1,3),ig=1,nkmary(ik))
     do ig=1,nkmary(ik)
        nvx=max(nvx, abs(kg(1,ig)))
        nvy=max(nvy, abs(kg(2,ig)))
        nvz=max(nvz, abs(kg(3,ig)))
     end do
     do is=1,nspin
        do ib=1,nw
           if (ifinvsymi) then
              read(nfin)
           else
              read(nfin)
           end if
        end do
     end do
  end do
  !
  ! return the file pointer to the head of the wave functions.
  !
  do ik=1,nks
     do is=1,nspin
        do ib=1,nw
           if (ifinvsymi) then
              backspace(nfin)
           else
              backspace(nfin)
           end if
        end do
     end do
     backspace(nfin)
  end do
  nrx = algn235(nvx+1,1)
  nry = algn235(nvy+1,1)
  nrz = algn235(nvz+1,1)
  nlx = FFT_PAD_NX(nrx*2)
  nly = FFT_PAD_NY(nry*2)
  nlz = FFT_PAD_NZ(nrz*2)
  allocate(FFT_A3M_DEFSZ(psir, nlx, nly, nlz, nspin), FFT_A3D_DEFSZ(ztmp, nlx, nly, nlz), rtmp(nlx, nly, nlz, 2, nspin))
  call fft3_init(nrx*2,nry*2,nrz*2,nlx,nly,nlz,fs)

  delta(:,1) = oneau*aa(:,1)/(nrx*2)
  delta(:,2) = oneau*aa(:,2)/(nry*2)
  delta(:,3) = oneau*aa(:,3)/(nrz*2)

  do ik=1,nks
     read(nfin) ((kg(i,ig),i=1,3),ig=1,nkmary(ik))

     do ib=1,nw
        psir = 0.0d0
        if (ifinvsymi) then
           read(nfin) (pr(i),i=1,nkmary(ik)), w(1), ee(1)
           do ig=1,nkmary(ik)
              i1 = kg(1,ig)
              i2 = kg(2,ig)
              i3 = kg(3,ig)
              k1 = modulo(i1, nrx*2) + 1
              k2 = modulo(i2, nry*2) + 1
              k3 = modulo(i3, nrz*2) + 1
              FFT_A3M_RL(psir, k1, k2, k3, 1) = pr(ig)
              FFT_A3M_IM(psir, k1, k2, k3, 1) = 0.0d0
           end do
        else
           read(nfin) ((p(i,j),i=1,ncomp),j=1,nkmary(ik)), w(1), ee(1)
           do ig=1,nkmary(ik)
              i1 = kg(1,ig)
              i2 = kg(2,ig)
              i3 = kg(3,ig)
              k1 = modulo(i1, nrx*2) + 1
              k2 = modulo(i2, nry*2) + 1
              k3 = modulo(i3, nrz*2) + 1
              FFT_A3M_RL(psir, k1, k2, k3, 1) = dble(p(1,ig))
              FFT_A3M_IM(psir, k1, k2, k3, 1) = aimag(p(1,ig))
           end do
        end if

        call fft3_bw(fs, FFT_A3M_RL(psir,1,1,1,1), ztmp)

        if (nspin .eq. 2) then
           !
           ! skip to the second spin
           !
           do ibt = 1,nw-1
              if (ifinvsymi) then
                 read(nfin)
              else
                 read(nfin)
              end if
           end do

           if (ifinvsymi) then
              read(nfin) (pr(i),i=1,nkmary(ik)), w(2), ee(2)
              do ig=1,nkmary(ik)
                 i1 = kg(1,ig)
                 i2 = kg(2,ig)
                 i3 = kg(3,ig)
                 k1 = modulo(i1, nrx*2) + 1
                 k2 = modulo(i2, nry*2) + 1
                 k3 = modulo(i3, nrz*2) + 1
                 FFT_A3M_RL(psir, k1, k2, k3, 2) = pr(ig)
                 FFT_A3M_IM(psir, k1, k2, k3, 2) = 0.0d0
              end do
           else
              read(nfin) ((p(i,j),i=1,ncomp),j=1,nkmary(ik)), w(2), ee(2)
              do ig=1,nkmary(ik)
                 i1 = kg(1,ig)
                 i2 = kg(2,ig)
                 i3 = kg(3,ig)
                 k1 = modulo(i1, nrx*2) + 1
                 k2 = modulo(i2, nry*2) + 1
                 k3 = modulo(i3, nrz*2) + 1
                 FFT_A3M_RL(psir, k1, k2, k3, 2) = dble(p(1,ig))
                 FFT_A3M_IM(psir, k1, k2, k3, 2) = aimag(p(1,ig))
              end do
           end if

           call fft3_bw(fs, FFT_A3M_RL(psir,1,1,1,2), ztmp)
           !
           ! return to the first spin
           !
           do ibt = 1,nw
              if (ifinvsymi) then
                 backspace(nfin)
              else
                 backspace(nfin)
              end if
           end do
        end if

        write(buf1,'(i0)') ib
        buf1=adjustl(buf1)
        write(buf2,'(i0)') ik
        buf2=adjustl(buf2)
        name='TAPP wfn'//achar(0)

        do is=1,nspin
           rtmp(:,:,:,1,is) = FFT_A3M_RL(psir,:,:,:,is)
           rtmp(:,:,:,2,is) = FFT_A3M_IM(psir,:,:,:,is)
        end do
     
        ofname=trim(ifname)//'.'//trim(buf1)//'.'//trim(buf2)//'.dx' &
             //achar(0)
        call outzfldprsk(ofname, name, nrx*2, nry*2, nrz*2, delta, &
                origin, nspin, nlx, nly, nlz, rtmp, 0, &
                sk(1,ik), ee, 1, iorg)
     end do
     !
     ! advance to the next k-point
     !
     if (nspin .eq. 2) then
        do ib=1,nw
           if (ifinvsymi) then
              read(nfin)
           else
              read(nfin)
           end if
        end do
     end if
  end do
  call fft3_done(fs)
end program wfn2dx
