!
! Copyright (c) 2013,2016,2017 Yoshihide Yoshimoto
!
module tetrafit
contains
  subroutine iksnst(sk,ng,rg,srg,bb,nksi,ski,nskip,skip,&
       iksret,skret,rmin)
    implicit none
    integer,intent(in):: ng, rg(3,3,ng), srg(ng), nksi
    integer,intent(out):: iksret
    real(kind=8),intent(out):: skret(3),rmin
    real(kind=8),intent(in):: sk(3),bb(3,3),ski(3,nksi)
    integer,intent(in):: nskip
    real(kind=8),intent(in):: skip(3,nskip)
    
    integer:: ig, iks, iiv, iskip
    real(kind=8):: r, tsk(3), d(3)
    logical:: ifcycle

    real(kind=8),parameter:: eps = 1.0d-8

    iksret = 0
    rmin = huge(rmin)
    do iks = 1, nksi
       do ig = 1, ng
          tsk = srg(ig)*matmul(rg(:,:,ig),ski(:,iks))
          tsk = tsk + nint(sk-tsk)

          ifcycle = .false.
          do iskip = 1,nskip
             if (maxval(abs(tsk-skip(:,iskip))) .lt. eps) then
                ifcycle = .true.
                exit
             end if
          end do

          if (.not. ifcycle) then
             d = sk - tsk
             d = matmul(bb,d)
             r = sqrt(d(1)*d(1) + d(2)*d(2) + d(3)*d(3))
             if (r .lt. rmin) then
                iksret = iks
                skret = tsk
                rmin = r
             end if
          end if
       end do
    end do
    if (iksret .eq. 0) then
       write(6,*) 'iksnst: iksret = 0'
       stop
    end if
    return
  end subroutine iksnst

  subroutine tetrafitee(nw,nspin,nks,bb,sk,ng,rg,pg,srg,ee,tsk,tee)
    implicit none
    integer,intent(in):: nw,nspin,nks,ng,rg(3,3,ng),pg(3,ng),srg(ng)
    real(kind=8),intent(in):: bb(3,3),sk(3,nks),ee(nw,nspin,nks)
    real(kind=8),intent(in):: tsk(3)
    real(kind=8):: tee(nspin,nw)

    real(kind=8):: skwgt(3),rmin(4),skvert(3,4),skip(3,10)
    integer:: vert(4)
    integer:: i,ib,is,nskip
    real(kind=8):: dsk(3,3),dski(3,3),v,w
    real(kind=8):: det(4)
    integer:: loc(1)

    real(kind=8),parameter:: eps = 1.0d-8

    nskip = 0
    call iksnst(tsk,ng,rg,srg,bb,nks,sk,nskip,skip, &
         vert(1),skvert(:,1),rmin(1))
    nskip = nskip + 1
    skip(:,nskip) = skvert(:,1)
    if (rmin(1).lt.eps) then
       do ib=1,nw
          do is=1,nspin
             tee(is,ib) = ee(ib,is,vert(1))
          end do
       end do
       return
    end if

    do i = 2,3
       call iksnst(tsk,ng,rg,srg,bb,nks,sk,nskip,skip, &
            vert(i),skvert(:,i),rmin(i))
       nskip = nskip + 1
       skip(:,nskip) = skvert(:,i) 
    end do
    dsk(:,1) = skvert(:,2)-skvert(:,1)
    dsk(:,2) = skvert(:,3)-skvert(:,1)
    v = dsk(1,1)*dsk(1,2)+dsk(2,1)*dsk(2,2)+dsk(3,1)*dsk(3,2)
    w = dsk(1,1)*dsk(1,1)+dsk(2,1)*dsk(2,1)+dsk(3,1)*dsk(3,1)
    w = w * ( dsk(1,2)*dsk(1,2)+dsk(2,2)*dsk(2,2)+dsk(3,2)*dsk(3,2) )
    v = v/sqrt(w)
    !
    ! check if dsk(:,1) and dsk(:,2) are parallel
    !
    if (abs(v-1.0d0).lt.eps .or. abs(v+1.0d0).lt.eps) then
       call iksnst(tsk,ng,rg,srg,bb,nks,sk,nskip,skip, &
            vert(3),skvert(:,3),rmin(3))
       nskip = nskip + 1
       skip(:,nskip) = skvert(:,3) 
    end if

    call iksnst(tsk,ng,rg,srg,bb,nks,sk,nskip,skip, &
         vert(4),skvert(:,4),rmin(4))
    nskip = nskip + 1
    skip(:,nskip) = skvert(:,3) 

    !
    ! check if tsk is on the plane spaned by
    ! skvert(:,1), skvert(:,2), skvert(:,3)
    !
    dsk(:,1) = skvert(:,2)-skvert(:,1)
    dsk(:,2) = skvert(:,3)-skvert(:,1)
    dsk(:,3) = tsk-skvert(:,1)
    det(1)=dsk(1,1)*(dsk(2,2)*dsk(3,3)-dsk(3,2)*dsk(2,3)) &
         +dsk(2,1)*(dsk(3,2)*dsk(1,3)-dsk(1,2)*dsk(3,3)) &
         +dsk(3,1)*(dsk(1,2)*dsk(2,3)-dsk(2,2)*dsk(1,3))

    !
    ! check if tsk is on the plane spaned by
    ! skvert(:,1), skvert(:,2), skvert(:,4)
    !
    dsk(:,1) = skvert(:,2)-skvert(:,1)
    dsk(:,2) = tsk-skvert(:,1)
    dsk(:,3) = skvert(:,4)-skvert(:,1)
    det(2)=dsk(1,1)*(dsk(2,2)*dsk(3,3)-dsk(3,2)*dsk(2,3)) &
         +dsk(2,1)*(dsk(3,2)*dsk(1,3)-dsk(1,2)*dsk(3,3)) &
         +dsk(3,1)*(dsk(1,2)*dsk(2,3)-dsk(2,2)*dsk(1,3))

    !
    ! check if tsk is on the plane spaned by
    ! skvert(:,1), skvert(:,3), skvert(:,4)
    !
    dsk(:,1) = tsk-skvert(:,1)
    dsk(:,2) = skvert(:,3)-skvert(:,1)
    dsk(:,3) = skvert(:,4)-skvert(:,1)
    det(3)=dsk(1,1)*(dsk(2,2)*dsk(3,3)-dsk(3,2)*dsk(2,3)) &
         +dsk(2,1)*(dsk(3,2)*dsk(1,3)-dsk(1,2)*dsk(3,3)) &
         +dsk(3,1)*(dsk(1,2)*dsk(2,3)-dsk(2,2)*dsk(1,3))

    !
    ! check if tsk is on the plane spaned by
    ! skvert(:,2), skvert(:,3), skvert(:,4)
    !
    dsk(:,1) = skvert(:,2)-skvert(:,4)
    dsk(:,2) = skvert(:,3)-skvert(:,4)
    dsk(:,3) = tsk-skvert(:,4)
    det(4)=dsk(1,1)*(dsk(2,2)*dsk(3,3)-dsk(3,2)*dsk(2,3)) &
         +dsk(2,1)*(dsk(3,2)*dsk(1,3)-dsk(1,2)*dsk(3,3)) &
         +dsk(3,1)*(dsk(1,2)*dsk(2,3)-dsk(2,2)*dsk(1,3))

    loc(1) = -1
    if (minval(abs(det)) .lt. eps) then
       loc = minloc(abs(det))

       if (loc(1).eq.1) then
          dsk(:,1) = skvert(:,2)-skvert(:,1)
          dsk(:,2) = skvert(:,3)-skvert(:,1)
          
          dsk(1,3) = dsk(2,1)*dsk(3,2)-dsk(3,1)*dsk(2,2)
          dsk(2,3) = dsk(3,1)*dsk(1,2)-dsk(1,1)*dsk(3,2)
          dsk(3,3) = dsk(1,1)*dsk(2,2)-dsk(2,1)*dsk(1,2)
       else if (loc(1).eq.2) then
          dsk(:,1) = skvert(:,2)-skvert(:,1)
          dsk(:,3) = skvert(:,4)-skvert(:,1)
          
          dsk(1,2) = dsk(2,3)*dsk(3,1)-dsk(3,1)*dsk(2,3)
          dsk(2,2) = dsk(3,3)*dsk(1,1)-dsk(1,1)*dsk(3,3)
          dsk(3,2) = dsk(1,3)*dsk(2,1)-dsk(2,1)*dsk(1,3)
       else if (loc(1).eq.3) then
          dsk(:,2) = skvert(:,3)-skvert(:,1)
          dsk(:,3) = skvert(:,4)-skvert(:,1)
          
          dsk(1,1) = dsk(2,2)*dsk(3,3)-dsk(3,3)*dsk(2,2)
          dsk(2,1) = dsk(3,2)*dsk(1,3)-dsk(1,3)*dsk(3,2)
          dsk(3,1) = dsk(1,2)*dsk(2,3)-dsk(2,3)*dsk(1,2)
       else if (loc(1).eq.4) then
          dsk(:,1) = skvert(:,2)-skvert(:,4)
          dsk(:,2) = skvert(:,3)-skvert(:,4)
          
          dsk(1,3) = dsk(2,1)*dsk(3,2)-dsk(3,1)*dsk(2,2)
          dsk(2,3) = dsk(3,1)*dsk(1,2)-dsk(1,1)*dsk(3,2)
          dsk(3,3) = dsk(1,1)*dsk(2,2)-dsk(2,1)*dsk(1,2)

       end if
    else
       dsk(:,1) = skvert(:,2)-skvert(:,1)
       dsk(:,2) = skvert(:,3)-skvert(:,1)
       dsk(:,3) = skvert(:,4)-skvert(:,1)
    end if
    v=dsk(1,1)*(dsk(2,2)*dsk(3,3)-dsk(3,2)*dsk(2,3)) &
         +dsk(2,1)*(dsk(3,2)*dsk(1,3)-dsk(1,2)*dsk(3,3)) &
         +dsk(3,1)*(dsk(1,2)*dsk(2,3)-dsk(2,2)*dsk(1,3))

    do i=5,10
       if (abs(v) .gt. eps) then
          dski(1,1)=(dsk(2,2)*dsk(3,3)-dsk(3,2)*dsk(2,3))/v
          dski(1,2)=(dsk(3,2)*dsk(1,3)-dsk(1,2)*dsk(3,3))/v
          dski(1,3)=(dsk(1,2)*dsk(2,3)-dsk(2,2)*dsk(1,3))/v
          dski(2,1)=(dsk(2,3)*dsk(3,1)-dsk(3,3)*dsk(2,1))/v
          dski(2,2)=(dsk(3,3)*dsk(1,1)-dsk(1,3)*dsk(3,1))/v
          dski(2,3)=(dsk(1,3)*dsk(2,1)-dsk(2,3)*dsk(1,1))/v
          dski(3,1)=(dsk(2,1)*dsk(3,2)-dsk(3,1)*dsk(2,2))/v
          dski(3,2)=(dsk(3,1)*dsk(1,2)-dsk(1,1)*dsk(3,2))/v
          dski(3,3)=(dsk(1,1)*dsk(2,2)-dsk(2,1)*dsk(1,2))/v

          skwgt = matmul(dski,tsk-skvert(:,1))
          if (minval(skwgt).lt.0.0d0 .or. maxval(skwgt).gt.1.0d0) then
             continue
          else
             exit
          end if
       end if

       !
       ! treat the case the found vertexes are liner dependent
       !
       call iksnst(tsk,ng,rg,srg,bb,nks,sk,nskip,skip, &
            vert(4),skvert(:,4),rmin(4))
       nskip = nskip + 1
       skip(:,nskip) = skvert(:,4)
       
       dsk(:,1) = skvert(:,2)-skvert(:,1)
       dsk(:,2) = skvert(:,3)-skvert(:,1)
       dsk(:,3) = skvert(:,4)-skvert(:,1)

       v=dsk(1,1)*(dsk(2,2)*dsk(3,3)-dsk(3,2)*dsk(2,3)) &
            +dsk(2,1)*(dsk(3,2)*dsk(1,3)-dsk(1,2)*dsk(3,3)) &
            +dsk(3,1)*(dsk(1,2)*dsk(2,3)-dsk(2,2)*dsk(1,3))
    end do

    if (abs(v) .lt. eps) then
       write(6,*) 'degenerated tetrahedron was found'
       write(6,*) 'the number of k-points may be too small'
       stop
    end if

    do ib=1,nw
       do is=1,nspin
          tee(is,ib) = ee(ib,is,vert(1)) &
               + skwgt(1)*( ee(ib,is,vert(2)) - ee(ib,is,vert(1)) ) &
               + skwgt(2)*( ee(ib,is,vert(3)) - ee(ib,is,vert(1)) ) &
               + skwgt(3)*( ee(ib,is,vert(4)) - ee(ib,is,vert(1)) )
       end do
    end do
    
  end subroutine tetrafitee
end module tetrafit

program wfn2ee
  use tetrafit
  implicit none
  integer, parameter:: nfin = 10, nfout = 11

  integer:: iorg(3)
  character(len=8):: sign
  integer:: ver
  integer:: nks, nw, ng, nnp, ncomp
  integer:: nkmmax
  integer,allocatable:: nkmary(:), kg(:,:), rg(:,:,:), pg(:,:), srg(:)
  real(kind=8),allocatable:: sk(:,:)
  real(kind=8),allocatable:: pr(:)
  complex(kind=8),allocatable:: p(:,:)
  integer:: n1, n2, n3, nspin
  real(kind=8):: bb(3,3), bbx(3,3), aa(3,3), tsk(3), x, y, z
  integer:: ix(3,3), i1, i2, i3, i, j, ios, ik, is, ig, ib
  character(len=256):: ifname, ofname, name, buf1, argv(6)
  logical:: error,ifinvsymi
  real(kind=8),allocatable:: w(:,:,:),ee(:,:,:),tee(:,:),aee(:,:,:,:,:)
  real(kind=8),parameter:: oneau = 0.529177249d0
  integer:: mode

  call getarg(1, argv(1))
  if (len_trim(argv(1)) .eq. 0) then
     write(6,*) 'invalid fname'
     stop
  end if
  ifname = trim(argv(1)) // '.wfn'

  call getarg(2, argv(2))
  if (trim(argv(2)).eq.'fermisurfer') then
     mode = 1
     ix = 0
     ix(1,1) = 1
     ix(2,2) = 1
     ix(3,3) = 1
     iorg = 0
     call getarg(3, argv(3))
     read(argv(3),*) n1
     call getarg(4, argv(4))
     read(argv(4),*) n2
     call getarg(5, argv(5))
     read(argv(5),*) n3
  else
     mode = 2
     read(argv(2),*) ix(1,1), ix(2,1), ix(3,1)
     call getarg(3, argv(3))
     read(argv(3),*) ix(1,2), ix(2,2), ix(3,2)
     call getarg(4, argv(4))
     read(argv(4),*) ix(1,3), ix(2,3), ix(3,3)
     call getarg(5, argv(5))
     read(argv(5),*) iorg(1), iorg(2), iorg(3)
     call getarg(6, argv(6))
     read(argv(6),*) n1, n2, n3
  end if

  open(nfin, file=ifname, status='old', action='read', form='unformatted')
  ifinvsymi = .false.
  read(nfin, iostat=ios) sign, ver
  if (ios.ne.0) then
     write(6,*) 'unknown format'
     stop
  end if
  if (ver.ne.3) then
     write(6,*) 'only version 3 format is supported by this program'
     stop
  end if
  if (sign .ne. 'TAPP:wfn') then
     if (sign .ne. 'TAPP:wfi') then
        write(6,*) 'Bad magic'
        stop
     else
        ifinvsymi = .true.
     end if
  end if
  read(nfin) aa, bb
  read(nfin) nspin, nks, nw, ng, nnp, ncomp
  allocate(nkmary(nks), sk(3,nks), rg(3,3,2*ng), pg(3,2*ng), srg(2*ng))
  read(nfin,iostat=ios) rg(:,:,1:ng), pg(:,1:ng), srg(1:ng)
  if (ios.ne.0) then
     !
     ! make it a grey group.
     !
     write(6,*) 'old symmetry dat: make the symmetry a grey group'
     srg = 1
     rg(:,:,ng+1:2*ng) = rg(:,:,1:ng)
     pg(:,ng+1:2*ng) = pg(:,1:ng)
     srg(1:ng) = 1
     srg(ng+1:2*ng) = -1
     ng = 2*ng
  end if
  read(nfin) (sk(:,ik),ik=1,nks)
  read(nfin) nkmary
  nkmmax=maxval(nkmary)
  allocate(kg(3,nkmmax), pr(nkmmax), p(ncomp,nkmmax))
  allocate(w(nw,nspin,nks),ee(nw,nspin,nks))

  do ik=1,nks
     read(nfin) ((kg(i,ig),i=1,3),ig=1,nkmary(ik))

     do is=1,nspin
        do ib=1,nw
           if (ifinvsymi) then
              read(nfin) (pr(i),i=1,nkmary(ik)), w(ib,is,ik), ee(ib,is,ik)
           else
              read(nfin) ((p(i,j),i=1,ncomp),j=1,nkmary(ik)), &
                   w(ib,is,ik), ee(ib,is,ik)
           end if
        end do
     end do
  end do

  allocate(aee(0:n1-1,0:n2-1,0:n3-1,nspin,nw))

!$omp parallel private(i1,i2,i3,x,y,z,tsk,tee)
!$omp critical
  allocate(tee(nspin,nw))
!$omp end critical
!$omp do
  do i3=0,n3-1
     z = dble(i3)/n3
     do i2=0,n2-1
        y = dble(i2)/n2
        do i1=0,n1-1
           x = dble(i1)/n1

           tsk(1) = ix(1,1)*x + ix(1,2)*y + ix(1,3)*z
           tsk(2) = ix(2,1)*x + ix(2,2)*y + ix(2,3)*z
           tsk(3) = ix(3,1)*x + ix(3,2)*y + ix(3,3)*z

           call tetrafitee(nw,nspin,nks,bb,sk,ng,rg,pg,srg,ee,tsk,tee)

           aee(i1,i2,i3,:,:) = tee
        end do
     end do
  end do
!$omp critical
  deallocate(tee)
!$omp end critical
!$omp end parallel

  if (mode .eq. 1) then
     do is=1,nspin
        write(buf1,'(i0)') is
        buf1=adjustl(buf1)
        ofname=trim(argv(1))//'.'//trim(buf1)//'.frmsf'//achar(0)
        open(nfout,file=ofname)
        write(nfout,*) n1,n2,n3
        write(nfout,*) 1
        write(nfout,*) nw
        write(nfout,*) bb(:,1)
        write(nfout,*) bb(:,2)
        write(nfout,*) bb(:,3)
        do ib=1,nw
           do i1=0,n1-1
              do i2=0,n2-1
                 do i3=0,n3-1
                    write(nfout,*) real(aee(i1,i2,i3,is,ib))
                 end do
              end do
           end do
        end do
        do ib=1,nw
           do i1=1,n1
              do i2=1,n2
                 do i3=1,n3
                    write(nfout,*) real(ib)
                 end do
              end do
           end do
        end do
        close(nfout)
     end do
  else
     name='TAPP ee'//achar(0)
     bbx = matmul(bb,ix)
     bbx = bbx/oneau
     do ib=1,nw
        write(buf1,'(i0)') ib
        buf1=adjustl(buf1)
        ofname=trim(argv(1))//'.'//trim(buf1)//'.ee.dx'//achar(0)
        call outfldpr(ofname, name, n1, n2, n3, bbx, iorg, nspin, &
             n1, n2, n3, aee(:,:,:,:,ib), 0)
     end do
  end if
end program wfn2ee
