#!/bin/sh

set -- `getopt r:b:d: "$@"`

if [ $? != 0 ]; then
    echo "Usage: $0 [-r path_to_rho2chgcar ] [-b path_to_bader ] [-d path_to_distpart ] base_name" 1>&2
    exit 1
fi

RHO2CHGCAR=rho2chgcar
BADER=bader
DISTPART=distpart

while [ $# -gt 0 ]; do
    case $1 in
        --) shift; break ;;
        -r) RHO2CHGCAR=`readlink -f $2`; shift;;
        -b) BADER=`readlink -f $2`; shift;;
        -d) DISTPART=`readlink -f $2`; shift;;
    esac
    shift
done

which $RHO2CHGCAR > /dev/null
if  [ $? = 1 ]; then
    echo "Error: $RHO2CHGCAR was not found in your PATH"
    echo "specify the path to rho2chgcar by -r option"
    exit 1
fi
which $BADER > /dev/null
if  [ $? = 1 ]; then
    echo "Error: $BADER was not found in your PATH"
    echo "specify the path to bader by -b option"
    exit 1
fi
which $DISTPART > /dev/null
if  [ $? = 1 ]; then
    echo "Error: $DISTPART was not found in your PATH"
    echo "specify the path to distpart by -d option"
    exit 1
fi

target=$1

rm -rf $target-bader
mkdir $target-bader
cd $target-bader

ln -s ../fort.10
rm -f $target.str $target.rho
ln -s ../$target.str
ln -s ../$target.rho
$RHO2CHGCAR $target
rm -f $target.atmcd.str $target.atmcd.rho
ln -s ../$target.str $target.atmcd.str
ln -s ../$target.atmcd $target.atmcd.rho
$RHO2CHGCAR $target.atmcd

ln -s $target.chgcar CHGCAR
ln -s $target.atmcd.chgcar CHGCAR.atmrho
$BADER -p atom_index -ref CHGCAR.atmrho CHGCAR > bader.log
sed -e '/^ Bader/,/^$/d' AtIndex.dat > $target.idxpar

rm -f fort.10; ln -s ../fort.10
for suf in bspot loceng rho ekinor excor hrtor evxcor sekinor shrtor ; do
        rm -f $target.$suf ; ln -s ../$target.$suf
done

mpirun -np 1 $DISTPART bader > distpart.bader.log
