#!/bin/sh

set -- `getopt r:w:d: "$@"`

if [ $? != 0 ]; then
    echo "Usage: $0 [-r path_to_rho2chgcar ] [-w path_to_weight_int ] [-d path_to_distpart ] base_name" 1>&2
    exit 1
fi

RHO2CHGCAR=rho2chgcar
WEIGHT_INT=weight_int
DISTPART=distpart

while [ $# -gt 0 ]; do
    case $1 in
	--) shift; break ;;
	-r) RHO2CHGCAR=`readlink -f $2`; shift;;
	-w) WEIGHT_INT=`readlink -f $2`; shift;;
	-d) DISTPART=`readlink -f $2`; shift;;
    esac
    shift
done

which $RHO2CHGCAR > /dev/null
if  [ $? = 1 ]; then
    echo "Error: $RHO2CHGCAR was not found in your PATH"
    echo "specify the path to rho2chgcar by -r option"
    exit 1
fi
which $WEIGHT_INT > /dev/null
if  [ $? = 1 ]; then
    echo "Error: $WEHGHT_INT was not found in your PATH"
    echo "specify the path to weight_int by -w option"
    exit 1
fi
which $DISTPART > /dev/null
if  [ $? = 1 ]; then
    echo "Error: $DISTPART was not found in your PATH"
    echo "specify the path to distpart by -d option"
    exit 1
fi

target=$1

if [ ! -d $target-trinkle ]; then
    mkdir $target-trinkle
fi
cd $target-trinkle

error=0
for suf in atmcd ekinor excor hrtor; do
	if [ ! -f ../$target.$suf ]; then
		echo "$target.$suf not found"
		error=1
	fi
	rm -f $target.$suf.rho ; ln -s ../$target.$suf $target.$suf.rho
	rm -f $target.$suf.str ; ln -s ../$target.str $target.$suf.str
	$RHO2CHGCAR $target.$suf
done
rm -f $target.car2xtp ; ln -s $target.atmcd.car2xtp $target.car2xtp

for suf in evxcor sekinor shrtor; do
	if [ ! -f ../$target.$suf ]; then
		echo "$target.$suf not found"
		error=1
	fi
	rm -f $target.$suf.rho ; ln -s ../$target.$suf $target.$suf.rho
	rm -f $target.$suf.str ; ln -s ../$target.str $target.$suf.str
	$RHO2CHGCAR -m $target.$suf
done
if [ $error = 1 ]; then
	exit 1
fi

$WEIGHT_INT -s $target.atmcd.chgcar $target.ekinor.chgcar $target.excor.chgcar $target.hrtor.chgcar > $target.wgt-bader

$WEIGHT_INT -s $target.atmcd.chgcar $target.evxcor.001.chgcar $target.evxcor.002.chgcar $target.evxcor.003.chgcar $target.evxcor.004.chgcar $target.evxcor.005.chgcar $target.evxcor.006.chgcar $target.evxcor.007.chgcar > $target.wgt-evxc-bader

$WEIGHT_INT -s $target.atmcd.chgcar $target.sekinor.001.chgcar $target.sekinor.002.chgcar $target.sekinor.003.chgcar $target.sekinor.004.chgcar $target.sekinor.005.chgcar $target.sekinor.006.chgcar > $target.wgt-sekin-bader

$WEIGHT_INT -s $target.atmcd.chgcar $target.shrtor.001.chgcar $target.shrtor.002.chgcar $target.shrtor.003.chgcar $target.shrtor.004.chgcar $target.shrtor.005.chgcar $target.shrtor.006.chgcar > $target.wgt-seht-bader

rm -f fort.10; ln -s ../fort.10
for suf in bspot loceng rho; do
	rm -f $target.$suf ; ln -s ../$target.$suf
done

mpirun -np 1 $DISTPART trinkle $target > distpart.trinkle.log
