!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!
module inspectstr
contains

  subroutine statangle(nsi,kd,tkd1,tkd2,rm,d,na,nst,wgt,astat)
    implicit none
    integer:: nsi,na,nst
    integer:: kd(nsi),tkd1,tkd2
    real(kind=8):: rm,d(4,nsi,nsi),wgt(nst),astat(na,nst)

    integer:: it,jt,kt,ia
    real(kind=8):: r1(3),r2(3),dela,c
    real(kind=8),parameter:: pi = 3.14159265358979323846d0

    dela = pi/(na-1)

    do kt = 1,nsi 
       if (tkd1.ne.-1 .and. kd(kt).ne.tkd1) cycle
       do jt = 1,nsi
          if ( (tkd2.ne.-1 .and. kd(jt).ne.tkd2) &
               .or. d(4,jt,kt).gt.rm .or. jt.eq.kt) cycle
          do it = 1,jt-1
             if ( (tkd2.ne.-1 .and. kd(it).ne.tkd2) &
                  .or. d(4,it,kt).gt.rm .or. it.eq.kt) cycle
             c = d(1,jt,kt)*d(1,it,kt) + d(2,jt,kt)*d(2,it,kt) &
                  + d(3,jt,kt)*d(3,it,kt)
             ia = nint( acos(c)/dela ) + 1
             if ( (ia.ge.1) .and. (ia.le.na) ) then
                astat(ia,:) = astat(ia,:) + wgt(:)
             else
                write(6,*) 'a = ', acos(c)
             end if
          end do
       end do
    end do
    return
  end subroutine statangle

  subroutine caldist(nsi,asi,aa,d)
    implicit none
    integer:: nsi
    real(kind=8):: asi(3,nsi),aa(3,3),d(4,nsi,nsi)

    real(kind=8):: dlat(3,8)
    data dlat / &
         0.0d0,  0.0d0,  0.0d0,   0.0d0,  0.0d0,  1.0d0, & 
         0.0d0,  1.0d0,  0.0d0,   0.0d0,  1.0d0,  1.0d0, & 
         1.0d0,  0.0d0,  0.0d0,   1.0d0,  0.0d0,  1.0d0, & 
         1.0d0,  1.0d0,  0.0d0,   1.0d0,  1.0d0,  1.0d0 /
    integer:: it,jt,k
    real(kind=8):: r(3),x(3),z(3),t,dmin,xmin(3)

    do it = 1,nsi
       do jt = 1,nsi
          z = asi(:,jt) - asi(:,it)
          z = z - nint(z)
          dmin = huge(1.0d0)
          do k=1,8
             r(1) = z(1) - dlat(1,k)*sign(1.0d0,z(1))
             r(2) = z(2) - dlat(2,k)*sign(1.0d0,z(2))
             r(3) = z(3) - dlat(3,k)*sign(1.0d0,z(3))
             x(1) = aa(1,1)*r(1) + aa(1,2)*r(2) + aa(1,3)*r(3)
             x(2) = aa(2,1)*r(1) + aa(2,2)*r(2) + aa(2,3)*r(3)
             x(3) = aa(3,1)*r(1) + aa(3,2)*r(2) + aa(3,3)*r(3)
             t = sqrt(x(1)*x(1) + x(2)*x(2) + x(3)*x(3))
             if (t .lt. dmin) then
                dmin = t
                xmin = x
             end if
          end do
          d(4,jt,it) = dmin
          d(1:3,jt,it) = xmin/dmin
       end do
    end do
    return
  end subroutine caldist

  subroutine histdist(nsi,d,dbase,deld,nd,dhist)
    implicit none
    integer:: nsi,nd,dhist(nd)
    real(kind=8):: d(4,nsi,nsi),dbase,deld

    integer:: it,jt,id

    do it=1,nsi
       do jt=1,it-1
          id = nint( (d(4,jt,it)-dbase)/deld ) + 1
          if ( (id.ge.1) .and. (id.le.nd) ) then
             dhist(id) = dhist(id) + 1
          end if
       end do
    end do
    return
  end subroutine histdist

  subroutine statdist(nsi,kd,tkd1,tkd2,d,dbase,deld,nd,nst,wgt,dstat)
    implicit none
    integer:: nsi,nd,nst
    integer:: kd(nsi),tkd1,tkd2
    real(kind=8):: d(4,nsi,nsi),dbase,deld,wgt(nst),dstat(nd,nst)

    integer:: it,jt,id

    do it=1,nsi
       if (tkd1 .ne. -1 .and. kd(it) .ne. tkd1) cycle
       do jt=1,nsi
          if ( (tkd2 .ne. -1 .and. kd(jt) .ne. tkd2) .or. jt .eq. it) cycle
          id = nint( (d(4,jt,it)-dbase)/deld ) + 1
          if ( (id.ge.1) .and. (id.le.nd) ) then
             dstat(id,:) = dstat(id,:) + wgt(:)
          end if
       end do
    end do
    return
  end subroutine statdist

  subroutine findbond(nsi,d,thr,lb,b)
    implicit none
    integer:: nsi
    real(kind=8):: d(4,nsi,nsi),thr
    integer:: lb,b(0:lb,nsi)

    integer:: i,j,k

    b = 0
    do j=1,nsi
       k = 0
       do i=1,nsi
          if (i.ne.j .and. d(4,i,j).lt.thr) then
             k = k + 1
             if (k.gt.lb) then
                write(6,*) 'too small lb for atom ', j
                stop
             end if
             b(k,j) = i
          end if
       end do
       b(0,j) = k
    end do
    return
  end subroutine findbond

  subroutine findring(nsi,lb,b,rnk,lr,rm,maxring,r,nr)
    implicit none
    integer:: nsi,lb,b(0:lb,nsi),nb(nsi),rnk,maxring,r(rnk,maxring),nr
    integer:: lr,rm(0:lr,nsi)

    integer:: i,j,k,m
    integer:: t(rnk)

    rm = 0
    r = 0
    nr = 0
    do j=1,nsi
       do i=1,b(0,j)
          m = 1
          t = 0
          t(1) = j
          call visitring(nsi,lb,b,rnk,m,t,lr,rm,maxring,r,nr)
       end do
    end do
    return
  end subroutine findring

  recursive subroutine visitring(nsi,lb,b,rnk,im,t,lr,rm,maxring,r,nr)
    implicit none
    integer:: nsi,lb,b(0:lb,nsi),rnk,im,t(rnk),lr,rm(0:lr,nsi)
    integer:: maxring,r(rnk,maxring),nr

    integer:: m,i,j,k,vertnew,vertmin,pos,bpos,npos
    integer:: s(rnk),sr

    if ( im.eq.rnk ) then
       do i=1,b(0,t(im))
          if ( t(1).eq.b(i,t(im)) ) then
             vertmin = t(1)
             pos = 1
             do j=2,rnk
                if (vertmin.gt.t(j)) then
                   vertmin = t(j)
                   pos = j
                end if
             end do
             npos = modulo(pos-1+1,rnk)+1
             bpos = modulo(pos-1-1,rnk)+1
             if (t(npos).lt.t(bpos)) then
                do j=0,rnk-1
                   k = modulo(j+pos-1,rnk)+1
                   s(j+1) = t(k)
                end do
             else
                do j=rnk,1,-1
                   k = modulo(j+pos-1,rnk)+1
                   s(rnk-j+1) = t(k)
                end do
             end if

             do j=1,rm(0,s(1))
                sr = 1
                do k=1,rnk
                   if (r(k,rm(j,s(1))).ne.s(k)) then
                      sr = 0
                      cycle
                   end if
                end do
                if (sr.eq.1) then
                   return
                end if
             end do

             nr = nr + 1
             if (nr.gt.maxring) then
                write(6,*) 'maxring is too small'
                stop
             end if
             r(:,nr) = s
             do j=1,rnk
                rm(0,s(j)) = rm(0,s(j)) + 1
                if (rm(0,s(j)).le.lr) then
                   rm(rm(0,s(j)),s(j)) = nr
                else
                   write(6,*) 'lr is too small for atom', s(j), lr, rm(1:,s(j))
                   write(6,*) 's', s
                   do k=1,rm(0,s(j))-1
                      write(6,*) r(:,rm(k,s(j)))
                   end do
                   stop
                end if
             end do
             return
          end if
       end do
    else
       trybond: do i=1,b(0,t(im))
          m = im + 1
          t(m) = b(i,t(im))
          do j=1,im
             if (t(j).eq.t(m)) then
                cycle trybond
             end if
          end do
          call visitring(nsi,lb,b,rnk,m,t,lr,rm,maxring,r,nr)
       end do trybond
    end if

    return
  end subroutine visitring

  real(kind=8) function calspecspw(nspecg,speckg,nsi,kd,asi,tkd,pwr)
    implicit none
    integer,intent(in):: nspecg,nsi,speckg(3,nspecg),kd(nsi),tkd
    real(kind=8),intent(in):: asi(3,nsi)
    integer,intent(in):: pwr

    real(kind=8),parameter:: pi = 3.14159265358979323846d0

    real(kind=8):: specstrpow,ph
    complex(kind=8):: strfac
    integer:: isg,it,igx,igy,igz,knsi

    knsi = 0
    do it=1,nsi
       if (kd(it).eq.tkd) knsi = knsi + 1
    end do

    specstrpow = 0.0d0
    do isg=1,nspecg
       igx = speckg(1,isg)
       igy = speckg(2,isg)
       igz = speckg(3,isg)
       strfac = 0.0d0
       do it=1,nsi
          if (kd(it).eq.tkd) then
             ph = 2.0d0*pi*(igx*asi(1,it)+igy*asi(2,it)+igz*asi(3,it))
             strfac = strfac + dcmplx(cos(ph),sin(ph))
          end if
       end do
       specstrpow = specstrpow + &
            ( dble(strfac)*dble(strfac) + aimag(strfac)*aimag(strfac) )**pwr
    end do
    specstrpow = (specstrpow/nspecg)/knsi

    calspecspw = specstrpow
    return
  end function calspecspw

  subroutine calstrpow(nkg,ngxyz,kg,kgmax,nsi,kd,asi,nkd,strpow)
    implicit none
    integer:: nkg,ngxyz,kg(3,ngxyz),kgmax(3),nsi,kd(nsi),nkd
    real(kind=8):: asi(3,nsi),strpow(nkd,ngxyz)

    real(kind=8),parameter:: pi = 3.14159265358979323846d0

    real(kind=8):: ph
    complex(kind=8):: strfac(nkd)
    integer:: knsi(nkd)
    integer:: ig,it,igx,igy,igz,ikd
    complex(kind=8):: strfx(-kgmax(1):kgmax(1),nsi)
    complex(kind=8):: strfy(-kgmax(2):kgmax(2),nsi)
    complex(kind=8):: strfz(-kgmax(3):kgmax(3),nsi)

    do igx=-kgmax(1),kgmax(1)
       do it=1,nsi
          ph = 2.0d0*pi*igx*asi(1,it)
          strfx(igx,it) = dcmplx(cos(ph),sin(ph))
       end do
    end do
    do igy=-kgmax(2),kgmax(2)
       do it=1,nsi
          ph = 2.0d0*pi*igy*asi(2,it)
          strfy(igy,it) = dcmplx(cos(ph),sin(ph))
       end do
    end do
    do igz=-kgmax(3),kgmax(3)
       do it=1,nsi
          ph = 2.0d0*pi*igz*asi(3,it)
          strfz(igz,it) = dcmplx(cos(ph),sin(ph))
       end do
    end do

    knsi = 0
    do it=1,nsi
       knsi(kd(it)) = knsi(kd(it)) + 1
    end do

    do ig=1,nkg
       igx = kg(1,ig)
       igy = kg(2,ig)
       igz = kg(3,ig)
       strfac = 0.0d0
       do it=1,nsi
          strfac(kd(it)) = strfac(kd(it)) &
               + strfx(igx,it)*strfy(igy,it)*strfz(igz,it)
       end do
       do ikd=1,nkd
          strpow(ikd,ig) = strpow(ikd,ig) + 1.0d0/knsi(ikd)* &
               ( dble(strfac(ikd))*dble(strfac(ikd)) &
               + aimag(strfac(ikd))*aimag(strfac(ikd)) )
       end do
    end do
    return
  end subroutine calstrpow
    
end module inspectstr
